#!/usr/bin/perl -w
#
# Name ............: simplebackup.pl
# Type ............: Perl Script
#
# Version .........: 1.5.1
# Last Update .....: 28 February 2005
#
# License .........: GNU GENERAL PUBLIC LICENSE Version 2, June 1991.
#
# Authors ..........: Miguel Angelo Martins Leite,
# Script E-Mail ....: miguel.angelo.public@migas.mine.nu
# Site .............: http://migas.mine.nu/index.php?pag=en.myapps&subpag=simplebackup
#
# Objective .......: Read a configuration file and do a backup of one or more directories
#                    and files and generate one or more a compress files, the compressed
#                    file(s) will be placed on a chosen directory, a remote ftp server
#                    or on a tape device (unix only).
#                    It supports full,incremental or differential backups using
#                    automatic backup session management.
#
# Depends on ......: Perl (v5.6 or over)
#                    Following archive programs (executables must be on your $PATH variable)
#                      * rar/winrar (v3.0 or over) : for rar backup type
#                      * info-zip (v2.3 or over) : for zip backup type
#                      * tar : for tar, tar compress, tar gzip, tar bzip2 backup types
#                      * compress : for tar compress backup type
#                      * gzip : for tar gzip backup type
#                      * bzip2 : for tar bzip2 backup type
#
# Runs On .........: Windows NT4, Windows 2000, Windows XP, 2003 and  All Unix systems
#
#
# Revision History :
#                    1.0.0 RC1 - 10 March 2004 - Miguel Angelo Martins Leite
#                      1. First version.
#
#                    1.0.0 RC2 - 13 March 2004 - Miguel Angelo Martins Leite
#                      1. Loads of tests on Unix (Linux) and Windows (XP) systems, most with
#                         success.
#                      2. Small fixes on Unix systems (Linux).
#                      3. Added support to run scripts/batch files/commands before and
#                         after a backup. Can be used to for example to shut-down a database
#                         do the backup, and in the end bring the database backup on-line.
#                      4. Improved the log messages.
#                      5. Improved a lot the documentation (readme.txt).
#                      6. Added support for zip backup files, using the info-zip software
#                         (http://www.info-zip.org/pub/infozip/) and included a Windows
#                         version to the simplebackup package, this will help Windows users
#                         get the backups running faster.
#
#                    1.0.0 RC3 - 17 March 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Small fixes on the documentation (readme.txt).
#                      3. Fixed a bug that reported a success backup even if the backup failed.
#                         It happened when the write move of the backup file into the final
#                         backup directory failed, fallowed by a success backup reported.
#                         Now if the move of the file fails (lack of space, permissions, etc)
#                         the backup is reported has failed.
#                      4. Added a new field on the log file message, the configuration file name,
#                         this should help people that only use a single log file for all backups.
#                         Thanks to my friend Alexandre Pereira for this suggestion.
#                      5. Added a new option on the configuration file ( backup_compression_ratio )
#                         which enables users to select the compression ratio that they want to use
#                         on their compression files. Options are generic for all archive formats, 
#                         none ; minimal ; average and maximal.
#                         Notice that not all formats support the compression ratio in that case
#                         the option is adapted or ignored, and simplebackup will not fail.
#                         Thanks to my friend Alexandre Pereira for this suggestion.
#
#                    1.0.0 RC4 - 19 March 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Small fixes on the documentation (readme.txt).
#
#                    1.0.0 RC5 - 21 March 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Improved the documentation (readme.txt).
#                      3. Fixed a small message that was miss representing the operation, it was
#                         reporting "Executing before backup procedure using command...", when it
#                         should be reporting "Executing after backup procedure using command..."
#                      4. Improved a bit more the log file, reporting the path him self.
#                      5. Added a new argument (--tconf) and function show_config(), --tconf tests
#                         the a given configuration file (example: simplebackup.pl --tconf file.conf),
#                         and the show_config() function will print the file configuration.
#                         And the most important reason, now i can use this to produce a screenshot
#                         to place on the program site hehehehe :).
#                      6. Centralized the operating system null device on the configuration hash
#                         (%global_config), and this value is now loaded on the read_config() function.
#                      7. Changed the simplebackup.pl intro screen to match the changes.
#
#                    1.0.0 RC6 - 27 March 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Improved the documentation (readme.txt).
#                      3. Small improvement on the log messages.
#                      4. Changed the Windows NT4 [ host_os ] parameter from nt4 into winnt4, it
#                         makes more sense this way, i think.
#                      5. Small fix on Windows NT4, simplebackup was letting Windows move command
#                         print a message when it shouldn't.
#                      6. Added a new warning, if simplebackup is unable to write into the backup
#                         log file, it will complain generally A LOT !, but will not fail.
#                      7. Fixed the zip support on Unix, info-zip does not support the -S parameter
#                         on Unix systems, -S is only for Microsoft Windows Systems, and Mac's.
#                      8. Remove code that added the null device into the command defined by
#                         "run_before_backup" and the "run_after_backup" on the configuration file,
#                         this might cause problems on some Windows programs that simply do not
#                         understand the concept of text redirect "command 2>nul", one example of
#                         this is Microsoft Outlook XP.
#                         If you want to redirect a output of a external command into the null
#                         device you must add that option into the "run_before_arguments" and the
#                         "run_afters_arguments" in the configuration file.
#                      9. Added support for directories and file names with spaces, example
#                         [ d:\test\name 1 with spaces\name 2 with spaces\final directory with spaces ]
#                         or [ /name 1 with spaces/name 2 with spaces/final directory with spaces ].
#
#                    1.0.0 28 March 2004 - Miguel Angelo Martins Leite
#                      1. Final stable version, simplebackup was tested on several operating systems
#                         and several situations. This is a copy of RC6.
#
#                    1.1.0 RC1 - 12 April 2004 - Miguel Angelo Martins Leite
#                      1. Fixed a bug, in the move_file() function, that in some situations could
#                         make it fail to report a missed moved file. In the previous simplebackup
#                         version the bug existed but failed to "work" hehehehe, but because of
#                         changes on the function the bug "worked" ;)
#                      2. Added support for backup's into FTP servers, with url check up, support
#                         for FTP passive mode and open FTP proxies.
#                      3. Added new options to the sample configuration file (simplebackup_example.conf).
#                      4. Added the function local_db_file_history() that manages a simple db file,
#                         that contains the backup file list managed by simplebackup. This is done
#                         to permit simplebackup to operate with "strange" ftp servers that produce
#                         weird file lists (ls/dir commands).
#                      5. Added support to show the FTP details on the --tconf option (test
#                         configuration file).
#                      6. Improvements on the program commentaries.
#                      7. Changed a bit the log messages.
#                      8. Previous backup configuration files (from simplebackup 1.0.0) remain 100%
#                         compatible.
#
#                    1.1.0 RC2 - 13 April 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Improved the documentation (readme.txt).
#                      3. General code change in order to simplify and improve the program performance
#                         in the function read_config().
#                      4. Added extra security to the move_file() function, in order to prevent
#                         writing over already existing files.
#                      5. Small improvement on the log messages.
#                      6. Changed a bit the simplebackup.pl intro screen to match the changes.
#
#                    1.1.0 RC3 - 16 April 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Improved the documentation (readme.txt).
#                      3. Bug fix on the session management, it was incorrectly reducing one backup
#                         session.
#                      4. Bug fix on a situation related to erasing data from the database file, if
#                         simplebackup was using the database file and working in "local file system"
#                         mode.
#                      5. Bug fix on a problem that prevented backup's into the / (root) path of a ftp
#                         server.
#                      6. To make it more coherent with the log_file name, changed the name of variables
#                         from local_db_file_history into local_db_history_file, and to prevent
#                         confusions also change the name of the function local_db_file_history() into
#                         work_local_db_history_file().
#                      7. Improvements on the program commentaries.
#                      8. Changed a bit the log messages.
#
#                    1.1.0 RC4 - 29 April 2004 - Miguel Angelo Martins Leite
#                      1. Added a new security mechanism on the ftp backups. If the upload of the backup
#                         file fails simplebackup will attempt to erase the partially upload file, since
#                         it is damaged.
#                         Plus in situations where simplebackup is using a local database file, there for
#                         using the function local_db_history_file(), if the erase of the partially uploaded
#                         uploaded file fails the file name will be added into the database file. This is
#                         done in order to prevent constant backup failures, backup's will recover more
#                         easily.
#                       2. Fixed the sample configuration file for windows.
#
#                    1.1.0 - 01 May 2004 - Miguel Angelo Martins Leite
#                      1. New stable version of simplebackup, this is a copy of RC4.
#                      2. Main features of this new version compared to version 1.0.0 are:
#                         . Support backup's into FTP servers,
#                         . Possibility to use a local database file to manage backup session,
#                         . Improved program commentaries,
#                         . Improved backup log,
#                         . Previous backup configuration files from version 1.0.0 remain 100%
#                           compatible.
#
#                    1.2.0 RC1 - 11 May 2004 - Miguel Angelo Martins Leite
#                      1. Added support for backup mail reporting, using SMTP mail servers only,
#                         simplebackup can now report the backup results into one or more e-mail
#                         addresses, this means small changes all over the program. Specially
#                         into the functions read_config() and write_file_log().
#                      2. Added new options to the sample configuration file (simplebackup_example.conf).
#                      3. Added support to show the mail details on the --tconf option (test
#                         configuration file).
#                      4. Added the function send_email() that actualy attempts to send the backup
#                         mail report.
#                      5  Added the function get_mail_date() that returns the system date in mail format,
#                         this function is used by the send_email() function.
#                      6. For security reasons the code that wrote the ftp password into the log
#                         file was removed.
#                      7. Changed the default ftp timeout value from 240 to 480, this is a attempt to make
#                         simplebackup even more tolerant to network failures.
#                      8. The mail server timeout value is also 480.
#                      9. Now simplebackup will report the hostname where it's running in the log file and
#                         in the e-mail subject.
#                     10. Bug fix on the readme.txt file, my friend is called Nuno Mu�oz and not Nuno Munhoz.
#                     11. Previous backup configuration files (from simplebackup 1.0.0 and 1.1.0)
#                         remain 100% compatible.
#
#                    1.2.0 RC2 - 13 May 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (XP) systems.
#                      2. Improved the documentation (readme.txt).
#                      3. Small improvement on the log messages.
#                      4. Improved a bit the backup report e-mail.
#                      5. Fixed the english on the sample configuration file.
#                      6. Bug fix on the e-mail subject field, simplebackup was not correctly reporting the
#                         configuration file name, it was reporting teste.conf over and over again.
#
#                    1.2.0 RC3 - 14 May 2004 - Miguel Angelo Martins Leite
#                      1. Continuing testes on Unix (Linux) and Windows (NT4, 2000, XP) systems.
#                      2. Improved a bit the backup report e-mail.
#
#                    1.2.0 - 22 May 2004 - Miguel Angelo Martins Leite
#                      1. New stable version of simplebackup, this is a copy of 1.2.0 RC3 with very little changes.
#                      2. Tested simplebackup under Microsoft Windows 2003 with 100% success, simplebackup is fully
#                         compatible with it.
#                      3. Added support for Microsoft Windows 2003 in the configuration files, where you reported
#                         "host = win2000/xp", you should now report "host = win2000/xp/2003", but simplebackup
#                         will continue to accepted old configuration files with "host = win2000/xp".
#                      4. Main features of this new version compared to version 1.1.0 are:
#                         . Simplebackup can now send backup reports using e-mail (smtp mail servers only),
#                         . Tested and found simplebackup 100% compatible with Microsoft Windows 2003,
#                         . Improved backup log,
#                         . Previous backup configuration files from version 1.0.0 and 1.1.0 remain 100%
#                           compatible.
#
#                    1.2.1 - 01 June 2004 - Miguel Angelo Martins Leite
#                      1. Bug fix on the function get_previous_filedir_name(), it was causing Perl warnings during
#                         unix backups if the backup directory was close to the root level, example /etc.
#                         The error passes without problems on Perl 5.6 but starts producing problems on Perl 5.8.
#
#                    1.3.0 RC1 - 29 June 2004 - Miguel Angelo Martins Leite
#                      1. Testes on Unix (Linux, FreeBsd) and Windows (XP) systems.
#                      2. Improved and re-worked the documentation (readme.txt) adding more stuff and changing
#                         the line size in order for it to be better view on standard console/shell/command line,
#                         meaning 24x80 chars.
#                      3. Added support to show the new options on the --tconf option (test configuration file),
#                         also improved a bit this entire option.
#                      4. Added support for incremental and differential backups. This meant new functions,
#                         check_add_to_backup(), build_present_backup_list(), read_write_management_file(),
#                         rebuild_full_path() is_today(), is_year_bisex(), and several changes all over,
#                         configuration file changes using configuration keys renames and new configuration keys.
#                      5. Added support of file rejections, simplebackup can now do backups selecting what file
#                         types not to backup, this can be used for example to do a user home backup rejecting all
#                         mp3,mov,wav,ogg files.
#                      6. Added support for file size limitation, simplebackup can reject files if they are over
#                         a given size.
#                      7. Added support for specific directory orders by using a special file named
#                         [ simplebackup_rejection.txt ].
#                      8. Added the function remove_excess_spaces() do deal with era tic user input on some
#                         configuration options.
#                      9. Added a new option that allows simplebackup to send the backup log into a external
#                         command using the standard input (stdin), the command or commands can be and do anything.
#                     10. Added a new function check_if_full_path() that will check if the user is using full path's
#                         in the configuration file.
#                     11. Added a pause of 60 seconds that will run after simplebackup runs with success the program
#                         defined by the configuration key [ run_before_backup ]. This is done to make safer backups
#                         since some programs don't free their files immediately after ending.
#                     12. Centralized the operating system file/directory separator, the "d:\my\path" or
#                         "/my/path" char) on the configuration hash (%global_config), and this value is now
#                         loaded in the read_config() function.
#                     13. Removed the all the truncate() perl commands since they where creating problems with perl
#                         5.8.0 causing some file corruption (log file, etc). This appears to be a perl bug.
#                     14. Bug fix, i was using the perl close() function to close directories when i should be using
#                         closedir(). This is tolerable but definitely wasn't good programming.
#                     15. Renamed the function compress_input_dir() into do_backup().
#                     16. Moved the actual compress commands (rar,zip,tar,etc) into the do_compress() function this
#                         will make it simpler maintain the program in the future.
#                     17. Small code improvements all over the program, in order to improve security, performance
#                         (where possible) and to allow for future improvements.
#                     18. Re-worked the error messages that are printed into the shell in order for them to fit
#                         better on a standard console/shell/command line, meaning 24x80 chars.
#                     19. Previous backup configuration files (from simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1)
#                         remain 100% compatible.
#
#                    1.3.0 RC2 - 30 June 2004 - Miguel Angelo Martins Leite
#                      1. Small improvement on the is_today() function.
#                      2. Improved the documentation (readme.txt).
#
#                    1.3.0 RC3 - 04 July 2004 - Miguel Angelo Martins Leite
#                      1. Added code to remove excess \ (Windows) or / (Unix) from the directory paths, this could
#                         cause problems in backup modes that used a backup list of files.
#                      2. Added a new check in order to prevent users from using the same file to do Database backup
#                         management (incremental/differential backups) and  Local Database file history (management
#                         of the backup file names).
#                      3. Bug fix a situation that was forcing full backups in Windows even if the simplebackup was
#                         working in incremental or differential backup modes.
#                      4. Bug fix a backup failure related to empty directories in incremental or differential
#                         backup modes if the tar format was being used.
#                      5. Corrected a small problem on the log file, that was causing a missed typed text in the
#                         last message of each backup.
#                      6. Renamed function is_year_bisex() to is_leap_year(), in portuguese bisesto is leap year, in
#                         english it's something about sex :) hehehe.
#                      7. Changed some log messages.
#                      8. Improved the documentation (readme.txt).
#
#                    1.3.0 RC4 - 11 July 2004 - Miguel Angelo Martins Leite
#                      1. Testes on Unix (Linux, FreeBsd) and Windows (XP) systems.
#                      2. Added a extra security, now if simplebackup is working in incremental or differential
#                         backup mode and if it's a "full backup day" the management file is erased. This is done
#                         to always force a full backup every time until a one finishes with success.
#                      3. Improved the documentation (readme.txt).
#
#                    1.3.0 - 25 July 2004 - Miguel Angelo Martins Leite
#                      1. New stable version of simplebackup, this is a copy of 1.3.0 RC4 with very little changes.
#                      2. Bug fixed a situation that would cause backup failure if simplebackup was doing full backups
#                         and using advanced backup file selections.
#                      3. Added a new field on the test configuration screen that shows up the operating system
#                         detected by simplebackup.
#                      4. For backups that require a backup list and if tar format is used the "-T" parameter will
#                         be used for Linux and Windows. The "-I" for all other operating systems. This will further
#                         improve compatibility with Unix operating systems.
#                      5. This is the most tested simplebackup version to date. Tested under SuSe Linux 8.1, SuSe
#                         Linux 9.1, Red Hat 8.0, Red Hat 9.0, Freebsd 4.1.0, FreeBsd 5.2.1, OpenBsd 3.5, Windows NT4
#                         Windows 2000, Windows XP, Windows 2003.
#                      6. Improved the documentation (readme.txt).
#                      7. Main features of this new version compared to version 1.2.1 are:
#                         . Several small bug fixes and compatibility improvements,
#                         . Added support for incremental and differential backup modes,
#                         . Advanced file selection, simplebackup can now reject backup files by type (extension)
#                           and by size,
#                         . Advanced backup directory selections,
#                         . Better control over the backup configuration files,
#                         . Previous backup configuration files from version 1.0.0, 1.1.0, 1.2.0, 1.2.1 remain
#                           100% compatible.
#
#                    1.3.1 - 06 August 2004 - Miguel Angelo Martins Leite
#                      1. Bug fix on the zip format support, function do_compress(), it was failing in Unix systems
#                         during incremental/differential backup modes due to bad parameters being passed to 
#                         info-zip, solution replace "-S" parameter with "-@".
#                      2. Improved the documentation (readme.txt).
#
#                    1.4.0 RC1 - 15 October 2004 - Miguel Angelo Martins Leite
#                      1. Performance improvement in the incremental, differential modes. This is noticeable on large
#                         backup directory trees that have many files and sub-directories mainly under Microsoft Windows.
#                         Preliminary testing results for a backup directory with 6232 files and directories using a
#                         pentium 4 - 1600 Mhz with 512 Mb ram and using Microsoft Windows XP. With simplebackup 1.3.1
#                         the process to build the backup list toke 6 minutes and 28 seconds to build the backup list,
#                         with simplebackup 1.4.0 RC1 the same list toke only 36 seconds.
#                      2. Added support to do backups into tape devices using the tar format and with the possibility
#                         to use any commands before and after writing to tape, example mt -t /dev/tape0 rewind.
#                         Changes where done a bit all over but mainly in the functions move_file() and
#                         clear_previous_backups().
#                      3. Added code to report the size of the current compressed backup file on the backup log,
#                         thanks to Thomas for this neat feature request.
#                      4. Added code to report the full size of all compressed backup files on the backup log
#                         again thanks to Thomas for this neat feature request.
#                      5. Added a new rejection option, backup_filenames_rejections, now it's possible to reject
#                         files with a given name.
#                      6. Added a new debug option, mail_debug_level, to enable a debug of the connection to the mail
#                         server.
#                      7. Added a new debug option, ftp_debug_level, to enable a debug of the connection to the ftp
#                         server.
#                      8. Renamed the function work_local_db_history_file() to read_write_history_file().
#                      9. Re-write of the read_write_history_file() function in order to make it generate files
#                         that are similar to the simplebackup management files (with file start and end line marker)
#                         this makes history files more robust, and makes it more easy to detect file corruption and
#                         to keep the size of the backup files. This makes the previous versions of the history files
#                         incompatible with this version of simplebackup. The history file is defined by the
#                         configuration key [ local_db_history_file ].
#                     10. Change a bit the function read_write_management_file() to improve the error detection in
#                         the management file. The management file is defined by the configuration key
#                         [ backup_db_management_file ].
#                     11. Better control over weird chars in the configuration file.
#                     12. Change a bit the simplebackup intro screen.
#                     13. Changed the way the temporary file is created, now the temporary file is always created in
#                         the path defined by temporary_dir with the name of the configuration file plus the .tmp
#                         extension, example my_backup.conf.tmp.
#                     14. Fixed a compatibility problem with some external commands that use the = char in their
#                         arguments, for example the oracle exp command makes used of this in the argument
#                         file=some_file_name.dmp.
#                     15. Bug fixed a obscure bug that would show-up if the remote ftp server would change his
#                         configuration in the middle of the backup. This is only related to ftp backups.
#                     16. Bug fixed a bug in the e-mail message system, huge thanxs to Renato Leon for his pacient
#                         testing.
#                     17. Fixed a error message that could miss represent the path into the temporary file. This
#                         only affected the advanced backup modes.
#                     18. Full code revision and improvement.
#                     19. Improved the backup log.
#                     20. Improved the documentation (readme.txt).
#                     21. Small improvement on the program commentaries.
#                     22. Previous backup configuration files from version 1.0.0, 1.1.0, 1.2.0, 1.2.1, 1.3.0,
#                         1.3.1 remain 100% compatible.
#
#                    1.4.0 RC2 - 22 October 2004 - Miguel Angelo Martins Leite
#                      1. Reactivated the log message that reported the total erased files in the function
#                         clear_previous_backups(), this was accidentally turn off in simplebackup 1.4.0 RC1.
#                      2. In tape backups simplebackup will now clear the history file.
#                      3. Improved the documentation (readme.txt).
#
#                    1.4.0 RC3 - 05 November 2004 - Miguel Angelo Martins Leite
#                      1. Changed the separator character used by simplebackup in the history and management
#                         files, the previous char was * and was causing problems with some programs, such as
#                         www.openwebmail.org, that generate file names with the * char, example
#                         miguel-session_look_*_.txt. This further improves simplebackup portability.
#                      2. Bug fixed a problem in the ftp backups that keep on detecting that the ftp debug
#                         level was wrong, uuups sorry.
#                      3. Improved the documentation (readme.txt).
#
#                    1.4.0 RC4 - 06 November 2004 - Miguel Angelo Martins Leite
#                      1. Improved error handling in differential/incremental backup modes, now if a full backup
#                         is forced because the management file is corrupted and if the full backup ends with success
#                         the previous backup session files will be erased and only the current full backup will be
#                         keep. This is the usual procedure of simplebackup during full backup days.
#
#                    1.4.0- 12 November 2004 - Miguel Angelo Martins Leite
#                      1. New stable version of simplebackup, this is a copy of 1.4.0 RC4 with very little changes.
#                      2. Bug fixed a problem in the mail authentication, that causes failures.
#                      3. Introduced some workaround code to deal with users that don't have the standard smtp
#                         authentication library (Authen::SASL), a warning will now be reported on the log file and
#                         simplebackup will no longer fail.
#                      4. Improved the warning messages that show up when there are problems in the mail reporting.
#                      5. Added copies of the Authen::SASL library for unix (from perl cpan) and for windows (from
#                         activestate), this will make life easier for simplebackup users.
#                      6  Improved documentation (readme.txt).
#                      7. Main features of this new version compared to version 1.3.1 are:
#                         . Several small bug fixes and compatibility improvements has a result of a full code revision
#                           and user feedback,
#                         . Improved the speed of incremental/differential backups mainly under windows,
#                         . Added tape backup support (under unix/linux only),
#                         . Added another advanced file selection, simplebackup can now reject backup files by
#                           filename,
#                         . Better control over the backup configuration files and other auxiliary files,
#                         . Greatly improved documentation (readme.txt).
#                         . Previous backup configuration files from version 1.0.0, 1.1.0, 1.2.0, 1.2.1, 1.3.0 and
#                           1.3.1 remain 100% compatible.
#
#                    1.4.1- 18 November 2004 - Miguel Angelo Martins Leite
#                      1. Fixed a bug that could cause a simplebackup crash during the ftp backup, if the ftp server
#                         blocks the connection right before the backup file upload. This was corrected in the function
#                         sum_all_backups() and was introduced in simplebackup 1.4.0, sorry !.
#                      2. Bug fixed several small problems in ftp backups that could occurred if simplebackup was
#                         running under a windows operating system.
#                      3. Now the message that reports "Disconnecting from ftp server" will only appear if we actually
#                         managed to connect in the first place.
#
#                    1.5.0 RC1 - 19 January 2005 - Miguel Angelo Martins Leite
#                      1. Huge speed improvement of incremental/differential under any supported operating system.
#                      2. Removed the requirement in simplebackup to always have the libnet perl library installed, this
#                         library is still required to use ftp backups and email reporting, but if the user doesn't use
#                         those features then simplebackup will work without errors.
#                      3. Added code to deal with perl installations that don't have libnet perl library installed, if ftp
#                         backups or email reporting are used simplebackup will report a error but will no longer crash.
#                      4. Removed the configuration parameters run_before_arguments and run_after_arguments, external
#                         commands and arguments can be fully declared in the configuration keys run_before and run_after.
#                         This makes simplebackup simpler ;).
#                      5. Bug fixed a problem on the function clear_previous_backups() that made simplebackup crash if the ftp
#                         connection dropped, notice that this bug did not affected any older simplebackup version.
#                      6. Added code so that simplebackup can now automatically detect the operating system, in the function
#                         get_os_type().
#                      7. Added code so that simplebackup can now automatically configure the temporary directory, in the
#                         function get_tmp_dir().
#                      8. Added code so that simplebackup can now automatically attribute the full path into the auxiliary
#                         files, the path is based on 3 things: the configuration parameter temporary_dir, the configuration
#                         file name and on the auxiliary file type.
#                      9. Improved the process that detects if a given backup file (the actual files generated by simplebackup
#                         that contain the backup data) is part or not of a backup. This detection logic is now centralized
#                         in the functions check_session_filename() and test_date().
#                     10. The backup file names will now contain a new meta data field that reflects some important extra data.
#                         Previously the logic was session_number.directory_name.extension, example [ 2.documents.zip ], the new
#                         logic is directory_name.meta_data.extension, example
#                         [ documents.2 full monday 14-12-2004 crc1234131.zip ].
#                     11. Added support for the backup of single files, previously simplebackup could only backup entire
#                         directories.
#                     12. Added multi-directory and multi-file backup options, now one can do a backup of more than one directory
#                         and file per configuration file.
#                     13. Added support to read the backup list from a separated text file.
#                     14. Added the function get_filename_session_metadata() that is used to get the session number and file
#                         metadata of any backup file, previously this code was spread and duplicated in several other functions.
#                     15. Added the function sum_backup_types() that counts the backup types (full, incremental, differential)
#                         available.
#                     16. Changed what goes into the label section of the management file, previously it was the backup directory
#                         name, now it's the crc32 number build from the path into the backup directory or file. This is part of
#                         the support of the multi-directory and multi-file backup option.
#                     17. Changed the error situation that made backups stop into a warning. If a problem is detected while building
#                         the backup list (incremental/differential and other advanced backup modes) a warning message is written
#                         into the log and the offending file or directory is ignored. This could occurred for example if the user
#                         is doing the backup of a directory and has lack of permissions for a given file.
#                     18. The function clear_previous_backups() was changed to support the multi-directory and multi-file backup
#                         options.
#                     19. Renamed the configuration key input_backup_dirs to input_backup.
#                     20. Updated the function remove_excess_spaces() to support the < separator char.
#                     21. Added the configuration parameter ftp_connect_retries. This defines the number of times that
#                         simplebackup will retry the connection into the backup ftp server, and has a default value of 6.
#                         This will make ftp backups more robust under bad network conditions.
#                     22. During ftp backup simplebackup will now automatically switch the passive mode value and retry if the
#                         ftp connection fails more than 3 times. Again this will make ftp backups more robust under bad
#                         network conditions.
#                     23. Moved the code that detects the missing Authen::SASL module from the send_email() function into the
#                         read_config() function. So now simplebackup will produce a error and not do a backup if one tries
#                         to do a backup using mail authentication without the required module available. The Authen::SASL
#                         module is required in order to support mail servers that request user authentication.
#                     24. Improved the process that compresses the backup data in function do_compress(), now simplebackup will
#                         erase the previous temporary archive files that could be left over if simplebackup crashed or if
#                         simplebackup was killed by the user.
#                     25. Added the -p parameter into the tar command backups, this will make simplebackup backups contain all
#                         permission information.
#                     26. Improved the test configuration option (--tconf parameter).
#                     27. Improved the log file with more error messages.
#                     28. General code improvements and simplifications.
#                     29. Improved the documentation (readme.txt).
#                     30. Improved the configuration file example (simplebackup_example.conf).
#                     31. Previous backup configuration files (from simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1, 1.3.0, 1.3.1, 1.4.0,
#                         1.4.1) remain 100% compatible.
#
#                    1.5.0 RC2 - 21 January 2005 - Miguel Angelo Martins Leite
#                      1. Bug fixed the problem with the backup list coming from external a file that prevented any valid
#                         file from being loaded. Huge thanks to Dr. A.M. Genaev for this bug fix.
#
#                    1.5.0 RC3 - 26 January 2005 - Miguel Angelo Martins Leite
#                      1. Bug fixed a non critical problem in the ftp backups. If a problem appeared on the initial ftp connection
#                         to the ftp server simplebackup would double the connections.
#                      2. Added the parameter --econf, function create_edit_config(), this will allow simplebackup create (if no file
#                         exists) and edit a given configuration file using the operating system text editor, vi under unix and notepad
#                         under windows.
#                      3. Improved the way simplebackup detects "out of sync" backups, a full backup is also forced if incremental or
#                         differential backups files are detected and the backup is working in the other backup mode (in differential
#                         or incremental).
#                      4. Improved the documentation (readme.txt).
#                      5. Uncompressed all files inside of the simplebackup package, this makes the installation a little bit easier.
#
#                    1.5.0 - 12 February 2005 - Miguel Angelo Martins Leite
#                      1. New stable version of simplebackup, this is a copy of 1.5.0 RC3 with very little changes.
#                      2. Bug fixed a problem in the way simplebackup processed the input_backup parameter in configurations
#                         with multiple directories/files and the use of space chars. Thanks to John for reporting this.
#                      3. Improved some log messages.
#                      4. Main features of this new version compared to version 1.4.1 are:
#                         . Huge speed improvement during incremental/differential backups under any supported operating system,
#                         . Support for the backup of multi-directory and multi-individual files on a single backup configuration,
#                         . Improved backup file names,
#                         . Option to include the hostname on the backup file names,
#                         . Simpler backup configuration with less configuration keys and the use of automatic detections, example
#                           detect the operating system,
#                         . Added the parameted --econf, that can create and edit configurations files, users are no longer required
#                           to use the example configuration files,
#                         . The backup list can be read from a simple text file,
#                         . Improved documentation (readme.txt),
#                         . Previous backup configuration files from version 1.0.0, 1.1.0, 1.2.0, 1.2.1, 1.3.0, 1.3.1, 1.4.0, 1.4.1
#                           remain 100% compatible.
#
#                    1.5.1 - 28 February 2005 - Miguel Angelo Martins Leite
#                      1. Bug fixed a problem that could lead simplebackup to dump warnings into the prompt if it was unable to
#                         get the file modification time, this occurred during incremental/differential backups. Simplebackup can
#                         fail to get the file modification time because of, lack of permissions, file system errors or if it gets files
#                         with "weird" dates, meaning dates before "00:00 January 1, 1970 GMT" (unix epoc), in those cases a warning is
#                         written into the backup log and the file is ignored. This is a very rare problem, and is simple perl warning.
#                         Thanks to Dave Eduards for reporting this bug.
#                      2. Now if the user is using the configuration key hostname_on_backup_files and the auto configuration options the
#                         hostname will be also added into the auxiliary files (log file, management file, history file).
#                      3. Improved some log messages.
#                      4. Improved documentation (readme.txt),
#
#
# Future Plans ....:
#                   * Adding support for doing restores, presently restores must be done manualy.
#                   * Study solutions to implement email backups.
#                   * Study a way to backup the cisco routers configuration.
#                   * Adding support for backups using the cpio file format.
#                   * Adding support for tape backups using more formats than tar.
#                   * Adding support scp or sftp (secure copy protocol), in order to backup
#                     over the net in a more secure way.
#
#

##########################
# Modules
##########################
use strict; # implement a more strict language
use Sys::Hostname; # to enable the possibility to get the machine hostname
use Fcntl qw(:DEFAULT :flock);  # file locking for the logs and other written files


##########################
# Global Variables
##########################
my $script_version = "1.5.1";
my $script_last_update = "28 February 2005";
my $script_license = "GNU GENERAL PUBLIC LICENSE Version 2, June 1991.";
my $script_authors = "Miguel Angelo Martins Leite";
my $script_email = "miguel.angelo.public\@migas.mine.nu";
my $script_site = "http://migas.mine.nu/index.php?pag=en.myapps&subpag=simplebackup";
my %global_config = (    # script configuration hash
    "configuration_file"        => "",    # Path into the configuration file, ex: /some/path/a_file.conf
    "configuration_file_name"   => "",    # Configuration file name, ex: a_file.conf
    "host_os"                   => "",    # Operating system running the script
    "host_null_device"          => "",    # Operating system null device
    "host_file_separator"       => "",    # Operating system file separator,   \ for windows and / for unix
    "backup_format"             => "",    # Backup type: rar, zip, tar.bz2, tar.gz, tar.Z, tar
    "backup_compression_ratio"  => "",    # Chose backup compression ratio:
                                          #   none = no compression = 0 (rar, info-zip) or 1 (gzip,
                                          #   bzip2)
                                          #   minimal = minimal compression = 1
                                          #   average = average type of compression = 3 (rar) or
                                          #             6 (info-zip,gzip,bzip2)
                                          #   maximal = maximal compression = 5 (rar) or
                                          #   9 (info-zip, gzip, bzip2)
                                          # Note tar and compress formats are not affected by the
                                          # argument
    "backup_real_ratio"         => "",    # The actual number to be passed into the compressor,
                                          # 0 up to 5 or 9
    "backup_mode"               => "",    # Type of backup : full, incremental or differental, comes
                                          # from the configuration file
    "build_backup_list"         => "",    # If build_present_backup_list() will be used to build the
                                          # backup list, values yes/no.
    "backup_real_mode"          => "",    # Actual type of backup used : full, incremental or differental,
                                          # for example it can be a full backup even in incremental or
                                          # differental backups if we have arrived at the full backup day
    "full_backup_day"           => "",    # When will the full backup occured.
    "backup_db_management_file" => "",    # Path into the management file that is used in incremental
                                          # or differental backups to match file/link/directories
                                          # modification times.
    "backup_extension_rejections" => "",  # List of types of backups that are suppose to be rejected
                                          # on backups... this is given in extensions, example
                                          # txt,zip,mp3
    "backup_filenames_rejections" => "",  # List of files to reject, example my_boss.jpg,secret.doc
    "backup_file_size_limit"      => "",  # Limit the maximal file size to backup (in kbytes)
    "backup_file_size_limit_bytes" => "", # Limit the maximal file size to backup (in bytes)
    "backup_directory_rejections" => "",  # Search and use simplebackup_rejection.txt files
    "backup_working"            => "",    # Where is the backup file is going, configured on the
                                          # read_config() function.
                                          #  local file system = for local file system directory, this
                                          #                      also included windows mounted drives
                                          #                      and unix mounted directories (devices,
                                          #                      nfs, samba, etc).
                                          #                ftp = if the backup is going into a remote
                                          #                      ftp server.
                                          #                scp = if the backup is going into a remote
                                          #                      server using scp (secure copy)
    "ftp_server"                => "",    # Remote ftp server.
    "ftp_debug_level"           => "",    # FTP debug level, none/full, default is none
    "ftp_debug_level_real"      => "",    # The actual parameter that is passed into the ftp module
    "ftp_server_port"           => "",    # Remote ftp server port number.
    "ftp_username"              => "",    # FTP Username.
    "ftp_password"              => "",    # FTP Password.
    "ftp_directory"             => "",    # FTP directory where we will upload the backup file.
    "ftp_passive_mode"          => "",    # FTP Passive mode [ 0 ; 1 ].
    "ftp_passive_mode_real"     => "",    # The real parameter passed into the Net:FTP, here is
                                          # only [ 0 ; 1 ].
    "ftp_proxy"                 => "",    # FTP Proxy, must work in open mode.
    "ftp_timeout"               => "",    # Ftp timeout
    "ftp_connect_retries"       => "",    # FTP retries that simplebackup will use during connect
    "tape_device"               => "",    # The tape device, example /dev/tp0
    "tape_command"              => "",    # The actual command that places the stuff on tar, default is tar
    "tape_before_backup"        => "",    # Command to execute before running the command that places
                                          # data on the tape, example mt -t /dev/tp0 rew
    "tape_after_backup"         => "",    # Command to execute after running the command that places
                                          # data on the tape, example mt -t /dev/tp0 offline
    "mail_type"                 => "",    # The backup server, empty for no e-mail reporting, smtp for smtp mail server type
                                          # other backup server types (example imap) are not yet supported
    "mail_server"               => "",    # Mail server name/address.
    "mail_debug_level"          => "",    # Mail debug level, none/full, default is none
    "mail_debug_level_real"     => "",    # The actual parameter that is passed into the mail module (0/1)
    "mail_port"                 => "",    # Mail server port (default is 25)
    "mail_timeout"              => "",    # Mail server timeout
    "mail_username"             => "",    # Mail username (might be required to smtp login).
    "mail_password"             => "",    # Mail password, if it's not empty simplebackup will attempt
                                          # to login into the mail server.
    "mail_from_address"         => "",    # E-Mail address from,    where the report comes from.
    "mail_to_addresses"         => "",    # E-Mail address(es) to,  where the report is/are going.
    "input_backup"              => "",    # The list of files and directories to backup
    "output_backup"             => "",    # Where the final compress file will be placed
    "temporary_dir"             => "",    # Where the data shall be compress, this is a temporary
                                          # destiny for the compressed file
    "keep_last_n_files"         => "",    # Number of the last backup files that shall be keep
                                          # 5 for example the script shall erase all compress backup
                                          # file except the last 5 sessions + the current backup
                                          # session
                                          # Notes : -1  -> all backup files shall be keep
                                          #          0  -> only the present backup file shall be keep
                                          #         If the backups are working in incremental or
                                          #         differential modes this value is automanaged.
    "run_before_backup"         => "",    # Optional script/batch file/command to run before backup
    "run_after_backup"          => "",    # Optional script/batch file/command to run after backup
    "pass_log_to_external_cmd"  => "",    # Pass the log into a external program via stdin... similar to
                                          # cat log.txt | my_command
    "log_file"                  => "",    # Where the log file shall live
    "hostname"                  => "",    # Machine name where simplebackup is running, replaced by "unavailable"
                                          # if the hostname cannot be obtained
    "hostname_on_backup_files"  => "",    # If were are going to include the hostname on each backup file
    "local_db_history_file"     => "",    # Database file that contains the backup files managed by
                                          # simplebackup, this USUALLY is not REQUIRED
);


##########################
# Functions
##########################

#
# Show a small command line help
# Will return: 0
# Explain: 0 is sucess
sub show_commands {

    my $my_script_version = $_[0]; # script version parameter
    my $my_script_last_update = $_[1]; # script lastupdate version
    my $my_script_license = $_[2]; # script license
    my $my_script_authors = $_[3]; # script authors
    my $my_script_email = $_[4]; # script email
    my $my_script_site = $_[5]; # script site

    print <<EOF;

  Name .....: simplebackup.pl v$my_script_version ($my_script_last_update)

  License ..: $script_license
  Author ...: $my_script_authors
  E-Mail ...: $my_script_email
  Site .....: $my_script_site

  Objective : Do a simple backup of directories and files by compressing each
              backup directory or file listed on the backup list into a
              separated archive file and placing each one, into some other
              path, machine or tape device (unix only), this also includes
              ftp servers.
              The archive file can be of one of the following types: rar,
              zip, tar, tar + compress, tar + gzip or tar + bzip2.

  Creating/Editing Configuration :
    Unix ...> simplebackup.pl --econf full/path/config_file.conf
    Windows > perl -w simplebackup.pl --econf full\\path\\config_file.conf

  Testing Configuration :
    Unix ...> simplebackup.pl --tconf full/path/config_file.conf
    Windows > perl -w simplebackup.pl --tconf full\\path\\config_file.conf

  Backup Execution :
    Unix ...> simplebackup.pl --conf full/path/config_file.conf
    Windows > perl -w simplebackup.pl --conf full\\path\\config_file.conf

  Notes :
    1 > Don't forget to edit and change the configuration file.
    2 > You can use this program to create automatic backups using Unix Cron,
        Windows Task Scheduler, etc.

   The crc32 algorithm used is based in the Swissknife library (crc32.pm
   module) from Wolfgang Fleischmann wfl\@ebi.ac.uk

EOF

    return 0; # always return sucess
}


#
# Create or Edit a configuration file using the default operating system editor
# vi for Unix and notepad for windows
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 is failure
sub create_edit_config {

    my $file_path = $_[0]; # this is the first function parameter $_[1] is the second etc etc etc
    my $my_script_version = $_[1]; # script version parameter
    my $error_flag = 0; # to mark errors

    print "\n";
    # test if the file is valid
    if ( -d $file_path or -l $file_path or -c $file_path or -b $file_path ) {
        print "Error - the path is not pointing into a valid file.\n\n";
        return 1; # "file" is a directory, link (unix), caracter device (unix) or block device (unix)
    }

    if ( ! -f $file_path ) { # if the file does not exists create a new one
        my $my_option =""; # to contain the user input
        while ( $my_option ne "y" and $my_option ne "n" ) {
            print "File does not exist would you like me create it ?\n";
            print "[ y/n ] ";
            $my_option = <STDIN>;
            chop $my_option; # remove the new line char
        }
        if ( $my_option eq "y" ) { # if user wants to create the file
            # create the log file in > write mode
            open(configuration_file_handle, ">$file_path") or $error_flag = 1;
            if ( $error_flag == 0 ) {
                #lock the file for exclusive script use
                flock (configuration_file_handle, LOCK_EX) or $error_flag = 1;
                # write the new configuration file (using default/standard values)
                print configuration_file_handle <<EOF, or $error_flag = 1;
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                                                   #
# Simplebackup.pl $my_script_version configuration file                          #
# Comments must be made with the char # on the start of the line to #
# prevent weird backup's always define full path's                  #
#                                                                   #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


#
# [ Backup mode ]
# Simplebackup supports 3 different backup modes:
#
# full         > In this mode simplebackup will always compress everything
#                that is suppose do be backup.
#                To restore your data, you only need to restore/uncompress a
#                single backup file, but the price is a higher usage of
#                backup space and time. If you use this mode don't forget to
#                check the configuration key [ keep_last_n_files ].
#   ** Logic: full_backup -> full_backup -> full_backup -> etc **
# incremental  > In this mode simplebackup will make a full backup (at a
#                chosen date), and from there on it will only make backups of
#                the data that was changed in relation to the previous
#                backup (full or incremental).
#                You gain a smaller and faster backups but the price is the
#                restore, to do a full restore you need to restore the full
#                backup and all the incremental backups in the correct
#                sequence.
#   ** Logic: full_backup -> incremental -> incremental -> etc **
# differential > In this mode simplebackup will make a full backup (at a
#                chosen date), and from there on it will only make backups of
#                the data that was changed in relation to the previous full
#                backup. This mode is a mixture between full backup and
#                incremental backup.
#                To do a full restore you need to restore the the full backup
#                and the most recent incremental backup.
#   ** Logic: full_backup -> differential -> differential -> etc **
#
# Default value is: full
# Valid values are: full ; incremental ; differential
backup_mode = full

#
# [ Full backup day ]
# When the full backup day occurs. This option is only used if simplebackup
# is working in incremental or differential backup mode, so check the
# [ backup_mode ].
# Note, when a successful full backup occurs all previous full, incremental
# or differential backups related are erased, also notice that you can do a
# single full backup per year... but one must be really nuts to use this.
#
# Default value is: monday
# Valid values are: monday, tuesday, wednesday, thursday, friday, saturday,
#                   sunday, first_day_month, last_day_month, first_day_year,
#                   last_day_year
full_backup_day = monday

#
# [ Database backup management file ]
# This is required if you are using incremental or differential backups, it
# must point into a file, or use the auto parameter. This file will contain
# the list of the backup data and it's used by simplebackup to keep a list of
# directories and files in order for it to know what do backup and what not.
# The management file is created and managed by simplebackup program.
# Notice: DO NOT share a single database file across different backup
# types/directories, use one database file per backup configuration file.
# The auto option creates a path based on the temporary_dir parameter on
# this configuration file name and by adding the .management.db extension.
#
# Default value is: auto
# Valid values are: auto (to automaticly define the file) ; any valid full
#                   path into a file
# Examples: c:\\temp\\some_management_file.db
#           /tmp/some_management_file.db
#           auto
backup_db_management_file = auto

#
# [ Backup format ]
#
# Valid values are: rar ; zip ; tar ; tar.Z ; tar.gz ; tar.bz2
backup_format = zip

#
# [ Backup compression ratio ]
# Note that on simple tar backup this value is not relevant, in that case you
# can set it to any value.
#
# Valid values are: none ; minimal ; average ; maximal
backup_compression_ratio = maximal

#
# [ Backup list ]
# Full path into directories and files to backup or a "url" into into a text
# file containing the backup list.
# Notice, if used, the backup file list must be a ascii text file containing
# one line for every directory or file to backup, also everything must be
# done declaring full path's.
#
# Valid values are: any valid directory, file or list of directories and
#                   files separated by the < char or a full path "url"
#                   into file that contains a list of what to backup
# Examples: c:\\some_directory
#           c:\\some_directory < d:\\another_directory
#           c:\\some_directory < d:\\another_directory < e:\\ok\\some_file.txt
#           c:\\a_file.doc < e:\\ok\\some_file.txt
#           /home/some_directory
#           /home/some_directory< /home/another_directory
#           /home/some_directory< /home/another_directory< /etc/my.conf
#           /etc/my.conf, /root/my_file.txt
#           file://\@/root/files/my_backup_list.txt
input_backup =

#
# [ Compressed backup files final destiny ]
# Where simplebackup places the backup data.
#
# Examples:
#      /some_directory      (under Unix type of systems)
#      d:\\some_directory    (under Microsoft type of systems)
#      FTP Backups
#       ftp://some_username:some_password\@some_server:8021/some_directory
#       ftp://some_username:some_password\@some_server/some_directory
#       Under ftp backup you must type a standard URL, here is the
#       explanation (the separated fields) :
#         ftp://          ; marks that is a ftp backup type,
#         some_username   ; The remote ftp user, the user that must be able
#                           to list,write and delete from the server,
#         some_password   ; The password associated with the ftp username,
#         some_server     ; Ftp server example ftp.netscape.com
#         8021            ; Ftp server port, you can leave this empty if
#                           the server is on the standard ftp port (21),
#         /some_directory ; the directory on the FTP server that will contain
#                           the backup file.
#      Tape Backups (Unix Only)
#       tape://\@/dev/some_tape_device
output_backup =

#
# [ Temporary directory ]
# Defines where the compress file and other temporary files are generated.
# This can be automaticly resolved by using the option "auto".
#
# Default value is: auto
# Valid values are: auto ; any valid path into a directory
temporary_dir = auto

#
# [ Backup management number ]
# Defined how many backup sessions to keep notice this option is only
# configurable if simplebackup is working in full backup mode
# [ backup_mode = full ]. In incremental or differential modes this value is
# automatically managed.
# The valid numbers are:
#   -1               > keep all full backup sessions, no previous backup
#                      are erased.
#   0                > just keep the current full backup session.
#   any_other_number > for example 5 will keep the last 5 + the current
#                      full backup session.
#
# Valid values are: -1 ; 0 ; any_other_number
keep_last_n_files = -1

#
# [Run before backup ]
# Optional script/batch file/command that you can run before starting to
# backup.
# Note that this command must end in sucess in order for the backup to start,
# if it fails simplebackup will not continue.
#
# Valid values are: any valid command ; nothing at all
# Examples: unix /etc/rc.d/mysql.server stop mysql ;
#           windows > net stop mysql
run_before_backup =

#
# [ Run after backup ]
# Optional script/batch file/command that you can run after the backup ended.
# Notice that this always runs even if the backup failed.
#
# Valid values are: any valid command ; nothing at all
# Examples: unix > /etc/rc.d/mysql.server start ;
#           windows > net start mysql
run_after_backup =

#
# [ Backup log file ]
# Full path into the backup log file or auto (to create it automaticly).
# This file is generated in ascii format (Unix or Windows depending on the
# host operating system).
# Note that you can share a single log file across all backup's, but it's not
# recommended since it will tend to mix the log messages, but it's up to you.
# The auto option creates a path based on the temporary_dir parameter on
# this configuration file name and by adding the .log extension.
#
# Valid values are: auto (to automaticly define the file) ; any valid full
#                   path into a file
# Examples: c:\\my_logfile.log
#           /home/my_user/my_backup_data.log
#           auto
log_file = auto

#
# [ Mail server ]
# If you want to send backup reports using email then you must configure here
# a valid mail server relay.
# Note that you must have the correct permissions on the server in order for
# this to work.
#
# Example: some_smtp_server.com
# Valid values are: any valid domain name, mail server or ip address
mail_server =

#
# [ Mail from address ]
# The e-mail address where the reports will be generated.
#
# Valid values are: any valid e-mail address
# Examples: backup_reporting\@mail_server.com
mail_from_address =

#
# [ Mail to addresses ]
# The e-mail address or addresses where the reports will be send.
#
# Valid values are: any valid e-mail address or addresses separated by the ,
# char.
# Examples: some_adress\@somewhere.com
#           some_adress_1\@somewhere.com, some_adress_2\@anywhere.com
mail_to_addresses =









# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   The following parameters are not required to be present,  #
#   use them ONLY if you absolutely require them.             #
#                                                             #
#   To use them uncomment and configure the option.           #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#
# [ Host Operating System ]
# Defines the operating system where simplebackup is running, you can use
# this if the autodetection is not working properly.
#
# Default value is: auto
# Valid values are: auto ; winnt4 ; win2000/xp/2003 ; unix
#host_os = auto

#
# [ Backup extension rejections (Backup advanced parameters) ]
# This option can be used to reject files of some type or types, the type is
# the file extension.
# Notice this option tends to make backups slower.
#
# Examples: zip
#           zip,mp3,mpg,mpeg
#backup_extension_rejections =

#
# [ Backup filenames rejections (Backup advanced parameters) ]
# This option can be used to reject files with a given name.
# Notice this option tends to make backups slower.
#
# Examples: my_boss_image.png
#           my_boss_image.png,boss_party.png,my secret.doc
#backup_filenames_rejections =

#
# [ Backup file size limit (Backup advanced parameters) ]
# This option can be used to limit the maximal size of the files that
# simplebackup will backup. File size is given in kbytes.
# Notice this option tends to make backups slower.
#
# Default value is: 0 (no limit)
# Valid values are: any valid positive number or 0 (no limit)
# Examples: 1024 (1 megabyte), 15360 (15 megabyte),
#           512 (half a megabyte), 102400 (100 megabytes)
#backup_file_size_limit = 0

#
# [ Backup directory rejections (Backup advanced parameters) ]
# If you want simplebackup to use [ simplebackup_rejection.txt ]
# files. This file contains orders on what NOT to backup on a given
# directory.
# Notice this option tends to make backups slower.
# Also notice if you use [ simplebackup_rejection.txt ] files you must write
# a valid order inside of it in a single line and add the file into the
# directory you want to implement it, for valid examples check the
# [ directory_rejections ] directory supplied with simplebackup.
# package.
# Valid orders for simplebackup_rejection.txt files are:
#                  all   -> to reject everything inside directory,
#                  files -> ignore all files,
#        sub-directories -> ignore all directories inside directory,
#                  touch -> dummy option just to add a empty directory.
#
# Default value is: no
# Valid values are: yes ;  no
#backup_directory_rejections = no

#
# [ Mail debug level ]
# The amount of messages related to the mail server connection that
# will be printed to the shell/prompt.
# This should only be useful to debug problems with the mail backup
# reporting, so in general you should accept the default configuration which
# is none.
#
# Default value is: none
# Valid values are: none ; full
#mail_debug_level = none

#
# [ Mail server port (Mail advanced parameters) ]
# The port where the mail server is running.
#
# Default value is: 25
# Valid values are: any numeric number over zero, example 25
#mail_port = 25

#
# [ Mail server Timeout (Mail advanced parameters) ]
#
# Default value is: 480
# Valid values are: any numeric number over zero, example 480
#mail_timeout = 480

#
# [ Mail username (Mail advanced parameters) ]
# If your mail server requires authentication then activate this option.
#
# Valid values are: any valid mail username
#mail_username =

#
# [ Mail password (Mail advanced parameters, to match the mail username) ]
# If your mail server requires authentication then activate this option.
#
# Valid values are: any valid mail password
#mail_password =

#
# [ FTP debug level ]
# The amount of messages related to the ftp server connection that will be
# printed to the shell/prompt.
# This should only be useful to debug problems with the ftp backup, so in
# general you should accept the default configuration which is none.
#
# Default value is: none
# Valid values are: none ; full
#ftp_debug_level = none

#
# [ FTP Passive mode (Ftp advanced parameters) ]
# Note that if value is not declared the default is [ no ], plus passive mode
# is only required for some firewall and network systems, first attempt a
# backup in "normal mode".
#
# Default value is: no
# Valid values are: yes ; no
#ftp_passive_mode = no

#
# [ FTP Timeout (Ftp advanced parameters) ]
#
# Default value is: 480
# Valid values are: any numeric number over zero, example 480
#ftp_timeout = 480

#
# [ FTP Connect Retries (Ftp advanced parameters) ]
# The number of retries that simplebackup will make connecting into a ftp
# server (during ftp backups). Please notice that after 3 connect attempts
# simplebackup will automatically switch the passive mode (ftp_passive_mode)
# reversing the configured value from yes or no.
#
# Default value is: 6
# Valid values are: any numeric number over zero, example 6
#ftp_connect_retries = 6

#
# [ FTP Proxy (Ftp advanced parameters) ]
# Note that if value is not declared simplebackup will not use a ftp proxy.
# The ftp proxy must be open, proxy authentication is not supported.
#
# Valid values are: any valid hostname + port
# examples: my_ftp_proxy:8021
#ftp_proxy = some_proxy_server:2121

#
# [ Execute before tape writing (Tape advanced parameters) ]
#
# Command or commands to execute before starting to write into the tape
# device.
#
# Default value is: [ mt -t your_tape_device rewind ]
# Valid values are: any command ; none (if you don't want to run anything)
#tape_before_backup = none

#
# [ Execute after tape writing (Tape advanced parameters) ]
#
# Command or commands to execute after ending writing into the tape
# device.
# Notice this only runs if you are doing a backup into a tape device
# and if the tape was used to backup at least one backup file.
#
# Default value is: [ mt -t your_tape_device offline ]
# Valid values are: any command ; none (if you don't want to run anything)
#tape_after_backup = none

#
# [ Local Database file history ]
# This should only be used if you are having problems with strange ftp
# servers that produce results that confuse simplebackup or if you are doing
# a backup into a tape device.
# Note DO NOT share a single database file across different backup
# types/directories, this can cause problems with the backup management,
# use one database file per backup configuration file. Also remember this
# file must be different from the one defined by backup_db_management_file.
# The auto option creates a path based on the temporary_dir parameter on
# this configuration file name and by adding the .history.db extension.
#
#
# Valid values are: auto ; any valid full path into a file
#local_db_history_file = auto

#
# [ Hostname on backup files ]
# If we are going to include your hostname on each generated backup
# file, and auxiliary files (if using the auto option).
#
# Default value is: no
# Valid values are: yes ; no
#hostname_on_backup_files = no

#
# [Pass log to external command ]
# This will enable simplebackup to pass the backup log into any desired
# command using the standard input (stdin).
# Your command must accept input from the stdin and then do what ever it
# want's to do with it.
# Notice this is similar to doing:
#   in unix ...: cat my_file.log | my_command -arguments
#   in windows : type my_file.log | my_command /arguments
#
# Valid values are: any valid command that acesses input from stdin
#                   (standard input)
#pass_log_to_external_cmd =
EOF
                print "File was successfully generated\n";
                close (configuration_file_handle); # close the file
            }
            else {
                print "Error - unable to write the new configuration file\n\n";
            }
        }
    }

    if ( $error_flag == 0 and -f $file_path ) { # if file exists and no error occured
        my $try_edit =""; # contain the operating system edit command
        $try_edit = "vi \"$file_path\"" if ( get_os_type() eq "unix" ); # attempt to use the operating system vi command
        $try_edit = "notepad \"$file_path\"" if ( get_os_type() ne "unix" ); # attempt to use the operating system notepad command
        print "Editing the file using [ $try_edit ]\n";
        exec ($try_edit) or $error_flag = 1; # execute the command, with exec so simplebackup will not detect a non existing problem
        print "Error - unable to edit the file, please attempt to edit the configuration file using any standard ascii text editor.\n\n" if ( $error_flag != 0);
    }

    return $error_flag; # always return sucess
}


#
# Test the command line parameters/commands
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 failure
sub test_commands {

    my @parameters = @_; # the prompt parameters
    # scalar counts a array
    if ( scalar(@parameters) != 2 or ($parameters[0] ne "--conf" and $parameters[0] ne "--tconf" and $parameters[0] ne "--econf") ) {
        return 1; # return failure
    }
    else { # --conf file.conf > is read config and do backup ; --tcont file.conf is read config and show is values
        return 0; # return sucess
    }
}


#
# Read (and test) the configuration file into the global var %GLOBAL_CONFIG
# Will return: 0 ; 1 ; 2 ; 3 ; 4 ; 5 ; 6 ; 7 ; 8 ; 9 ; 10 ; 11 ; 12 ; 13 ; 14 ; 15 ; 16 ; 17 ; 18 ; 19 ; 20 ; 21 ; 22 ; 23 ; 24 ; 25 ; 26 ; 27 ; 28 ; 29 ; 30 ; 31 ; 32 ; 33 ; 34 ; 35 ; 36 ; 37 ; 38 ; 39 ; 40 ; 41 ; 42 ; 43 ; 44
# Explain: 0 is sucess ; any other is failure
sub read_config {

    my $file_path = $_[0]; # this is the first function parameter $_[1] is the second etc etc etc
    my $global_config_p = $_[1]; # pointer into the configuration hash
    my $error_flag = 0; # file error flag
    my $error_file_directory = ""; # to keep the description of which directory/file failed

    $$global_config_p{configuration_file} = $file_path; # path into the configuration file used

    # open file in read mode ( < )
    open(config_file_handle, "<$file_path") or $error_flag = 1;
    if ( $error_flag == 0 ) {
        # read the entire config file line by line
        while(<config_file_handle>) {
            # enter only if there is the = char on the set and the char # is NOT on the begining of the line
            if ( $_ =~ /=/ and not $_ =~ /^#/ ) {

                my ( $config_key, $config_value ) = split ( /=/, $_,2); # split the line into the configuration key and the configuration value, limit is 2 meaning it will only split the first = char, the oracle exp command uses = (file=named.dmp) and causes problems
                $config_key =~ s/^\s+//;  # Delete leading whitespace ^ is the first caracter  \s is space then // replace with nothing if there was /a/ then replace with a
                $config_key =~ s/\s+$//;  # Delete trailing whitespace same thing kind of but $ represents end of line.
                $config_value =~ s/^\s+//;  # Delete leading whitespace ^ is the first caracter  \s is space then // replace with nothing if there was /a/ then replace with a
                $config_value =~ s/\s+$//;  # Delete trailing whitespace same thing kind of but $ represents end of line

                # use a switch / case this is not a real perl function but it works
                # this will be used to load the global configuration hash
                switch: {
                    /host_os/   && do {
                                        $$global_config_p{host_os} = $config_value;
                                        $$global_config_p{host_os} = "win2000/xp/2003" if ($$global_config_p{host_os} eq "win2000/xp"); # this is done to maintain compatibility with previous simplebackup versions
                                    };
                   /backup_mode/ && do {
                                       $$global_config_p{backup_mode} = $config_value;
                                       $$global_config_p{backup_real_mode} = $config_value;
                                    };
                   /full_backup_day/ && do {
                                       $$global_config_p{full_backup_day} = $config_value;
                                    };
                   /backup_db_management_file/ && do {
                                       $$global_config_p{backup_db_management_file} = $config_value;
                                    };
                   /backup_extension_rejections/ && do {
                                       $$global_config_p{backup_extension_rejections} = $config_value;
                                       $$global_config_p{backup_extension_rejections} = remove_excess_spaces($$global_config_p{backup_extension_rejections}, "," ) if $$global_config_p{backup_extension_rejections} ne "";
                                    };
                   /backup_filenames_rejections/ && do {
                                       $$global_config_p{backup_filenames_rejections} = $config_value;
                                       $$global_config_p{backup_filenames_rejections} = remove_excess_spaces($$global_config_p{backup_filenames_rejections}, "," ) if $$global_config_p{backup_filenames_rejections} ne "";
                                    };
                    /backup_file_size_limit/ && do {
                                       $$global_config_p{backup_file_size_limit} = $config_value;
                                    };
                    /backup_directory_rejections/ && do {
                                       $$global_config_p{backup_directory_rejections} = $config_value;
                                    };
                    /backup_format/   && do {
                                       $$global_config_p{backup_format} = $config_value;
                                     };
                    /backup_type/   && do { # backward compatibility with simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1
                                        if ( $$global_config_p{backup_format} eq "" ) {
                                            $$global_config_p{backup_format} = $config_value;
                                        }
                                     };
                    /backup_compression_ratio/   && do {
                                        $$global_config_p{backup_compression_ratio} = $config_value;
                                     };
                    /input_backup/   && do {
                                        $$global_config_p{input_backup} = $config_value;
                                        $$global_config_p{input_backup} = remove_excess_spaces($$global_config_p{input_backup}, "<" ) if $$global_config_p{input_backup} ne "";
                                    };
                    /input_backup_dirs/   && do { # backward compatibility with simplebackup 1.3.0, 1.3.1, 1.4.0, 1.4.1
                                        if ( $$global_config_p{input_backup} eq "" ) {
                                            $$global_config_p{input_backup} = $config_value;
                                            $$global_config_p{input_backup} = remove_excess_spaces($$global_config_p{input_backup}, "<" ) if $$global_config_p{input_backup} ne "";
                                        }
                                    };
                    /input_backup_dir/   && do { # backward compatibility with simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1
                                        if ( $$global_config_p{input_backup} eq "" ) {
                                            $$global_config_p{input_backup} = $config_value;
                                            $$global_config_p{input_backup} = remove_excess_spaces($$global_config_p{input_backup}, "<" ) if $$global_config_p{input_backup} ne "";
                                        }
                                    };
                    /output_backup/   && do {
                                        $$global_config_p{output_backup} = $config_value;
                                    };
                    /output_backup_dir/   && do { # backward compatibility with simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1
                                        if ( $$global_config_p{output_backup} eq "" ) {
                                            $$global_config_p{output_backup} = $config_value;
                                        }
                                    };
                    /temporary_dir/   && do {
                                        $$global_config_p{temporary_dir} = $config_value;
                                    };
                    /keep_last_n_files/   && do {
                                        $$global_config_p{keep_last_n_files} = $config_value;
                                    };
                    /run_before_backup/   && do {
                                        $$global_config_p{run_before_backup} = $config_value;
                                    };
                    /run_before_arguments/   && do { # backward compatibility simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1, 1.3.0, 1.3.1, 1.4.0, 1.4.1
                                        $config_value =~ s/( )\1/$1/g; # s/ search  /( ) for spaces  \1 first double chars... first double spaces     /$1 replace with first char in this case always a space  /g global to all string
                                        if ( $config_value ne "" ) { # detect empty arguments
                                            $$global_config_p{run_before_backup} = $$global_config_p{run_before_backup} . " " . $config_value;
                                        }
                                    };
                    /run_after_backup/   && do {
                                        $$global_config_p{run_after_backup} = $config_value;
                                    };
                    /run_after_arguments/   && do { # backward compatibility simplebackup 1.0.0, 1.1.0, 1.2.0, 1.2.1, 1.3.0, 1.3.1, 1.4.0, 1.4.1
                                        $config_value =~ s/( )\1/$1/g; # s/ search  /( ) for spaces  \1 first double chars... first double spaces     /$1 replace with first char in this case always a space  /g global to all string
                                        if ( $config_value ne "" ) { # detect empty arguments
                                            $$global_config_p{run_after_backup} = $$global_config_p{run_after_backup} . " " . $config_value;
                                        }
                                    };
                    /log_file/   && do {
                                        $$global_config_p{log_file} = $config_value;
                                    };
                    /ftp_debug_level/   && do {
                                        $$global_config_p{ftp_debug_level} = $config_value;
                                    };
                    /ftp_passive_mode/   && do {
                                        $$global_config_p{ftp_passive_mode} = $config_value;
                                    };
                    /ftp_timeout/   && do {
                                        $$global_config_p{ftp_timeout} = $config_value;
                                    };
                    /ftp_proxy/   && do {
                                        $$global_config_p{ftp_proxy} = $config_value;
                                    };
                    /ftp_connect_retries/   && do {
                                        $$global_config_p{ftp_connect_retries} = $config_value;
                                    };
                    /tape_before_backup/   && do {
                                        $$global_config_p{tape_before_backup} = $config_value;
                                    };
                    /tape_after_backup/   && do {
                                        $$global_config_p{tape_after_backup} = $config_value;
                                    };
                    /mail_server/   && do {
                                        $$global_config_p{mail_server} = $config_value;
                                    };
                    /mail_debug_level/   && do {
                                        $$global_config_p{mail_debug_level} = $config_value;
                                    };
                    /mail_port/   && do {
                                        $$global_config_p{mail_port} = $config_value;
                                    };
                    /mail_timeout/   && do {
                                        $$global_config_p{mail_timeout} = $config_value;
                                    };
                    /mail_from_address/   && do {
                                        $$global_config_p{mail_from_address} = $config_value;
                                    };
                    /mail_to_addresses/   && do {
                                        $$global_config_p{mail_to_addresses} = $config_value;
                                        $$global_config_p{mail_to_addresses} = remove_excess_spaces($$global_config_p{mail_to_addresses}, ",") if $$global_config_p{mail_to_addresses} ne "";
                                    };
                    /mail_username/   && do {
                                        $$global_config_p{mail_username} = $config_value;
                                    };
                    /mail_password/   && do {
                                        $$global_config_p{mail_password} = $config_value;
                                    };
                    /pass_log_to_external_cmd/   && do {
                                        $$global_config_p{pass_log_to_external_cmd} = $config_value;
                                    };
                    /hostname_on_backup_files/   && do {
                                        $$global_config_p{hostname_on_backup_files} = $config_value;
                                    };
                    /local_db_history_file/   && do {
                                        $$global_config_p{local_db_history_file} = $config_value;
                                    };
                }
            }
        }
        # close file
        close (config_file_handle);
    }

    # configure the apropriate backup mode
    switch: {
            ( substr($$global_config_p{output_backup},0,6) ne "ftp://" and  substr($$global_config_p{output_backup},0,6) ne "scp://" and  substr($$global_config_p{output_backup},0,7) ne "tape://" ) && do {
                $$global_config_p{backup_working} = "local file system"; # local file system backup mode (ex. /my_directory or d:\my_directory)
            };
            substr($$global_config_p{output_backup},0,6) eq "ftp://" && do {
                $$global_config_p{backup_working} = "ftp"; # ftp backup mode
            };
            substr($$global_config_p{output_backup},0,6) eq "scp://" && do {
                $$global_config_p{backup_working} = "scp"; # scp backup mode
                $error_flag = 17; # this mode is not yet supported
            };
            substr($$global_config_p{output_backup},0,7) eq "tape://" && do {
                $$global_config_p{backup_working} = "tape"; # tape backup mode
            };
    }

    # test operating system host values, valid is auto ; winnt4 ; win2000/xp/2003 ; unix
    $$global_config_p{host_os} = get_os_type() if ( $$global_config_p{host_os} eq "" or $$global_config_p{host_os} eq "auto" ); # autodetect the operating system if required
    if ( $error_flag == 0 and ($$global_config_p{host_os} ne "winnt4" and $$global_config_p{host_os} ne "win2000/xp/2003" and $$global_config_p{host_os} ne "unix") ) {
        $error_flag = 3; # invalid operating system
    }
    elsif ( $error_flag == 0 ) { # configure the null device
        if ( $$global_config_p{host_os} eq "unix" ) { # unix null device
            $$global_config_p{host_null_device} = "2>/dev/null";
            $$global_config_p{host_file_separator} = "/";
        }
        else { # microsoft null device and microsoft file separator
            $$global_config_p{host_null_device} = "2>nul";
            $$global_config_p{host_file_separator} = "\\";
        }
        # get just the configuration file name
        $$global_config_p{configuration_file_name} = get_filedir_name("$$global_config_p{configuration_file}", $global_config_p);
    }

    $$global_config_p{hostname} = hostname(); # configure the hostname
    $$global_config_p{hostname} = "unavailable" if ( !(defined $$global_config_p{hostname}) or $$global_config_p{hostname} eq ""); # if we cannot get the hostname of this machine replaced it by "unavailable"
    if ( $error_flag == 0 ) {
        # Check the hostname on backup files
        if ( $$global_config_p{hostname_on_backup_files} eq "" ) {
            $$global_config_p{hostname_on_backup_files}  = "no"; # change to default value
        }
        if ( $$global_config_p{hostname_on_backup_files} ne "no" and $$global_config_p{hostname_on_backup_files} ne "yes" ) {
            $error_flag = "43"; # mark error invalid value
        }
    }

    # read the backup file list (if required)
    if ( substr($$global_config_p{input_backup},0,8) eq "file://@" ) {
        my ( $list_config_key, $list_file_path ) = split ( /@/, $$global_config_p{input_backup}); # get the file path
        $error_file_directory = "$list_file_path";
        if ( check_if_full_path($list_file_path, $$global_config_p{host_os}) != 0 or ! -f $list_file_path ) {
            $error_flag = 12; # not a full path or file does not exists
        }
        else { # file is a full path, open and load the backup list
            open(list_file_handle, "<$list_file_path") or $error_flag = 13; # open file in read mode
            if ( $error_flag == 0 ) {
                $$global_config_p{input_backup} = ""; # clear the file path
                while(<list_file_handle>) { # read the entire file into the array
                    chomp($_); # remove the new line chars
                    if ( $_ !~ /^#/ ) { # read only if there is no # char on the start of the line
                        $_ =~ s/^\s+//;  # Delete leading whitespace ^ is the first caracter  \s is space then // replace with nothing if there was /a/ then replace with a
                        $_ =~ s/\s+$//;  # Delete trailing whitespace same thing kind of but $ represents end of line
                        if ( $$global_config_p{input_backup} ne "" ) { # if not first line of file
                            $$global_config_p{input_backup} = $$global_config_p{input_backup} . "<" . $_;
                        }
                        else { # if first line
                            $$global_config_p{input_backup} = $_;
                        }
                    }
                }
                close(list_file_handle);
            }
        }
    }

    # test if the file was correctly loaded   used eq and not == because we are using strings
    if ( $error_flag == 0 and ( $$global_config_p{backup_format} eq "" or $$global_config_p{input_backup} eq "" or $$global_config_p{output_backup} eq "" or $$global_config_p{keep_last_n_files} eq "" ) ) {
        $error_flag = 2; # mark corrupted file
    }

    # search for invalid chars, in this case the * char
    if ( $error_flag == 0 and ( $$global_config_p{backup_db_management_file} =~ /\*/ or $$global_config_p{backup_extension_rejections} =~ /\*/ or $$global_config_p{backup_filenames_rejections} =~ /\*/ or $$global_config_p{input_backup} =~ /\*/ or $$global_config_p{temporary_dir} =~ /\*/ or $$global_config_p{run_before_backup} =~ /\*/ or $$global_config_p{run_after_backup} =~ /\*/ or $$global_config_p{log_file} =~ /\*/ or $$global_config_p{local_db_history_file}  =~ /\*/ ) ) { # protection against using the * char
        $error_flag = 2; # mark corrupted file
    }

    # test backup type is valid
    if ( $error_flag == 0 and ($$global_config_p{backup_format} ne "rar" and $$global_config_p{backup_format} ne "zip" and $$global_config_p{backup_format} ne "tar.Z" and $$global_config_p{backup_format} ne "tar.bz2" and $$global_config_p{backup_format} ne "tar.gz" and $$global_config_p{backup_format} ne "tar") ) {
        $error_flag = 4; # invalid backup type
    }
    $$global_config_p{backup_real_ratio} = 0; # default no compression
    # test if backup ratio is valid
    if ( $error_flag == 0 and ($$global_config_p{backup_compression_ratio} ne "none" and $$global_config_p{backup_compression_ratio} ne "minimal" and $$global_config_p{backup_compression_ratio} ne "average" and $$global_config_p{backup_compression_ratio} ne "maximal" )) {
        $error_flag = 5; # invalid backup ratio
    }
    elsif ( $error_flag == 0 ) { # valid backup ratio set the real value that is going to get passed into the compressor
        switch: {
            $$global_config_p{backup_compression_ratio} eq "none" && do { # no compression
                            if ( $$global_config_p{backup_format} ne "rar" and $$global_config_p{backup_format} ne "zip" ) {
                                $$global_config_p{backup_real_ratio} = 1; # almost no compression, none is not avaliable for bzip2, and gzip
                            }
                        };
            $$global_config_p{backup_compression_ratio} eq "minimal" && do { # minimal
                            $$global_config_p{backup_real_ratio} = 1; # equal to all compressors
                        };
            $$global_config_p{backup_compression_ratio} eq "average" && do { # average compression
                            if ( $$global_config_p{backup_format} eq "rar" ) {
                                $$global_config_p{backup_real_ratio} = 3; # average, just for rar
                            }
                            else {
                                $$global_config_p{backup_real_ratio} = 6; # all others use this
                            }
                        };
            $$global_config_p{backup_compression_ratio} eq "maximal" && do { # average compression
                            if ( $$global_config_p{backup_format} eq "rar" ) {
                                $$global_config_p{backup_real_ratio} = 5; # max just for rar
                            }
                            else {
                                $$global_config_p{backup_real_ratio} = 9; # all others use this
                            }
                        };
        }
    }

    # check if we are using valid path's ... / for unix host system \ for microsoft systems
    if ( $error_flag == 0 and ( $$global_config_p{host_os} eq "unix" and ($$global_config_p{input_backup} =~ /\\/ or $$global_config_p{output_backup} =~ /\\/ or $$global_config_p{temporary_dir} =~ /\\/ or $$global_config_p{log_file} =~ /\\/ or $$global_config_p{local_db_history_file} =~ /\\/ or $$global_config_p{backup_db_management_file} =~ /\\/) )) {
        $error_flag = 6; # Using microsoft type of paths \ on a unix system
    }
    # check if we are using valid path's ... / for unix host system \ for microsoft systems
    if ( $error_flag == 0 and ( $$global_config_p{host_os} ne "unix" and ( $$global_config_p{input_backup} =~ /[\/]/ or ($$global_config_p{output_backup} =~ /[\/]/ and $$global_config_p{backup_working} eq "local file system") or $$global_config_p{temporary_dir} =~ /[\/]/ or $$global_config_p{log_file} =~ /[\/]/ or $$global_config_p{local_db_history_file} =~ /[\/]/ or $$global_config_p{backup_db_management_file} =~ /[\/]/) )) {
        $error_flag = 7; # Unix unix type of paths / on a microsoft system
    }
    # test if the directory(ies) and files to backup exist and if the their path is a full path
    if ( $error_flag == 0 ) {
        my @input_backup_split = split(/</,  $$global_config_p{input_backup}); # real all backup directories
        foreach my $test_file_directory (@input_backup_split) { # test each backup directory for full path and if it exists.
            if ( $test_file_directory eq "/" ) {
                $error_flag=10; # mark error you cannot backup your root directory
            }
            elsif ( check_if_full_path($test_file_directory, $$global_config_p{host_os}) != 0 or (!-d $test_file_directory and !-f $test_file_directory ) ) {
                $error_flag = 40; # mark file or directory does not exit or it's not a full path
            }
            # search for repeted directory paths
            my $file_directory_count =0; # search for
            foreach my $test_file_directory_namesearch (@input_backup_split) { # test each backup directory for full path and if it exists.
                $file_directory_count++ if ( $test_file_directory eq $test_file_directory_namesearch );
            }
            $error_flag = 42 if $file_directory_count > 1; # mark error repeted file or directory path
            $error_file_directory = $test_file_directory if $error_flag != 0; # name of the directory to cause the error will be show on screen
            last if $error_flag != 0; # exit at first error
        }
    }

    # test if the backup destiny directory exists
    if ( $error_flag == 0 and $$global_config_p{backup_working} eq "local file system" and (!-d $$global_config_p{output_backup} or check_if_full_path($$global_config_p{output_backup}, $$global_config_p{host_os}) != 0) ) {
        $error_file_directory = "Backup destination directory";
        $error_flag = 8; # mark directory does not exit
    }

    # test if the backup temporary directory exists
    $$global_config_p{temporary_dir} = get_tmp_dir($$global_config_p{host_os}) if ( $$global_config_p{temporary_dir} eq "" or $$global_config_p{temporary_dir} eq "auto");
    if ( $error_flag == 0 and (!-d $$global_config_p{temporary_dir} or check_if_full_path($$global_config_p{temporary_dir}, $$global_config_p{host_os}) != 0) ) {
        $error_file_directory = "Temporary directory";
        $error_flag = 8; # mark directory does not exit
    }
    elsif ( $error_flag == 0 ) { # temporary directory exists, configure (if required by using the auto or empty parameter) the auxiliar files
        # if the log file is configured to be automaticly named
        if ( $$global_config_p{log_file} eq "" or $$global_config_p{log_file} eq "auto" ) {
            $$global_config_p{log_file} = rebuild_full_path("$$global_config_p{temporary_dir}", "$$global_config_p{configuration_file_name}.log", $global_config_p) if $$global_config_p{hostname_on_backup_files} eq "no"; # rebuild the full path pointing also into the path defined by temporary_dir and if we are not incluing the hostname on the backup files
            $$global_config_p{log_file} = rebuild_full_path("$$global_config_p{temporary_dir}", "$$global_config_p{hostname}.$$global_config_p{configuration_file_name}.log", $global_config_p) if $$global_config_p{hostname_on_backup_files} eq "yes"; # rebuild the full path pointing also into the path defined by temporary_dir and if we are incluing the hostname on the backup files
        }
        # if the management file is configured to be automaticly named and the backup mode is not full
        if ( ($$global_config_p{backup_db_management_file} eq "" or $$global_config_p{backup_db_management_file} eq "auto") and $$global_config_p{backup_mode} ne "full" ) {
            $$global_config_p{backup_db_management_file} = rebuild_full_path("$$global_config_p{temporary_dir}", "$$global_config_p{configuration_file_name}.management.db", $global_config_p) if $$global_config_p{hostname_on_backup_files} eq "no"; # rebuild the full path pointing also into the path defined by temporary_dir and if we are not incluing the hostname on the backup files
            $$global_config_p{backup_db_management_file} = rebuild_full_path("$$global_config_p{temporary_dir}", "$$global_config_p{hostname}.$$global_config_p{configuration_file_name}.management.db", $global_config_p) if $$global_config_p{hostname_on_backup_files} eq "yes"; # rebuild the full path pointing also into the path defined by temporary_dir and if we are incluing the hostname on the backup files
        }
        # if the history file is configured to be automaticly named or the path is not defined and we are doing a backup into a tape device
        if ( $$global_config_p{local_db_history_file} eq "auto" ) {
            $$global_config_p{local_db_history_file} = rebuild_full_path("$$global_config_p{temporary_dir}", "$$global_config_p{configuration_file_name}.history.db", $global_config_p) if $$global_config_p{hostname_on_backup_files} eq "no"; # rebuild the full path pointing also into the path defined by temporary_dir and if we are not incluing the hostname on the backup files
            $$global_config_p{local_db_history_file} = rebuild_full_path("$$global_config_p{temporary_dir}", "$$global_config_p{hostname}.$$global_config_p{configuration_file_name}.history.db", $global_config_p) if $$global_config_p{hostname_on_backup_files} eq "yes"; # rebuild the full path pointing also into the path defined by temporary_dir and if we are incluing the hostname on the backup files
        }
    }

    # check if the value is valid, it must be >= -1  /^[+-]?\d+$/   /^\d+$/
    if (  $error_flag == 0 and ( $$global_config_p{keep_last_n_files} !~ /^[+-]?\d+$/ or $$global_config_p{keep_last_n_files} < -1 ) ) {
        $error_flag = 9; # Invalid value
    }

    # test if the backup destiny directory conflits with the temporary or input directory
    if ( $error_flag == 0 and ( $$global_config_p{input_backup} eq ($$global_config_p{temporary_dir} or $$global_config_p{output_backup}) )) {
        $error_file_directory = "Backup directory";
        $error_flag = 11; # mark directory does not exit
    }

    # Ftp setup
    if ( $error_flag == 0 and $$global_config_p{backup_working} eq "ftp" ) { # if we are using ftp backup mode
        eval {
            require Net::FTP; # attempt to detect if the  smtp authentication module is installed
        };
        if ($@) {
            $error_flag = 38; # required module is not available
        }
        if ( $error_flag == 0 ) {
            # configure the debug level
            $$global_config_p{ftp_debug_level} = "none" if $$global_config_p{ftp_debug_level} eq ""; # default value
            if ( $$global_config_p{ftp_debug_level} ne "none" and $$global_config_p{ftp_debug_level} ne "full" ) {
                $error_flag = 37; # mark error invalid ftp debug level
            }
        }
        # Check if ftp_passive mode is correcly defined, must be yes/no or empty ""
        if ( $error_flag == 0 and ($$global_config_p{ftp_passive_mode} ne "" and $$global_config_p{ftp_passive_mode} ne "yes" and $$global_config_p{ftp_passive_mode} ne "no") ) {
            $error_flag = 14; # mark error invalid passive mode
        }
        else { # configure the passive mode parameter (to pass into NET::FTP)
                if ( $$global_config_p{ftp_passive_mode} eq "" ) {
                    $$global_config_p{ftp_passive_mode}  = "no"; # change to default value
                }
                if ( $$global_config_p{ftp_passive_mode} eq "no" ) { # passive mode no
                    $$global_config_p{ftp_passive_mode_real} = 0; # The real parameter passed into the NET:FTP, here is [ 0 ; 1 ]
                }
                else { # passive mode yes
                    $$global_config_p{ftp_passive_mode_real} = 1; # The real parameter passed into the NET:FTP, here is [ 0 ; 1 ]
                }
        }
        # check if the ftp url is correct, and place all other ftp values in the correct places
        if ( $error_flag == 0 ) {
            my @ftp_array = split(/:/, $$global_config_p{output_backup}); # split the url into fields
            my $ftp_array_size = scalar(@ftp_array); # get the array size
            my $ftp_username = "";
            my $ftp_password = "";
            my $ftp_server = "";
            my $ftp_port = "";
            my @ftp_port_servername_ftp_dir = "";
            # general url check: if there is only one @  ; at least 3  / ( ftp:// and the / from the remote ftp dir) and there must be 2 or 3 :
            if ( (($$global_config_p{output_backup} =~ tr/@//) != 1) or ($$global_config_p{output_backup} =~ tr/\///) < 3 or !(($$global_config_p{output_backup} =~ tr/://) >= 2 and ($$global_config_p{output_backup} =~ tr/://) <= 3) ) {
                $error_flag = 15; # mark error
            }
            else {
                if ( $ftp_array_size ==  3 or $ftp_array_size == 4 ) { # get username, password and ftp server
                    $ftp_username = $ftp_array[1];
                    $ftp_username =~ s/\///g; # remove the // in order to only get the ftp username
                }
                if ( scalar(@ftp_array) == 4 ) { # ftp url with ftp port declared
                    ( $ftp_password, $ftp_server ) = split ( /@/, $ftp_array[2]); # get the ftp password and server name
                    @ftp_port_servername_ftp_dir = split ( /\//, $ftp_array[3]); # split the port and directory paths
                    if ( !($ftp_port_servername_ftp_dir[0] =~ /^\d+$/) ) { # check if port number is a integer number without + or -
                        $error_flag = 15; # mark error
                    }
                    else {
                        $ftp_port = $ftp_port_servername_ftp_dir[0];
                    }
                }
                elsif ( scalar(@ftp_array) == 3 ) { # ftp url without ftp port declared
                    my $ftp_server_ftp_directory = ""; # to countaing ftp_server + directories
                    my $trash = ""; # simple dump
                    ( $ftp_password, $ftp_server_ftp_directory ) = split ( /@/, $ftp_array[2]); # get the ftp password server name + directories
                    ( $ftp_server, $trash ) = split ( /\//, $ftp_server_ftp_directory); # split the servername from the directories
                    $ftp_port = 21; # standard ftp port
                    @ftp_port_servername_ftp_dir = split ( /\//, $ftp_array[2]); # split the directory paths
                }
                else { # URL is incorrect
                    $error_flag = 15; # mark error
                }
            }
            if ( $error_flag == 0 ) {
                # rebuild the correct ftp directory path
                my $ftp_directory = "";
                my $i;
                for ($i = 1; $i < scalar(@ftp_port_servername_ftp_dir); $i++) {
                    $ftp_directory = "$ftp_directory/$ftp_port_servername_ftp_dir[$i]";
                }
                if ( $ftp_directory eq "" ) { # fix in case the user has typed / on the backup directory
                    $ftp_directory = "/";
                }
                # final check
                if ( $ftp_server ne "" and $ftp_port ne "" and $ftp_username ne "" and $ftp_password ne "" and $ftp_directory ne "" ) {
                    # update the main configuration hash
                    $$global_config_p{ftp_server} =  $ftp_server;        # Remote ftp server.
                    $$global_config_p{ftp_server_port} = $ftp_port;      # Remote ftp server port number.
                    $$global_config_p{ftp_username} = $ftp_username;     # FTP Username.
                    $$global_config_p{ftp_password} = $ftp_password;     # FTP Password
                    $$global_config_p{ftp_directory} = $ftp_directory;   # FTP directory where we will upload the backup file.
                }
                else {
                    $error_flag = 15; # mark error
                }
            }
        }
        if ( $error_flag == 0 ) { # if there was no error get the ftp timeout value
            if ( $$global_config_p{ftp_timeout} eq "" ) { # If it's required to put the default Ftp timeout value
                $$global_config_p{ftp_timeout} = 480; # default value
            }
            elsif ( !($$global_config_p{ftp_timeout} =~ /^\d+$/) or $$global_config_p{ftp_timeout} <= 0 ) { # check if the number is a integer
                $error_flag = 16; # mark error
            }
        }
        if ( $error_flag == 0 ) { # if there was no error get the ftp connect retry value
            if ( $$global_config_p{ftp_connect_retries} eq "" ) {
                $$global_config_p{ftp_connect_retries} = 6; # default value
            }
            elsif ( !($$global_config_p{ftp_connect_retries} =~ /^\d+$/) or $$global_config_p{ftp_connect_retries} <= 0 ) { # check if the number is a integer
                $error_flag = 41; # mark error
            }
        }
    }

    # Check if we are using Windows and Tape backups
    if ( $error_flag == 0 and $$global_config_p{backup_working} eq "tape" and $$global_config_p{host_os} ne "unix" ) {
        $error_flag = 18; # mark error, tape backups are not compatible with windows
    }

    # Check if are not using windows and if we have the before writing to tape and after writing to tape configured
    if ( $error_flag == 0 and $$global_config_p{backup_working} ne "tape" and ( ( $$global_config_p{tape_before_backup} ne "" and $$global_config_p{tape_before_backup} ne "none" ) or ( $$global_config_p{tape_after_backup} ne "" and $$global_config_p{tape_after_backup} ne "none" ) ) ) {
        $error_flag = 34; # mark error using tape procedures and not doing a tape backup
    }

    if ( $error_flag == 0 and $$global_config_p{backup_working} eq "tape" ) {
        if ( ($$global_config_p{output_backup} =~ tr/@//) != 1 ) {
            $error_flag = 19; # mark error, invalid tape url, must be something similar to tape://@/dev/my_tape_device
        }
        else {
            my $trash; # ignore this
            ($trash, $$global_config_p{tape_device} ) = split (/@/, $$global_config_p{output_backup});
            if ( !(-l $$global_config_p{tape_device} or -b $$global_config_p{tape_device} or -c $$global_config_p{tape_device}) ) { # tape device must be a link a block device or a caracter device
                $$global_config_p{tape_device} = ""; # clear the value since it's wrong
                $error_flag = 19; # mark error, invalid tape device
            }
            if ( $$global_config_p{local_db_history_file} eq "" ) {
                $error_flag = 35; # mark error, not using a local history file
            }
            if ( $error_flag == 0 ) { # all is well, configure the remaining tape stuff
                if ( $$global_config_p{tape_command} eq "" ) { # configure the default tape backup command
                    $$global_config_p{tape_command} = "tar"; # default tape backup command
                }
                # configure the default command to execute before writing and after writing to tape
                if ( $$global_config_p{tape_before_backup} eq "" ) {
                    $$global_config_p{tape_before_backup} = "mt -f " . $$global_config_p{tape_device} . " rewind ";
                }
                if ( $$global_config_p{tape_after_backup} eq "" ) {
                    $$global_config_p{tape_after_backup} = "mt -f " . $$global_config_p{tape_device} . " offline ";
                }
                # if user doesn't want any command to be runned
                $$global_config_p{tape_before_backup} = "" if $$global_config_p{tape_before_backup} eq "none";
                $$global_config_p{tape_after_backup} = "" if $$global_config_p{tape_after_backup} eq "none";
            }
        }
    }

    # Mail stuff
    if ( $$global_config_p{mail_type} eq "" ) {
        $$global_config_p{mail_type} = "smtp"; # put in the default mail server type
    }
    if ( $error_flag == 0 and $$global_config_p{mail_type} ne "smtp" ) {
        $error_flag = 20; # mark error, invalid smtp mail server type
    }
    elsif ( $error_flag == 0 and $$global_config_p{mail_server} ne "" ) { # if there was no previous error and a mail server is defined
        if ($$global_config_p{mail_type} eq "smtp" ) { # if the mail server is SMTP type
            eval {
                require Net::SMTP; # attempt to detect if the  smtp module is installed
            };
            if ($@) {
                $error_flag = 39; # required module is not available
            }
        }
        if ( $$global_config_p{mail_username} ne "" ) { # If mail requires authentication
            eval {
                require Authen::SASL; # attempt to detect if the mail authentication module is installed
            };
            if ($@) {
                $error_flag = 44; # required module is not available
            }
        }

        # check mail server
        if ( $error_flag == 0 and $$global_config_p{mail_server} =~ /\s/ ) { # if the mail server name contains a invalid char... \n \r, tabs or spaces
            $error_flag = 21; # mark error, invalid smtp mail server
        }
        # check the mail debug level
        if ( $$global_config_p{mail_debug_level} eq "" ) {
            $$global_config_p{mail_debug_level} = "none";
        }
        if ( $error_flag == 0 and ( $$global_config_p{mail_debug_level} ne "none" and $$global_config_p{mail_debug_level} ne "full" ) ) {
            $error_flag = 36; # mark error, invalid mail debug level
        }
        # check username + password, if username is used and password is not used or password is used and username not used
        if ( $error_flag == 0 and (($$global_config_p{mail_username} ne "" and $$global_config_p{mail_password} eq "") or ($$global_config_p{mail_username} eq "" and $$global_config_p{mail_password} ne "")) ) {
            $error_flag = 22; # mark error, lack of smtp password or lack of smtp username
        }
        # check the from address, must have one @ char and no , char
        if ( $error_flag == 0 and (($$global_config_p{mail_from_address} =~ tr/@//) != 1 or ($$global_config_p{mail_from_address} =~ tr/,//) != 0) ) {
            $error_flag = 23; # mark error, invalid smtp from address
        }
        # check the to addresses, must have at least has many @ has the , char -1
        # valid is: count(@) - count (,) = 1;
        if ( $error_flag == 0 and (($$global_config_p{mail_to_addresses} =~ tr/@//) - ($$global_config_p{mail_to_addresses} =~ tr/,//)) != 1 ) {
            $error_flag = 24; # mark error, invalid smtp to addresses
        }
        # check the mail port value
        if ( $$global_config_p{mail_port} eq "" ) { # not declared by user on config file
            $$global_config_p{mail_port} = 25; # put in the default value
        }
        elsif ( !($$global_config_p{mail_port} =~ /^\d+$/) or $$global_config_p{mail_port} <= 0 ) { # check if the number is a integer
            $error_flag = 26; # mark error user declared a invalid mail server port
        }
        # check the mail timeout value
        if ( $$global_config_p{mail_timeout} eq "" ) { # not declared by user on config file
            $$global_config_p{mail_timeout} = 480; # put in the default value
        }
        elsif ( !($$global_config_p{mail_timeout} =~ /^\d+$/) or $$global_config_p{mail_timeout} <= 0 ) { # check if the number is a integer
            $error_flag = 27; # mark error user declared a invalid mail server port
        }
    }
    elsif ( $error_flag == 0 ) { # no mail server defined
        if ( $$global_config_p{mail_from_address} ne "" or $$global_config_p{mail_to_addresses} ne "" or $$global_config_p{mail_username} ne "" or $$global_config_p{mail_password} ne "" or $$global_config_p{mail_port} ne "" or $$global_config_p{mail_timeout} ne "" ) {
            $error_flag = 25; # mark error, we have some mail configuration but no mail server is declared
        }
    }
    # End of Mail stuff

    # check backup mode
    $$global_config_p{backup_mode} = "full" if $$global_config_p{backup_mode} eq ""; # backward compatibity
    if ( $error_flag == 0 and ($$global_config_p{backup_mode} ne "full" and $$global_config_p{backup_mode} ne "incremental" and $$global_config_p{backup_mode} ne "differential") ) {
        $error_flag = 28; # mark error, invalid backup mode
    }

    # check the full backup day
    if ( $error_flag == 0 and $$global_config_p{backup_mode} ne "full" ) { # if we are using full backups this parameter is ignored
        $$global_config_p{keep_last_n_files} = -1; # by default don't delete any files
        my $today_result = is_today($$global_config_p{full_backup_day});
        if ( $today_result eq "yes" ) { # if it's the full backup
            $$global_config_p{keep_last_n_files} = 0; # doing a full backup, keep only the present backup and delete all the rest
            $$global_config_p{backup_real_mode} = "full"; # reset the backup type to FULL
        }
        elsif ( $today_result eq "error" ) {
            $error_flag = 29; # mark error, invalid full backup day
        }
    }
    else { # if user is in full backup mode in the configuration file then the real mode is always full
        $$global_config_p{backup_real_mode} = "full"; # reset the backup type to FULL
    }

    # check if the path into the management file is declared
    if ( $error_flag == 0 and ($$global_config_p{backup_mode} eq "incremental" or $$global_config_p{backup_mode} eq "differential") and $$global_config_p{backup_db_management_file} eq "") {
        $error_flag = 30; # mark error, missing path into management file
    }

    # check if the path into the management file is not equal to the path defined by the  Local Database file history
    if ( $error_flag == 0 and $$global_config_p{local_db_history_file} ne "" and $$global_config_p{backup_db_management_file} eq $$global_config_p{local_db_history_file} ) {
        $error_flag = 33; # mark error, management file is the same has the local database file
    }

    # Check the file size
    $$global_config_p{backup_file_size_limit} = 0 if $$global_config_p{backup_file_size_limit} eq ""; # if not reported on the configuration file, change into default value  this is also importante for backward compatibility
    if ( $error_flag == 0 and ( !($$global_config_p{backup_file_size_limit} =~ /^\d+$/) and $$global_config_p{backup_file_size_limit} ne 0 ) ) {
        $error_flag = 31; # mark error, invalid size limit
    }

    # Check directory rejection
    $$global_config_p{backup_directory_rejections} = "no" if $$global_config_p{backup_directory_rejections} eq ""; # if not reported on the configuration file, change into default value  this is also importante for backward compatibility
    if ( $error_flag == 0 and ($$global_config_p{backup_directory_rejections} ne "yes" and $$global_config_p{backup_directory_rejections} ne "no") ) {
        $error_flag = 32; # mark error, backup directory configuration
    }

    if ( $error_flag == 0 ) { # if there was not errors configure the operating system null device, which is used all over the program and some other things and the file separator / or \

        # Delete trailing \ or / char same thing kind of but $ represents end of line + represents any \ or / chars
        $$global_config_p{temporary_dir} =~ s/[\\\/]+$// if ( $$global_config_p{temporary_dir} ne "/" and ( (length($$global_config_p{temporary_dir}) != 3 and $$global_config_p{host_os} ne "unix") or $$global_config_p{host_os} eq "unix" ) ); # if not the / or a local drive (in windows) example D:\
        $$global_config_p{output_backup} =~ s/[\\\/]+$// if ( $$global_config_p{backup_working} eq "local file system" and $$global_config_p{output_backup} ne "/" and ( (length($$global_config_p{output_backup}) != 3 and $$global_config_p{host_os} ne "unix") or $$global_config_p{host_os} eq "unix") );
        $$global_config_p{input_backup} =~ s/[\\\/]+$// if ( $$global_config_p{input_backup} ne "/" and ( (length($$global_config_p{input_backup}) != 3 and $$global_config_p{host_os} ne "unix") or $$global_config_p{host_os} eq "unix") );

        # add the null device into the tape commands
        $$global_config_p{tape_before_backup} = $$global_config_p{tape_before_backup} . $$global_config_p{host_null_device} if $$global_config_p{tape_before_backup} ne "";
        $$global_config_p{tape_after_backup} = $$global_config_p{tape_after_backup} . $$global_config_p{host_null_device} if $$global_config_p{tape_after_backup} ne "";

        # converter the file size limit in kbytes into bytes
        $$global_config_p{backup_file_size_limit_bytes} = $$global_config_p{backup_file_size_limit} * 1024; # 1024 bytes = 1 Kbyte
        # if we are working in full backup mode some other things are irrelevent
        if ( $$global_config_p{backup_mode} eq "full" ) {
            $$global_config_p{full_backup_day} = "everyday";
            $$global_config_p{backup_db_management_file} = "";
        }
        else { # if in incremental or differential backup mode
            if ( !-f $$global_config_p{backup_db_management_file} ) { # if the management file does not exists force full backup
                $$global_config_p{keep_last_n_files} = 0; # doing a full backup, keep only the present backup and delete all the rest
                $$global_config_p{backup_real_mode} = "full"; # reset the backup type to FULL
            }
        }
        # Check if we need to build a backup list
        if ( $$global_config_p{backup_extension_rejections} ne "" or $$global_config_p{backup_filenames_rejections} ne "" or $$global_config_p{backup_file_size_limit} != 0 or $$global_config_p{backup_directory_rejections} ne "no" or $$global_config_p{backup_mode} ne "full" ) {
            $$global_config_p{build_backup_list} = "yes";
        }
        else {
            $$global_config_p{build_backup_list} = "no";
        }
        # check the real email debug level
        if ( $$global_config_p{mail_debug_level} eq "none" ) {
            $$global_config_p{mail_debug_level_real} = 0; # no debug
        }
        else {
            $$global_config_p{mail_debug_level_real} = 1; # full debug
        }
        # check the real ftp debug level
        if ( $$global_config_p{ftp_debug_level} eq "none" ) {
            $$global_config_p{ftp_debug_level_real} = 0; # no debug
        }
        else {
            $$global_config_p{ftp_debug_level_real} = 1; # full debug
        }
    }

    # print apropriate messages
    switch: {
        $error_flag > 0 && do {
                            print "\nUsing configuration file: [ $file_path ]\n\n";
                        };
        $error_flag == 1 && do {
                            print "Error - opening the configuration file,\ncheck if file exists and if you have read permissions.\n\n";
                        };
        $error_flag == 2 && do {
                            print "Error - configuration file corrupted,\ncheck configuration file.\n\n";
                        };
        $error_flag == 3 && do {
                            print "Error - invalid operating system,\nvalid values are: auto ; winnt4 ; win2000/xp/2003 or unix,\ncheck [ host_os value ].\n\n";
                        };
        $error_flag == 4 && do {
                            print "Error - invalid backup type,\nvalid values are: rar ; zip ; tar ; tar.Z ; tar.gz ; tar.bz2,\ncheck [ backup_format value ].\n\n";
                        };
        $error_flag == 5 && do {
                            print "Error - invalid backup compression ratio,\nvalid values are: none ; minimal, average ; maximal,\ncheck [ backup_compression_ratio ].\n\n";
                        };
        $error_flag == 6 && do {
                            print "Error - using microsoft type of paths (\\) on a unix system,\ncheck [ host_os ; input_backup ; output_backup ;\ntemporary_dir values ; log_file ; local_db_history_file (if defined) ;\nbackup_db_management_file (if defined) ].\n\n";
                        };
        $error_flag == 7 && do {
                            print "Error - using unix type of paths (/) on a microsoft system,\ncheck [ host_os ; input_backup ; output_backup ;\ntemporary_dir values ; log_file ; local_db_history_file (if defined) ;\nbackup_db_management_file (if defined) ].\n\n";
                        };
        $error_flag == 8 && do {
                            print "Error - [ $error_file_directory ],\ndoes not exist, it's not defined has a full path\nor you do not have permissions to acess it,\ncheck [ output_backup ; temporary_dir values ].\n\n";
                        };
        $error_flag == 9 && do {
                            print "Error - value is invalid, it must be >= -1,\ncheck [ keep_last_n_files value ].\n\n";
                        };
        $error_flag == 10 && do {
                            print "Error - you cannot use simplebackup to do a root (/) backup of a unix system,\ncheck [ input_backup value ].\n\n";
                        };
        $error_flag == 11 && do {
                            print "Error - [ $error_file_directory ],\ncannot be the same has the\nbackup destination directory or the temporary directory,\ncheck [ input_backup value ].\n\n";
                        };
        $error_flag == 12 && do {
                            print "Error - Backup file list [ $error_file_directory ],\ndoes not exist, it's not defined has a full path\nor you do not have permissions to acess it,\ncheck [ input_backup ].\n\n";
                        };
        $error_flag == 13 && do {
                            print "Error - Unable to read the Backup file list [ $error_file_directory ],\ncheck [ input_backup ].\n\n";
                        };
        $error_flag == 14 && do {
                            print "Error - ftp passive mode is badly defined,\nvalid values are: yes/no or not defined at all,\ncheck [ ftp_passive_mode ].\n\n";
                        };
        $error_flag == 15 && do {
                            print "Error - ftp url is badly defined,\ncheck [ output_backup ].\n\n";
                        };
        $error_flag == 16 && do {
                            print "Error - ftp timeout value is invalid,\nit must be a valid number meaning a number over zero, example 480,\ncheck [ ftp_timeout ].\n\n";
                        };
        $error_flag == 17 && do {
                            print "Error - scp (secure copy) backups are not yet supported,\ncheck [ output_backup ].\n\n";
                        };
        $error_flag == 18 && do {
                            print "Error - tape backups are not supported in the windows operating system,\ncheck [ output_backup ].\n\n";
                        };
        $error_flag == 19 && do {
                            print "Error - badly defined tape backup or it's not pointing into a block device,\ncheck [ output_backup ].\n\n";
                        };
        $error_flag == 20 && do {
                            print "Error - invalid mail server type,\nonly smtp mail servers are compatible with simplebackup,\ncheck [ mail_type ].\n\n";
                        };
        $error_flag == 21 && do {
                            print "Error - invalid mail server name or ip,\ncheck [ mail_server ].\n\n";
                        };
        $error_flag == 22 && do {
                            print "Error - incomplete mail authentication,\nyou must fully define the mail username and\npassword or not defining them at all,\ncheck [ mail_username ; mail_password ].\n\n";
                        };
        $error_flag == 23 && do {
                            print "Error - invalid e-mail address,\ncheck [ mail_from_address ].\n\n";
                        };
        $error_flag == 24 && do {
                            print "Error - invalid e-mail address(es),\ncheck [ mail_to_addresses ].\n\n";
                        };
        $error_flag == 25 && do {
                            print "Error - using mail configuration but no mail server is configured,\ncheck [ mail_server ; mail_from_address ; mail_to_addresses ;\nmail_username ; mail_password ].\n\n";
                        };
        $error_flag == 26 && do {
                            print "Error - mail port value is invalid,\nit must be a valid number meaning a number over zero, example 25,\ncheck [ mail_port ].\n\n";
                        };
        $error_flag == 27 && do {
                            print "Error - mail timeout value is invalid,\nit must be a valid number meaning a number over zero, example 480,\ncheck [ mail_timeout ].\n\n";
                        };
        $error_flag == 28 && do {
                            print "Error - invalid backup mode,\nvalid values are: full ;  incremental ; differential,\ncheck [ backup_mode ].\n\n";
                        };
        $error_flag == 29 && do {
                            print "Error - invalid full backup day,\nvalid values are: monday ; tuesday ;\nwednesday ; thursday ; friday ;\nsaturday ; sunday ; first_day_month ;last_day_month ; first_day_year ;\nlast_day_year,\ncheck [ full_backup_day ].\n\n";
                        };
        $error_flag == 30 && do {
                            print "Error - the path into a valid management file must be declared if\nsimplebackup is working with incremental or differential backups,\ncheck [ backup_mode ; backup_db_management_file ].\n\n";
                        };
        $error_flag == 31 && do {
                            print "Error - invalid file limit size,\nvalid values are: 0 ; or any valid positive number,\ncheck [ backup_file_size_limit ].\n\n";
                        };
        $error_flag == 32 && do {
                            print "Error - invalid directory configuration,\nvalid values are: yes ; no,\ncheck [ backup_directory_rejections ].\n\n";
                        };
        $error_flag == 33 && do {
                            print "Error - the path to the database backup management file\nand the Local Database file history cannot be the same,\ncheck [ backup_db_management_file ; local_db_history_file ].\n\n";
                        };
        $error_flag == 34 && do {
                            print "Error - cannot have defined tape procedures if not doing a tape backup,\ncheck [ tape_before_backup ; tape_after_backup ].\n\n";
                        };
        $error_flag == 35 && do {
                            print "Error - a history file must be defined if doing a backup into a tape device,\ncheck [ local_db_history_file ; output_backup ].\n\n";
                        };
        $error_flag == 36 && do {
                            print "Error - invalid mail debug level,\nvalid values are: none ; full,\ncheck [ mail_debug_level ].\n\n";
                        };
        $error_flag == 37 && do {
                            print "Error - invalid ftp debug level,\nvalid values are: none ; full,\ncheck [ mail_debug_level ].\n\n";
                        };
        $error_flag == 38 && do {
                            print "Error - Net::FTP module wasn't found,\nthis is required in order to support ftp backups,\nyou must install the libnet package that contains the support for ftp backups\nand email reporting and is included in the original simplebackup package.\n\n";
                        };
        $error_flag == 39 && do {
                            print "Error - Net::SMTP module wasn't found,\nthis is required in order to support email reporting,\nyou must install the libnet package that contains the support for ftp backups\nand email reporting and is included in the original simplebackup package.\n\n";
                        };
        $error_flag == 40 && do {
                            print "Error - Backup directory/file [ $error_file_directory ],\ndoes not exist, it's not defined has a full path\nor you do not have permissions to acess it,\ncheck [ input_backup ].\n\n";
                        };
        $error_flag == 41 && do {
                            print "Error - ftp connect retries value is invalid,\nit must be a valid number meaning a number over zero, example 6,\ncheck [ ftp_connect_retries ].\n\n";
                        };
        $error_flag == 42 && do {
                            print "Error - repeated backup directory/file [ $error_file_directory ],\ncheck [ input_backup ].\n\n";
                        };
        $error_flag == 43 && do {
                            print "Error - invalid hostname on backup files value,\nvalid values are: yes/no or not defined at all,\ncheck [ hostname_on_backup_files ].\n\n";
                        };
        $error_flag == 44 && do {
                            print "Error - Authen::SASL module wasn't found,\nthis is required in order to support mail servers that request user\nauthentication,\nthis module is included in the original simplebackup package.\n\n";
                        };

    }

    return $error_flag;
}


#
# Show the configuration read from the conf file,
# NOTE !! this should only be run if the configuration is correct (passed sucessufly in the
# read_config() function!!)
# Will return: 0
# Explain: 0 is sucess
sub show_config {

    my $global_config_p = $_[0]; # configuration pointer

    # put variables in local ones to make them easy to place inside strings
    my $running_host_os = $^O; # type of the host operating system
    my $configuration_file = $$global_config_p{configuration_file};
    my $configuration_file_name = $$global_config_p{configuration_file_name};
    my $host_os = $$global_config_p{host_os};
    my $host_null_device = $$global_config_p{host_null_device};
    my $backup_format = $$global_config_p{backup_format};

    my $backup_mode = $$global_config_p{backup_mode};
    my $backup_real_mode = $$global_config_p{backup_real_mode};
    my $full_backup_day = $$global_config_p{full_backup_day};
    my $backup_db_management_file = $$global_config_p{backup_db_management_file};
    my $backup_extension_rejections = $$global_config_p{backup_extension_rejections};
    my $backup_filenames_rejections = $$global_config_p{backup_filenames_rejections};
    my $backup_file_size_limit = $$global_config_p{backup_file_size_limit};
    my $backup_directory_rejections = $$global_config_p{backup_directory_rejections};
    my $backup_compression_ratio = $$global_config_p{backup_compression_ratio};
    my $backup_real_ratio = $$global_config_p{backup_real_ratio};
    my $input_backup = $$global_config_p{input_backup};
    my @input_backup_split;
    my $input_backup_files_count = 0; # count the number of backup files
    my $input_backup_directories_count = 0; # count the number of backup directories
    my $output_backup = $$global_config_p{output_backup};
    my $temporary_dir = $$global_config_p{temporary_dir};
    my $keep_last_n_files = $$global_config_p{keep_last_n_files};
    my $run_before_backup = $$global_config_p{run_before_backup};
    my $run_after_backup = $$global_config_p{run_after_backup};
    my $log_file = $$global_config_p{log_file};
    my $pass_log_to_external_cmd = $$global_config_p{pass_log_to_external_cmd};
    my $local_db_history_file = $$global_config_p{local_db_history_file};
    my $tape_device = $$global_config_p{tape_device};
    my $tape_command = $$global_config_p{tape_command};
    my $tape_before_backup = $$global_config_p{tape_before_backup};
    my $tape_after_backup = $$global_config_p{tape_after_backup};
    my $hostname_on_backup_files = $$global_config_p{hostname_on_backup_files};
    my $hostname = $$global_config_p{hostname};
    my $mail_debug_level = "";
    my $explain_keep_last_n_files = "";
    my $ftp_debug_level = "";
    my $ftp_server = "";
    my $ftp_port = "";
    my $ftp_username = "";
    my $ftp_password = "";
    my $ftp_directory = "";
    my $ftp_passive_mode = "";
    my $ftp_proxy = "";
    my $ftp_timeout = "";
    my $mail_server = "";
    my $mail_type = "";
    my $mail_port = "";
    my $mail_timeout = "";
    my $mail_from_address = "";
    my $mail_to_addresses = "";
    my $mail_username = "";
    my $mail_password = "";

    @input_backup_split = split(/</,  $$global_config_p{input_backup}); # real all backup directories
    foreach my $my_file_dir (@input_backup_split) { # count the number of files and directories
        if ( -d "$my_file_dir" ) {
            $input_backup_directories_count++; # it's a directory count it
        }
        else {
            $input_backup_files_count++; # it's a file count it
        }
    }

    if ( $backup_extension_rejections eq "" ) {
        $backup_extension_rejections = "None configured";
    }

    if ( $backup_filenames_rejections eq "" ) {
        $backup_filenames_rejections = "None configured";
    }

    if ( $backup_file_size_limit == 0 ) {
        $backup_file_size_limit = "None configured";
    }
    else {
        $backup_file_size_limit = "$backup_file_size_limit Kbytes";
    }

    if ( $$global_config_p{backup_working} eq "ftp" ) { # if the backup type is ftp
        $ftp_server = $$global_config_p{ftp_server};        # Remote ftp server.
        $ftp_port = $$global_config_p{ftp_server_port};      # Remote ftp server port number.
        $ftp_username = $$global_config_p{ftp_username};     # FTP Username.
        $ftp_password = $$global_config_p{ftp_password};     # FTP Password
        $ftp_directory = $$global_config_p{ftp_directory};   # FTP directory where we will upload the backup file.
        $ftp_passive_mode = $$global_config_p{ftp_passive_mode}; # FTP passive mode or nor
        $ftp_debug_level = $$global_config_p{ftp_debug_level}; # FTP debug level
        $ftp_proxy = $$global_config_p{ftp_proxy}; # FTP proxy server (if available)
        if ( $ftp_proxy eq "" ) {
            $ftp_proxy = "None configured";
        }
        $ftp_timeout = $$global_config_p{ftp_timeout}; # FTP timeout
    }

    if ( $run_before_backup eq "" ) {
        $run_before_backup = "None configured";
    }
    if ( $run_after_backup eq "" ) {
        $run_after_backup = "None configured";
    }

    if ( $tape_before_backup eq "" ) {
        $tape_before_backup = "None configured";
    }

    if ( $tape_after_backup eq "" ) {
        $tape_after_backup = "None configured";
    }

    # deal with the external command that can process the current log
    if ( $pass_log_to_external_cmd eq "" ) {
        $pass_log_to_external_cmd = "None configured";
    }
    else {
        $pass_log_to_external_cmd = "current_log | $pass_log_to_external_cmd";
    }
    
    if ( $backup_mode eq "full" ) { # if backup is in full backup mode
        if ( $keep_last_n_files == -1 ) {
            $explain_keep_last_n_files = "Keep ALL"; # keep all backup files
        }
        elsif ( $keep_last_n_files == 0 ) {
            $explain_keep_last_n_files = "Keep just the current"; # keep just the current backup file
        }
        else {
            $explain_keep_last_n_files = "$keep_last_n_files plus the current";
        }
    }

    # mail server stuff
    $mail_server = $$global_config_p{mail_server};
    if ( $mail_server ne "" ) {
        $mail_port = $$global_config_p{mail_port};
        $mail_timeout = $$global_config_p{mail_timeout};
        $mail_type = $$global_config_p{mail_type};
        $mail_from_address = $$global_config_p{mail_from_address};
        $mail_to_addresses = $$global_config_p{mail_to_addresses};
        $mail_username = $$global_config_p{mail_username};
        $mail_password = $$global_config_p{mail_password};
        $mail_debug_level = $$global_config_p{mail_debug_level};
    }

    print "\nHostname ..........: [ $hostname ]\n";
    print "Configuration file : [ $configuration_file ]\n\n";
    print "Directories & Files...\n";
    print " To backup .....: [ $input_backup ]\n";
    if ( ($input_backup_files_count + $input_backup_directories_count) == 1 ) {
        print "                  Will do the backup of $input_backup_files_count file\n" if $input_backup_files_count == 1;
        print "                  Will do the backup of $input_backup_directories_count directory\n" if $input_backup_directories_count == 1;
    }
    else {
        print "                  Will do the backup of $input_backup_directories_count directori(es)\n";
        print "                  and $input_backup_files_count file(s)\n";
    }
    print " Temporary .....: [ $temporary_dir ]\n";
    print " Backup destiny : [ $output_backup ]\n";
 
    if ( $$global_config_p{backup_working} eq "ftp" ) {
        print "\n  FTP connection details...\n";
        print "   Debug level .....: $ftp_debug_level\n";
        print "   Server ..........: $ftp_server\n";
        print "   Server Port .....: $ftp_port\n";
        print "   Username ........: $ftp_username\n";
        print "   Password ........: $ftp_password\n";
        print "   Remote Directory : $ftp_directory\n";
        print "   Passive mode ....: $ftp_passive_mode\n";
        print "   Proxy ...........: $ftp_proxy\n";
        print "   Timeout .........: $ftp_timeout\n";
    }
    elsif ( $$global_config_p{backup_working} eq "tape" ) {
        print "\n  Tape details...\n";
        print "   Device ..................: $tape_device\n";
        print "   Tape program ............: $tape_command\n";
        print "   Run before writting tape : $tape_before_backup\n";
        print "   Run after writing tape ..: $tape_after_backup\n";
    }
    print "\nGeneric...\n";
    print " Selected O.S. (host_os) .................: $host_os ($running_host_os)\n";
    print " O.S. null device is .....................: $host_null_device\n";
    print " Include hostname on the backup files ....: $hostname_on_backup_files\n";
    print " Backup mode is ..........................: $backup_mode\n";
    if ( $backup_mode ne "full" ) {
        print " Today's backup will be ..................: $backup_real_mode\n";
        print " Full backup day .........................: $full_backup_day\n";
    }
    print " Backup rejections (file type) ...........: $backup_extension_rejections\n";
    print " Backup rejections (file name) ...........: $backup_filenames_rejections\n";
    print " Backup file size limit ..................: $backup_file_size_limit\n";
    print " Use [ simplebackup_rejection.txt ] files : $backup_directory_rejections\n";
    print " Backup archiver format ..................: $backup_format\n";
    print " Backup compression ratio ................: $backup_compression_ratio\n";
    print " Real backup ratio (passed to archiver) ..: $backup_real_ratio\n";
    print " Backup sessions to keep .................: ";
    if ( $backup_mode eq "full" ) {
        print "$keep_last_n_files ($explain_keep_last_n_files)\n";
    }
    else {
        print "Auto managed\n";
    }
    print " Log is passed to external command using .: $pass_log_to_external_cmd\n";
    print "\nExternal stuff...\n";
    print " Run before backup : $run_before_backup\n";
    print " Run after backup .: $run_after_backup\n";
    if ( $local_db_history_file ne "" ) {
        print "\nLocal history file is [ $local_db_history_file ]\n";
    }
    if ( $backup_db_management_file ne "" ) {
        print "\nBackup management file is [ $backup_db_management_file ]\n";
    }
    print "\nE-mail reporting...\n";
    if ( $mail_server ne "" ) {
        print "   Server type ..........: $mail_type\n";
        print "   Debug level ..........: $mail_debug_level\n";
        print "   Mail server ..........: $mail_server\n";
        print "   Server port ..........: $mail_port\n";
        print "   Timeout ..............: $mail_timeout\n";
        print "   From e-mail address ..: $mail_from_address\n";
        print "   To e-mail address(es) : $mail_to_addresses\n";
        print "   Mail Authentication...\n";
        if ( $mail_username ne "" ) {
            print "      Username ..........: $mail_username\n";
            print "      Password ..........: $mail_password\n";
        }
        else {
            print "      -- None configured --\n";
        }
    }
    else {
        print "   -- None configured --\n";
    }
    print "\nLog file is [ $log_file ]\n\n";

    return 0; # always return sucess
}


#
# Attempt to determin what is the operating system we are running, and we always return
# a operating system value
# Will return: unix ; winnt4 ; win2000/xp/2003
# Explain: unix is for any operating system than is not windows; all other and windows
sub get_os_type {

    my $running_host_os = $^O; # type of the host operating system

    if ( $running_host_os eq "MSWin32" ) { # under some kind of windows
        $running_host_os = "winnt4"; #defaults to winnt4
        my $try_ver = `ver 2>nul`; # attempt to use the operating system ver command
        if ( defined $try_ver and ($try_ver =~ /Windows XP/ or $try_ver =~ /Windows 2000/ or $try_ver =~ /Windows 2003/) ) {
            $running_host_os = "win2000/xp/2003"; #defaults to winnt4
        }
    }
    else { # not windows defaults to unix
        $running_host_os = "unix";
    }

    return $running_host_os; # return the operating system type
}


#
# Return the most apropriate temporary directory for the running operating system
# Will return: a path (always)
# Explain: the path is the temporary directory
sub get_tmp_dir {

    my $host_type = $_[0]; # configuration pointer
    my $temporary_path = "";

    if ( $host_type eq "unix" ) { # if running on unix
        $temporary_path = "/tmp"; # In unix /tmp is always the default path
    }
    else { # if under windows
        $temporary_path = $ENV{TEMP} if ( $temporary_path eq "" and defined $ENV{TEMP} and $ENV{TEMP} ne "" and -d $ENV{TEMP} ); # if using TEMP
        $temporary_path = $ENV{temp} if ( $temporary_path eq "" and defined $ENV{temp} and $ENV{temp} ne "" and -d $ENV{temp} ); # same, but temp
        $temporary_path = $ENV{TMP} if ( $temporary_path eq "" and defined $ENV{TMP} and $ENV{TMP} ne "" and -d $ENV{TMP} ); # if using TMP
        $temporary_path = $ENV{tmp} if ( $temporary_path eq "" and defined $ENV{tmp} and $ENV{tmp} ne "" and -d $ENV{tmp} ); # same, but tmp
        $temporary_path = "c:\\windows\\temp" if ( $temporary_path eq "" and -d "c:\\windows\\temp" ); # try c:\windows\temp
        $temporary_path = "c:\\winnt\\temp" if ( $temporary_path eq "" and -d "c:\\winnt\\temp" ); # try c:\winnt\temp
        $temporary_path = "c:\\" if ( $temporary_path eq "" ); # last attempt try c:\
    }

    return $temporary_path; # return the temporary dir path
}


#
# Receive a string containing a day order, and check if that order matches the system day
# Will return: yes ; no ; error
# Explain: yes is the today ; no it's not today ; invalid order !
sub is_today {

    my $check_order = $_[0];
    my $return_value = "no"; # by default is no

    if ( $check_order eq "monday" or $check_order eq "tuesday" or $check_order eq "wednesday" or $check_order eq "thursday" or $check_order eq "friday" or $check_order eq "saturday" or $check_order eq "sunday" or $check_order eq "first_day_month" or $check_order eq "last_day_month" or $check_order eq "first_day_year" or $check_order eq "last_day_year" ) {
        switch: {
            # weekday backup
            ($check_order eq "monday" or $check_order eq "tuesday" or $check_order eq "wednesday" or $check_order eq "thursday" or $check_order eq "friday" or $check_order eq "saturday" or $check_order eq "sunday") && do {
                    my $wday = (localtime(time))[6]; # get the weekday number
                    my @days   = ('sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'); # define the weekdays names
                    if ( $days[$wday] eq $check_order ) { # if the present weekday  is equal to the one configured by the user
                        $return_value = "yes";
                    }
            };
            # montly backup/year backup
            ($check_order eq "first_day_month" or $check_order eq "last_day_month" or $check_order eq "first_day_year" or $check_order eq "last_day_year") && do {
                my ($mday, $month) = (localtime(time))[3,4]; # get the monday day 1..28/29/30/31 + the month  0..11
                $month++; # just to make it simple... month 1..12
                if ( $mday == 1 and $check_order eq "first_day_month" ) { # backup on the first day of the month
                    $return_value = "yes";
                }
                elsif ( $mday == 1 and $month == 1 and $check_order eq "first_day_year" ) { # backup on the first day of year
                    $return_value = "yes";
                }
                elsif ( $check_order eq "last_day_month" ) { # let's look for the last day of the month
                    if ( $mday == 30 and ( $month == 4 or $month == 6 or $month == 9 or $month == 11 ) ) { # look for months with the last day has 30 ! remember $month is 1..12
                        $return_value = "yes";
                    }
                    elsif ( $mday == 31 and ( $month == 1 or $month == 3 or $month == 5 or $month == 7 or $month == 8 or $month == 10 or $month == 12 ) ) { # look for months with the last day has 31 ! remember $month is 0..12
                        $return_value = "yes";
                    }
                    elsif ( $month == 2 and $mday == 29 ) { # last day of february on a bi year
                        $return_value = "yes";
                    }
                    elsif ( $month == 2 and $mday == 28 and is_leap_year() eq "no" ) { # check if this is really the last day of the february
                        $return_value = "yes";
                    }
                }
                elsif ( $month == 12 and $mday == 31 and $check_order eq "last_day_year" ) { # let's look for the last day of the year
                    $return_value = "yes";
                }
            };
        }
    }
    else {
        $return_value = "error"; # error, invalid order
    }

    return $return_value;

}


#
# Return if the present year is a leap year or not
# Will return: yes ; no
# Explain: yes is is ; no it's not
sub is_leap_year {

    my $return_value = "no"; # by default it's not
    my $year = (localtime(time))[5];

    $year += 1900; # Add 1900 to the year to get the full 4 digit year
    if ( ($year % 4  == 0 and $year % 100 > 0) or $year % 400 == 0 ) { # check if the a leap year
        $return_value = "yes"; # year has 366 days
    }

    return $return_value;
}


#
# Receive a string, and a host type and return if the string is a full path or not
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 is failure
sub check_if_full_path {

    my $my_path = $_[0]; # string that get's the path
    my $host_type = $_[1]; # windows / unix
    my $error_flag = 0; # return value

    if ( $host_type eq "unix" and substr($my_path,0,1) ne "/" ) {
        $error_flag = 1;
    }
    elsif ( $host_type ne "unix" and substr($my_path,1,1) ne ":" ) {
        $error_flag = 1;
    }

    return $error_flag;

}


#
# Receive a string separated by a special char, and remove extra spaces bettwen that char
# and return that new string, the separator char can be , or <
# example : *hello , i'm separating,suff, ok*    will became = *hello,i'm separating,suff,ok*
# example : *hello<  i'm separating<suff< ok*    will became = *hello<i'm separating<suff<ok*
# Will return: a string
sub remove_excess_spaces {

    my $input_string = $_[0]; # the string to fix
    my $separator_char  = $_[1]; # the separator char, supported are , and <
    my $fixed_string = ""; # the fixed string to return

    # deal with the separator char
    my @my_rsplit;
    switch: {
        $separator_char eq ',' && do { # separator char is ,
                            @my_rsplit = split(/,/, $input_string);
                        };
        $separator_char eq '<' && do { # separator char is <
                            @my_rsplit = split(/</, $input_string);
                        };
    }
    # remove extra spaces
    foreach my $entry ( @my_rsplit ) {
        $entry =~ s/^\s+//;  # Delete leading whitespace ^ is the first caracter  \s is space then // replace with nothing if there was /a/ then replace with a
        $entry =~ s/\s+$//;  # Delete trailing whitespace same thing kind of but $ represents end of line.
        if ( $fixed_string ne "" ) {
            $fixed_string = "$fixed_string$separator_char$entry";
        }
        else { # first entry
            $fixed_string = "$entry";
        }
    }

    return $fixed_string; # return the fixed string
}


#
# Write a log file into the script log file, this should be used to log error messages
# Will also add the message into a message variable define by a pointer
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 is failure
sub write_file_log {

    my $global_config_p = $_[0]; # configuration pointer
    my $error_msg = $_[1]; # error message
    my $error_flag = 0; # write file control flag
    my $full_date = (localtime(time())); # present date in the default format
    my $configuration_file_name = $$global_config_p{configuration_file_name};
    my $new_log = "$full_date - $error_msg\n";
    if  ( defined $_[2] ) { # pointer into a variable that contains all the error messages,
                            # this will be used in the mail backup report, it can be safely ignore
                            # if required
        my $p_all_error_msg = $_[2];
        $$p_all_error_msg = "$$p_all_error_msg$new_log"; # add message into the general variable of all the error messages, the e-mail message will not contain the configuration file name
    }

    $new_log = "[ $configuration_file_name ] $new_log"; # add the configuration file name into the file log, this info will not appear in the e-message
    my $log_file = $$global_config_p{log_file};
    # open or create the log file in append >> write mode
    open(log_file_handle, ">>$log_file") or $error_flag = 1;
    if ( $error_flag == 0 ) {
        #lock the file for exclusive script use
        flock (log_file_handle, LOCK_EX) or $error_flag = 1;
        # write the log message
        print log_file_handle "$new_log", or $error_flag = 1;
        close (log_file_handle); # close the file
    }
    else {
        print "Warning - unable to write into the log file\n\n";
    }

    return $error_flag; # return the result
}


#
# Work with the local db file which will contain the list of the files managed by simplebackup
# on the current backup session, this is used to work with some more "strange" ftp servers
# can: read and return the entire db in a array,file names and sizes (read_all) ; read and
#      return all file names in a array (real_all_filenames) ; add a new entry (add) ;
#      remove one entry (delete).
# Will return: undef ; "" (empty string/array) ; some array
# Explain: undef is failure, "" (empty string/array) is sucess and marks a empty file, some array is sucess and the array is the data of file;
sub read_write_history_file {

    my $global_config_p = $_[0]; # configuration pointer
    my $file_name = $_[1]; # file name to be added or removed (it can be empty !)
    my $file_size = $_[2]; # file size to be added (it can be empty !)
    my $log_body_p = $_[3]; # pointer into the e-mail body string
    my $db_operation = $_[4]; # db operation to be done, valid values are: read_all ; add ; delete

    my $local_db_history_file = $$global_config_p{local_db_history_file};
    my $error_flag = 0;
    my @db_array;

    if ( -f $local_db_history_file ) { # if file exists

        open(local_db_history_file_handle, "<$local_db_history_file") or $error_flag = 1; # open file in read mode
        if ( $error_flag == 0 ) {
            my $last_line = ""; # to keep the last read line
            my $line_counter = 0; # counter to count the line number

            while(<local_db_history_file_handle>) { # read the entire file into the array
                $line_counter++; # add a line number
                chomp($_); # remove the new line chars
                if ( $line_counter == 1 and $_ ne "<-- start of history file -->" ) { # search if first line marks the start of file
                    $error_flag = 1; # mark error missing start of file line
                }
                elsif ( $line_counter > 1 and $_ ne "<-- end of history file -->" ) {
                    my @my_split = split( /</, $_);
                    # check if there is ONLY one : char and if there is something like 1.compress_file.zip*5234
                    if ( scalar(@my_split) == 2 and ($my_split[1] =~ /^\d+$/ and $my_split[1] ne 0) ) {
                        # all ok in this line keep it
                        if ( $db_operation eq "read_all_filenames" ) { # read only the file names
                            push @db_array, $my_split[0]; # extract only the filename
                        }
                        else { # extract the entire line
                            push @db_array, $_;
                        }
                    }
                    else { # corrupted line
                        $error_flag = 1; # mark error
                    }
                }
                if ( $last_line eq "<-- end of history file -->" ) { # detect duplicated end of file markers
                    $error_flag = 1
                }
                $last_line = $_; # backup the read line
                last if $error_flag != 0; # break cicle if any error occured
            }
            close (local_db_history_file_handle); # close file
            if ( $last_line ne "<-- end of history file -->" ) { # search if first line marks the start of file
                $error_flag = 1; # mark error missing end of file line
            }
        }
        if ( $error_flag != 0 ) { #  detect any errors
            write_file_log($global_config_p, "Error - unable to read the local history file or the file is corrupted [ $local_db_history_file ]", $log_body_p); # write the error message into the script error file
        }
    }
    if ( $error_flag == 0 ) {
        # deal with the several options
        switch: {
            ( $db_operation eq "read_all" or $db_operation eq "read_all_filenames" ) && do { # return the entire file in a array
                    if ( scalar(@db_array) > 0 ) { # if data was found on the file
                        return @db_array; # return data
                    }
                    else {
                        return ""; # file does not exist or it's empty
                    }
                };
            ( $db_operation eq "add" or $db_operation eq "delete" ) && do { # add a new entry on the db file
                    my $mark_do_write = "no"; # mark if we are going to rewrite the file or not
                    if ( $db_operation eq "add" ) { # if we are adding
                        push @db_array, "$file_name<$file_size"; # build the new line
                        $mark_do_write = "yes"; # rewrite the file
                    }
                    else { # if we are deleting
                        my $line_erase_marker = -1; # used the mark the array entry to erase
                        my $counter = 0;
                        # search
                        foreach my $array_line (@db_array)
                        {
                            my ($my_filename, $my_filesize) = split( /</, $array_line); # get the file name
                            if ( "$my_filename" eq "$file_name" ) { # if the filename on the file matches the one we want to erase
                                $line_erase_marker = $counter;
                                last; # found it now break from the foreach cicle
                            }
                            $counter++;
                        }
                        if ( $line_erase_marker > -1 ) { # erase only if it was found on the file!
                            splice(@db_array, $line_erase_marker, 1); # delete the value from the array
                            $mark_do_write = "yes"; # rewrite the file
                        }
                    }
                    if ( $mark_do_write eq "yes" ) { # if data is marked for rewriting on the file

                        open(local_db_history_file_handle, ">$local_db_history_file") or $error_flag = 1;
                        if ( $error_flag == 0 ) {
                            flock (local_db_history_file_handle, LOCK_EX) or $error_flag = 1; #lock the file for exclusive script use
                            # write the file header
                            print local_db_history_file_handle "<-- start of history file -->\n", or $error_flag = 1;
                            # write the line
                            foreach my $my_line ( @db_array ) { # re-write the entire file list
                                print local_db_history_file_handle "$my_line\n", or $error_flag = 1;
                                last if $error_flag != 0; # break cicle if any error occured
                            }
                            if ( $error_flag == 0 ) { # write file file end marker
                                print local_db_history_file_handle "<-- end of history file -->\n", or $error_flag = 1;
                            }
                            close (local_db_history_file_handle); # close the file
                        }
                        if ( $error_flag != 0 ) {
                            write_file_log($global_config_p, "Error - unable to erase or write into local history file [ $local_db_history_file ], $log_body_p"); # write the error message into the script error file
                        }
                    }
                };
        }
    }

    return 0 if $error_flag == 0; # return sucess
    return undef if $error_flag != 0; # return failure

}


#
# moving files from one directory to another,
# there is a choice to destinct the system from windows NT, Windows2000/XP ; Unix ; ftp servers
# Will return: 0 ; 1 ; 2
# Explain: 0 is sucess ; 1 is failure ; 2 is failure when attempted to erase a damaged backup file
#          from a ftp server
sub move_file {

    # note we used the \" to pass " into the variable to allow usage of names with spaces inside
    my $from_file_system = "\"$_[0]\""; # origin file, to pass to the operating system move/mv
    my $from_file_perl = $_[0]; # origin file, to be passed to perl commands
    my $to_file_system = "\"$_[1]\""; # destiny file, to pass to the operating system move/mv
    my $to_file_perl = $_[1]; # destiny file, to be passed to perl commands
    my $ftp_connection_p = $_[2]; # ftp connection object
    my $global_config_p = $_[3]; # configuration pointer
    my $log_body_p = $_[4]; # pointer into the e-mail body string

    my $null_device = $$global_config_p{host_null_device}; # operating system null device
    my $error_flag = 0; # to mark errors
    my @mv_command; # will contain the move command and the arguments

    if ( $$global_config_p{backup_working} eq "local file system" ) { # local file system move (mv / move)
        if ( !-f $to_file_perl ) { # only move files if there is no duplicated copy on the destiny location
        # deal with Operating System dependend stuff, error messages on windows are redirectry into the virtual file nul on unix into /dev/null
            switch: {
                $$global_config_p{host_os} eq "winnt4" && do {
                                                        @mv_command = ("move", "$from_file_system", "$to_file_system", ">nul"); # here null device must be done this way
                                                    };
                $$global_config_p{host_os} eq "win2000/xp/2003" && do {
                                                        @mv_command = ("move", "/y", "$from_file_system", "$to_file_system", "$null_device", ">nul");
                                                    };
                $$global_config_p{host_os} eq "unix" && do {
                                                        @mv_command = ("mv", "-f", "$from_file_system", "$to_file_system", "$null_device");
                                                    };
            }
            system("@mv_command") == 0 or $error_flag = 1; # if we return 1 something failed
        }
        else { # file with the same name already exists on the destiny directory
            $error_flag = 1; # mark error
        }
    }
    elsif ( $$global_config_p{backup_working} eq "ftp" ) { # ftp move (NET::FTP) and unlink command
        if ( $ftp_connection_p->size($to_file_perl) ) { # check if the file already exists on the ftp server
            write_file_log($global_config_p, "Error - the backup file already exists in the ftp server, the file is [ $to_file_perl ]", $log_body_p); # write the error message into the script error file
            $error_flag = 1; # mark error
        }
        else {
            $ftp_connection_p->binary(); # put in binary download mode
            if ( !$ftp_connection_p->put( "$from_file_perl", "$to_file_perl" ) ) {
                write_file_log($global_config_p, "Error - unable to upload the backup file into the ftp server, it might be lack of permissions, lack of free space or the ftp connection might have dropped, the file is [ $to_file_perl ]", $log_body_p); # write the error message into the script error file
                $error_flag = 1; # mark error
                # now attempt to delete the uploaded file from the ftp server, it might have been partly uploaded
                if ( $ftp_connection_p->size($to_file_perl) and $ftp_connection_p->size($to_file_perl) >= 0 ) { # check if the file exists
                    write_file_log($global_config_p, "Attempting to erase existing incomplete backup file on the ftp server", $log_body_p); # write the error message into the script error file
                    if ( !($ftp_connection_p->delete($to_file_perl)) ) { # erase the file
                        # warn that it failed to erase the previous backup file on the ftp server
                        write_file_log($global_config_p, "Error - failed to erase the damaged backup file on the remote ftp server, the connection might have dropped or you might not have the correct permissions", $log_body_p); # write the error message into the script error file
                        write_file_log($global_config_p, "Warning - the backup file [ $to_file_perl ] available on the ftp server is corrupted", $log_body_p); # write the error message into the script error file
                        return 2; # return error
                    }
                }
             }
            else { # upload sucess, erase the local file
                unlink "$from_file_perl"; # erasing the temporary file
            }
        }
    }
    elsif ( $$global_config_p{backup_working} eq "tape" ) { # tape backup
        my $final_file_name = get_filedir_name("$from_file_perl", $global_config_p);
        my $full_path_final_file_name = rebuild_full_path("$$global_config_p{temporary_dir}", "$final_file_name", $global_config_p); # rebuild the full path pointing also into the path defined by temporary_dir
        if ( ! chdir "$$global_config_p{temporary_dir}" ) { # attempt to change into the temporary directory
            $error_flag = 1; # return failure
        }
        else {
            rename($from_file_perl, $full_path_final_file_name) or $error_flag =1; # rename the file from something with _tmp_ into some someone without... example _tmp_1.myname.zip into 1.myname.zip
        }
        if ( $error_flag == 0 ) {

            my $local_tape_device = $$global_config_p{tape_device}; # tape device
            my @tape_command; # will contain the command that places the backup data on a the tape
            switch: {
                $$global_config_p{tape_command} eq "tar" && do { # if we are using tar to place data on tape
                            @tape_command = ("tar", "-cf", "\"$local_tape_device\"", "\"$full_path_final_file_name\"", "$null_device");
                            };
            }
            system("@tape_command") == 0 or $error_flag = 1; # do backup and if we return 1 something failed
        }
        unlink "$full_path_final_file_name" if ( -f "$full_path_final_file_name" ); # erase the left over file
    }
    # check if move worked, for backup modes not equal to tape
    if ( $$global_config_p{backup_working} ne "tape" ) {
        if ( $error_flag == 1 or (-f "$from_file_perl" and (!-f "$to_file_perl" and $$global_config_p{backup_working} eq "local file system") ) ) { # if there where any errors moving the file show them, being carefull to consider ftp backups
            $error_flag = 1; # mark error
        }
    }

    return $error_flag; # return the result
}


#
# Receive a full file/directory path and return the name of the directory or filename.
# Will return: a string
# Explain: the string it a directory path
sub get_filedir_name {

    my $filedir_name_path = $_[0]; # full path into directory or file
    my $global_config_p = $_[1]; # pointer into the configuration
    my $file_separator = $$global_config_p{host_file_separator};
    my @my_split;

    # deal with Operating System dependend stuff
    if ( $$global_config_p{host_os} eq "unix" ) {
        @my_split = split ( "/", $filedir_name_path ); # unix type paths
    }
    else {
        @my_split = split ( /\\/, $filedir_name_path ); # microsoft type paths
    }
    my $size_value = scalar(@my_split) - 1 ; # read the array size, note array starts at zero size is always bigger
    my $filedir_name = $my_split[$size_value]; # file or directory name

    # deal with Operating System dependend stuff again... check if it's backup up of / or d: ( drive: )
    if ( $$global_config_p{host_os} eq "unix" and ! defined $filedir_name ) { # unix type paths and the path is / (in this case it's detected as undefined)
        $filedir_name = "root"; # backup starting from the root drive /
    }
    elsif ( $filedir_name =~ ":" ) { # microsoft type paths and if it countains the : caracter example D: 
        my @new_my_split = split ( /:/, $filedir_name);
        $filedir_name = "disk_$new_my_split[0]"; # backup of a entire windows partition
    }

    return $filedir_name;
}


#
# Receive a full directory path and return the name of the last directory or filename,
# in a remote server (ftp/scp).
# Will return: a string
# Explain: the string it a directory path
sub get_filedir_name_remote {

    my $filedir_name_path = $_[0]; # full path into directory or file
    my @my_split;

    @my_split = split ( "/", $filedir_name_path ); # unix type paths

    my $size_value = scalar(@my_split) - 1 ; # read the array size, note array starts at zero size is always bigger
    my $filedir_name = $my_split[$size_value]; # file or directory name

    if ( $filedir_name eq "" ) { # if there is no / on the ftp file nameq
        $filedir_name = $filedir_name_path;
    }

    return $filedir_name;
}


#
# Return the previous directory name, example  /directory/another/directory the previous
# is /directory/another
# Will return the same directory if the path is a root directory example D:\ or /
# Will return: a string
# Explain: the string it a directory path
sub get_previous_filedir_name {

    my $filedir_name_path = $_[0]; # full path into directory or file
    my $global_config_p = $_[1]; # pointer into the configuration
    my @my_split;
    my $file_separator = $$global_config_p{host_file_separator};
    my $previous_dir_fullpath;

    # deal with Operating System dependend stuff
    if ( $$global_config_p{host_os} eq "unix" ) {
       @my_split = split ( "/", $filedir_name_path ); # unix type paths
    }
    else {
        @my_split = split ( /\\/, $filedir_name_path ); # microsoft type paths
    }
    my $size_value = scalar(@my_split); # array size
    if ( ( $size_value == 0 and $$global_config_p{host_os} eq "unix" ) or ( $size_value == 1 and $$global_config_p{host_os} ne "unix" ) ) { # we are at a starting point... D:\  or / for example
        $previous_dir_fullpath = $filedir_name_path; # unix paths
    }
    $size_value = $size_value - 2; # array size up to the previous backup directory
    if ( $size_value == 0 and $$global_config_p{host_os} ne "unix" ) { # if microsoft path's
        $previous_dir_fullpath = "$my_split[0]$file_separator";
        return $previous_dir_fullpath;
    }

    for my $i ( 0 .. $size_value ) {
        if ( $i == 0 and $size_value > 0 ) { # first part
            $previous_dir_fullpath = "$my_split[$i]";
        }
        elsif ( $i == 0 and $size_value == 0 and $$global_config_p{host_os} eq "unix" ) { # second part unix only
            $previous_dir_fullpath = $file_separator; # backuping up a unix directory close to root, example /etc
        }
        else { # next path's
            $previous_dir_fullpath = "$previous_dir_fullpath$file_separator$my_split[$i]";
        }
    }

    return $previous_dir_fullpath;
}


#
# Receive a date in the format DD-MM-YYYY (01-12-2004) and test if it's valid or not
# Please notice that this does not test the date in 100% !!!
# Will return: 0 ; 1
# Explain: 0 (valid date) ; 1 (invalid date)
sub test_date {

    my $date_to_test = $_[0]; # string representing the date
    my @date_to_test_parts = split(/-/, $date_to_test);
    # test the date
    if ( scalar(@date_to_test_parts) == 3 and (length($date_to_test_parts[0]) == 2 and $date_to_test_parts[0] =~ /^\d+$/ and $date_to_test_parts[0] > 0 and $date_to_test_parts[0] <= 31) and (length($date_to_test_parts[1]) == 2 and $date_to_test_parts[1] =~ /^\d+$/ and $date_to_test_parts[1] > 0 and $date_to_test_parts[1] <= 12) and (length($date_to_test_parts[2]) == 4 and $date_to_test_parts[2] =~ /^\d+$/ and $date_to_test_parts[2] >= 1980) ) {
        return 0; # valid date
    }

    return 1; # invalid date
}


#
# Receive the backup file/directory name, and a full filename and crc32 and test
# if the file is part of the backup session
# Will return: yes / no
# Explain: yes the file name is part ; no  the filename is not part
sub check_session_filename {

    my $backup_directory = $_[0]; # directory name and full path
    my $test_filename = $_[1]; # filename to test
    my $crc = $_[2]; # crc that represents the full path into the directory name represented by backup_directory_name
    my $global_config_p = $_[3]; # pointer into the configuration

    my @file_parts = split( /\./, "$test_filename" );
    my $file_parts_count = scalar(@file_parts); # count the number of . in the filename
    my $file_parts_count_minimal = 3; # the default 3, example 4.backup_name.meta label.zip
    # remove the full path, i only want the backup directory name
    $backup_directory = get_filedir_name("$backup_directory", $global_config_p);

    # if the hostname is on the start of the file name , example "myserver.5.perl.full monday 3-1-2005 crc97074048.zip"
    if ( $$global_config_p{hostname_on_backup_files} eq "yes" ) { 
        my $hostname_dotcount = 1;
        while ($$global_config_p{hostname} =~ /\./g) { $hostname_dotcount++ } # count the number of . chars
        $file_parts_count_minimal = $file_parts_count_minimal + $hostname_dotcount; # fix the value
    }
    # prelimilary chech on the filename
    if ( $file_parts_count >= $file_parts_count_minimal ) {

        my $backup_format = $$global_config_p{backup_format}; # extract the backup format
        my $backup_format_count = 1; # at least count 1 part (zip tar rar)
        while ($backup_format =~ /\./g) { $backup_format_count++ } # count the number of . chars
        my $file_meta_info = $file_parts[$file_parts_count-$backup_format_count-1]; # extract the simplebackup meta info (example full monday 12-12-2004)
        my @file_meta_info_parts =  split( / /, "$file_meta_info" );
        # test the file meta information
        if ( scalar(@file_meta_info_parts) == 5 and ($file_meta_info_parts[0] =~ /^\d+$/ and $file_meta_info_parts[1] eq "full" or $file_meta_info_parts[1] eq "incremental" or $file_meta_info_parts[1] eq "differential") and ($file_meta_info_parts[2] eq "monday" or $file_meta_info_parts[2] eq "tuesday" or $file_meta_info_parts[2] eq "wednesday" or $file_meta_info_parts[2] eq "thursday" or $file_meta_info_parts[2] eq "friday" or $file_meta_info_parts[2] eq "saturday" or $file_meta_info_parts[2] eq "sunday") and test_date($file_meta_info_parts[3]) == 0 and $file_meta_info_parts[4] eq "crc$crc" ) {
            # the file meta information is valid so rrebuild a full named file based on the file present session on the directory name
            # and on the configured extesions in order to to a full string match
            my $virtual_filename = "";
            # if with OUT the hostname on the backup file
            if ( $$global_config_p{hostname_on_backup_files} ne "yes" ) {
                $virtual_filename = "$backup_directory." . "$file_meta_info." . $backup_format;
            }
            else { # if with the hostname on the backup file
                $virtual_filename = $$global_config_p{hostname} . ".$backup_directory." . "$file_meta_info." . $backup_format;
            }
            if ( $virtual_filename eq $test_filename ) {
                return "yes"; # file is part of the backup
            }
        }
    }

    return "no"; # file is not part of the backup
}


#
# Receive the backup file/directory name and return the session number + the file meta information
# in a string separated by the space char. Please notive that the file name must be in the correct
# format so use the check_session_filename() before using this.
# Will return: a string, example "3 full monday 3-1-2005 crc97074048.zip"
sub get_filename_session_metadata {

    my $test_filename = $_[0]; # filename to test
    my $global_config_p = $_[1]; # pointer into the configuration
    my $return_string = ""; # what we are returning

    my @file_parts = split( /\./, "$test_filename" );
    my $file_parts_count = scalar(@file_parts); # count the number of . in the filename

    my $backup_format = $$global_config_p{backup_format}; # extract the backup format
    my $backup_format_count = 1; # at least count 1 part (zip tar rar)
    while ($backup_format =~ /\./g) { $backup_format_count++ } # count the number of . chars

    return $file_parts[$file_parts_count-$backup_format_count-1]; # return the values
}


# 
# Get the present backup id session number, to use on the file name... example 5 for the
# full name 5.backup_directory_name.rar
# It will also acess into the ftp server to get the session number or the local history
# file if required
# Will return: positive number >= 0
# Explain: 0 is failure , any other is sucess
sub get_present_session_number {

    my $ftp_connection_p = $_[0]; # pointer into the ftp connection object
    my $global_config_p = $_[1]; # pointer into the configuration
    my $log_body_p = $_[2]; # pointer into the e-mail body string

    my $session_number = 0; # default session number
    my $error_flag = 0; # error flag
    my @file_list; # to contain the file list

    if ( $$global_config_p{local_db_history_file} ne "" ) { # if we are using a local db file and not reading the actual file list
        @file_list = read_write_history_file( $global_config_p, "", "", $log_body_p, "read_all_filenames" ); # read the entire db file
        if ( !defined $file_list[0] ) {
            return 0; # return failure
        }
    }

    if ( $$global_config_p{backup_working} eq "local file system" and $$global_config_p{local_db_history_file} eq "" ) {
        opendir(dir_handle, "$$global_config_p{output_backup}") or $error_flag = 1; # open the load directory
        @file_list = readdir dir_handle; # read the file list
        closedir(dir_handle);# close the directory handle
    }
    elsif ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ){
        if ( ! defined $ftp_connection_p ) { # if the ftp connection exists
            return 0; # return error
        }
        my $ftp_directory = $$global_config_p{ftp_directory}; # prepare to change into the ftp remote directory
        # change into the upload directory
        if ( !$ftp_connection_p->cwd("$ftp_directory") ) {
            write_file_log($global_config_p, "Error - unable to acess the remote ftp directory [ $ftp_directory ]", $log_body_p); # write the error message into the script error file
            return 0; # return aborted
        }
        else { # if worked get the file list
            @file_list = $ftp_connection_p->ls;
        }
    }
    if ( $error_flag == 0 ) {
        foreach  my $file_name ( @file_list ) {
            if ( $file_name ne "." and $file_name ne ".." ) { # ignore the system file names . and ..
                if ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ) { # deal with anoying ftp servers that return "strange" names on the ls command, one example of this is the famouse windows war ftp deamon
                    $file_name = get_filedir_name_remote($file_name);
                }
                # Check if the file belongs to the backup (any backup directory)
                my @my_split = split(/</,  $$global_config_p{input_backup}); # real all backup directories
                foreach my $my_backup_file_directory( @my_split ) { # compare filename against all backup directories
                    my $my_crc = build_crc32("$my_backup_file_directory"); # build the crc32 number for this backup directory
                    if ( check_session_filename("$my_backup_file_directory","$file_name", "$my_crc", $global_config_p) eq "yes" ) { # match the present file against one backup directory
                        my ($file_session_number, $file_type, $file_weekday, $file_date, $file_crc32 ) = split ( / /, get_filename_session_metadata($file_name, $global_config_p)); # get the file session number
                        if ( $session_number < $file_session_number ) { # get the largest session id
                            $session_number = $file_session_number;
                        }
                    }
                }
            }
        }
    }
    else {
        my $bck_directory = $$global_config_p{output_backup};
        write_file_log($global_config_p, "Error - opening the backup destiny directory [ $bck_directory ]", $log_body_p); # write the error message into the script error file
        return 0; # return aborted
    }

    return $session_number + 1; # return the session id

}


#
# Clear the previous backup's
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 failure
sub clear_previous_backups {

    my $session_number = $_[0]; # current session number
    my $input_backup_done_p = $_[1]; # pointer into the array that countains the list of the data that was backup
    my $ftp_connection_p = $_[2]; # pointer into the ftp connection object
    my $global_config_p = $_[3]; # pointer into the configuration
    my $log_body_p = $_[4]; # pointer into the e-mail body string
    my $file_separator = $$global_config_p{host_file_separator}; # file separator
    my $error_flag = 0; # error flag
    my $erase_file_counter = 0;

    if ( $$global_config_p{keep_last_n_files} == "-1" ) { # nothing to erase
        write_file_log($global_config_p, "No previous backup sessions files were erased, configuration file disables backup sessions erases", $log_body_p); # write the error message into the script error file
        return 0;
    }
    else {
        my @file_list; # to contain the file list
        if ( $$global_config_p{local_db_history_file} ne "" ) { # if we are using a local db file and not reading the actual file list
            @file_list = read_write_history_file( $global_config_p, "", "", $log_body_p, "read_all_filenames" ); # read the entire db file
            if ( !defined $file_list[0] ) {
                return 1; # return failure
            }
        }

        if ( $$global_config_p{backup_working} eq "local file system" and $$global_config_p{local_db_history_file} eq "") {
            opendir(dir_handle, "$$global_config_p{output_backup}") or $error_flag = 1; # open the local backup directory
            @file_list = readdir dir_handle; # read the file list
            closedir(dir_handle);# close the directory handle
        }
        elsif ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ){ # get the list from the ftp server
            if ( ! defined $ftp_connection_p ) { # if the ftp connection exists
                return -1; # return error
            }
            my $ftp_directory = $$global_config_p{ftp_directory}; # prepare to change into the ftp remote directory
            # change into the upload directory
            if ( !$ftp_connection_p->cwd("$ftp_directory") ) {
                write_file_log($global_config_p, "Error - unable to acess the remote ftp directory [ $ftp_directory ]", $log_body_p); # write the error message into the script error file
                return 1; # return aborted
            }
            else { # if worked get the file list
                @file_list = $ftp_connection_p->ls;
            }
        }
        if ( $error_flag == 0 ) {
             my $check_session_number = $session_number - $$global_config_p{keep_last_n_files}; # calculate the first session to erase
             if ( $check_session_number <= 1 ) {
                 write_file_log($global_config_p, "No previous backup sessions files were erased, no previous backup sessions are in conditions to be erased", $log_body_p); # write the error message into the script error file
                 return 0; # there nothing to erase
             }
             foreach  my $file_name ( @file_list ) {
                 if ( $file_name ne "." and $file_name ne ".." ) { # ignore the system file names . and ..
                     if ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ) { # deal with anoying ftp servers that return "strange" names on the ls command, one example of this is the famouse windows war ftp deamon
                        $file_name = get_filedir_name_remote($file_name);
                     }
                     foreach my $my_backup_file_directory ( @$input_backup_done_p ) { # compare filename against all backup directories

                         my $my_crc = build_crc32("$my_backup_file_directory"); # build the crc32 number for this backup directory
                         if ( check_session_filename("$my_backup_file_directory","$file_name", "$my_crc", $global_config_p) eq "yes" ) {

                             my $session_number_id = 0;
                             my $erase_file = "yes";
                             my ($file_session_number, $file_type, $file_weekday, $file_date, $file_crc32 ) = split ( / /, get_filename_session_metadata($file_name, $global_config_p)); # get the file session number
                             for $session_number_id ( $check_session_number .. $session_number ) { # search if the file is to be erased
                                 if ( "$file_session_number" eq "$session_number_id" ) {
                                     $erase_file = "no"; # file is on the NON ERASE LIST
                                 }
                             }
                             if ( $erase_file eq "yes" ) { # file is marked to erase
                                 my $erase_me = ""; # to contain the full path to the file that is supose to be erased
                                 if ( $$global_config_p{backup_working} eq "local file system" ) { # local file system backup
                                     $erase_me = $$global_config_p{output_backup};
                                 }
                                 elsif ( $$global_config_p{backup_working} eq "ftp" ) { # ftp backup server
                                     $erase_me = $$global_config_p{ftp_directory}; # ftp directory
                                 }
                                 $erase_me =~ s/[\\\/]$// if "$erase_me" ne "/";;  # Delete trailing \ or / char same thing kind of but $ represents end of line
                                 if ( $erase_me eq "/" and ($$global_config_p{host_os} eq "unix" or $$global_config_p{backup_working} eq "ftp") ) { # unix type paths or ftp server
                                     $erase_me = ""; # protection against ftp paths that are / (root)
                                 }
                                 if ( $$global_config_p{backup_working} eq "ftp" ) {
                                     $erase_me = "$erase_me/$file_name"; # if ftp paths
                                 }
                                 else { # if unix or windows path's
                                     $erase_me = "$erase_me$file_separator$file_name";
                                 }
                                 if ( $$global_config_p{backup_working} ne "tape" ) {
                                     write_file_log($global_config_p, "Erasing previous backup session file [ $erase_me ]", $log_body_p); # write the error message into the script error file
                                 }
                                 else { # if "erasing from tape" this will only erase the record from the history file
                                     write_file_log($global_config_p, "Erasing previous backup session file from the history file (only !!) [ $erase_me ]", $log_body_p); # write the error message into the script error file
                                 }
                                 # if we are erasing on the local file system
                                 if ( $$global_config_p{backup_working} eq "local file system" ) { # erase the local file
                                     if ( unlink "$erase_me" ) { # if the erase went ok
                                         $erase_file_counter++; # add counter to erased file
                                         if ( $$global_config_p{local_db_history_file} ne "" and !-f $erase_me ) { # if we are using a local db file and not reading the actual file list
                                             if ( ! defined read_write_history_file( $global_config_p, "$file_name", "", $log_body_p, "delete" ) ) {  # erase the file from the database
                                                 $error_flag = 1; # some error occured mark error
                                                 last; # break cicle
                                             }
                                         }
                                     }
                                     else { # warn that it failed to erase the previous backup file on the local server
                                         write_file_log($global_config_p, "Warning - failed to erase the previous backup file, you might not have the correct permissions", $log_body_p); # write the error message into the script error file
                                     }
                                 }
                                 elsif ($$global_config_p{backup_working} eq "ftp" ) { # erase the remote ftp file
                                     if ( $ftp_connection_p->size($erase_me) ) { # check if the file exists
                                         if ( $ftp_connection_p->delete($erase_me) ) { # check if the erase of the remote file worked
                                             $erase_file_counter++; # add counter to erased file
                                         }
                                         else { # warn that it failed to erase the previous backup file on the ftp server
                                             write_file_log($global_config_p, "Warning - failed to erase the previous backup file on the remote ftp server, the connection might have dropped or you might not have the correct permissions", $log_body_p); # write the error message into the script error file
                                         }
                                     }
                                     else { # file does not exists
                                         write_file_log($global_config_p, "Warning - previous backup file does not exists on the remote ftp server", $log_body_p); # write the error message into the script error file
                                     }
                                     if ( $$global_config_p{local_db_history_file} ne "" and !$ftp_connection_p->size($erase_me)) { # if we are using a local db file and the remote file does not exists
                                         if ( ! defined read_write_history_file( $global_config_p, "$file_name", "", $log_body_p, "delete" ) ) {  # erase the file from the database
                                             $error_flag = 1; # some error occured mark error
                                             last; # break cicle
                                         }
                                     }
                                 }
                                 elsif ($$global_config_p{backup_working} eq "tape" ) { # "erase" from the tape device, this will only erase the entry from the history file
                                     if ( ! defined read_write_history_file( $global_config_p, "$file_name", "", $log_body_p, "delete" ) ) {  # erase the file from the database
                                         $error_flag = 1; # some error occured mark error
                                         last; # break cicle
                                     }
                                 }
                             }
                         }
                     }
                 }
             }
        }
    }
    if ( $error_flag == 0 ) { # only write this last message if there where no errors
        if ( $$global_config_p{backup_working} ne "tape" ) { # if actualy erasing session files
            write_file_log($global_config_p, "Erased [ $erase_file_counter ] previous backup session files", $log_body_p); # write the error message into the script error file
        }
        else {
            write_file_log($global_config_p, "Erased [ $erase_file_counter ] previous backup sessions from the history file", $log_body_p); # write the error message into the script error file
        }
    }

    return $error_flag; # return the result
}


#
# Sum all the current backup files sizes and return the total value (in bytes)
# Will return: -1 ; any_valid_positive_value
# Explain: -1 failure ; any_valid_positive_value is the total size of the current backup file(s) in bytes
sub sum_all_backups {

    my $ftp_connection_p = $_[0]; # pointer into the ftp connection object
    my $global_config_p = $_[1]; # pointer into the configuration
    my $log_body_p = $_[2]; # pointer into the e-mail body string
    my $file_separator = $$global_config_p{host_file_separator}; # file separator
    my $error_flag = 0; # error flag

    my @file_list; # to contain the file list
    if ( $$global_config_p{local_db_history_file} ne "" ) { # if we are using a local db file and not reading the actual file list
        @file_list = read_write_history_file( $global_config_p, "", "", $log_body_p, "read_all" ); # read the entire db file
        if ( ! defined $file_list[0] ) { # detect any possible errors
            return -1; # return error
        }
    }
    if ( $$global_config_p{backup_working} eq "local file system" and $$global_config_p{local_db_history_file} eq "") {
        opendir(dir_handle, "$$global_config_p{output_backup}") or $error_flag = -1; # open the local backup directory
        @file_list = readdir dir_handle; # read the file list
        closedir(dir_handle);# close the directory handle
    }
    elsif ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ){ # get the list from the ftp server
        if ( ! defined $ftp_connection_p ) { # if the ftp connection exists
            return -1; # return error
        }
        my $ftp_directory = $$global_config_p{ftp_directory}; # prepare to change into the ftp remote directory
        # change into the upload directory
        if ( !$ftp_connection_p->cwd("$ftp_directory") ) {
            write_file_log($global_config_p, "Error - unable to acess the remote ftp directory [ $ftp_directory ]", $log_body_p); # write the error message into the script error file
            $error_flag = -1; # mark error
        }
        else { # if worked get the file list
            @file_list = $ftp_connection_p->ls;
        }
    }
    if ( $error_flag == 0 ) {

        foreach  my $file_name ( @file_list ) {
            if ( $file_name ne "." and $file_name ne ".." ) { # ignore the system file names . and ..
                if ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ) { # deal with anoying ftp servers that return "strange" names on the ls command, one example of this is the famouse windows war ftp deamon
                    $file_name = get_filedir_name_remote($file_name);
                }
                my $local_history_filesize = ""; # if we are using a local history file, this will contain each file size
                if ( $$global_config_p{local_db_history_file} ne "" ) {
                    ($file_name, $local_history_filesize ) = split( /</, $file_name); # the data comes from the history file, there file it's something similar to... 1.filename.zip*1234
                }
                # Check if the file belongs to the backup (any backup directory)
                my @my_split = split(/</,  $$global_config_p{input_backup}); # real all backup directories
                foreach my $my_backup_file_directory ( @my_split ) { # compare filename against all backup directories
                    my $my_crc = build_crc32("$my_backup_file_directory"); # build the crc32 number for this backup directory
                    # check if the filename is correct
                    if ( check_session_filename("$my_backup_file_directory","$file_name", "$my_crc", $global_config_p) eq "yes" ) {
                        # if we are reading from a history file... in tape backups this is always used
                        if ( $$global_config_p{local_db_history_file} ne "" ) {
                            $error_flag = $error_flag + $local_history_filesize; # add the current file size to the total
                        }
                        else { # if working without a db list file
                            my $full_file_path = ""; # to contain the full path to the file
                            if ( $$global_config_p{backup_working} eq "local file system" ) { # local file system backup
                                $full_file_path = $$global_config_p{output_backup};
                            }
                            elsif ( $$global_config_p{backup_working} eq "ftp" ) { # ftp backup server
                                $full_file_path = $$global_config_p{ftp_directory}; # ftp directory
                            }
                            $full_file_path =~ s/[\\\/]$// if "$full_file_path" ne "/";;  # Delete trailing \ or / char same thing kind of but $ represents end of line
                            if ( $full_file_path eq "/" and ($$global_config_p{host_os} eq "unix" or $$global_config_p{backup_working} eq "ftp") ) { # unix type paths or ftp server
                                $full_file_path = ""; # protection against ftp paths that are / (root)
                            }
                            if ( $$global_config_p{backup_working} eq "ftp" ) {
                                $full_file_path = "$full_file_path/$file_name"; # if ftp paths
                            }
                            else { # if windows or unix path's
                                $full_file_path = "$full_file_path$file_separator$file_name";
                            }
                            eval {
                                if ( $$global_config_p{backup_working} eq "local file system" ) { # if reading from the local file system or a local mounted drive
                                    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$present_mtime,$ctime,$blksize,$blocks) = stat("$full_file_path");
                                    $error_flag = $error_flag + $size; # add the current file size to the total
                                }
                                elsif ( $$global_config_p{backup_working} eq "ftp" ) { # if reading from a ftp server
                                    $error_flag = $error_flag + $ftp_connection_p->size($full_file_path); # add the current file size to the total
                                }
                            };
                            if ( $@ ) { # detect if any error occured
                                $error_flag = -1; # mark error
                            }
                        }
                    }
                }
             }
        }
    }

    return $error_flag; # return the apropriated value
}


#
# Sum the types (full, incremental, differencial) of backup files based on a list (array)
# that contains the full path of the backup data
# Will return: undef ; a string
# Explain: undef is error ; a string containing the count off all types separated by the space char,
# example "10, 20, 0, 30" 10 is the full backups count ; 20 is the count of all incremental ; 0 is
# the count of all differencial and 30 is the count of all backups tapes 10 + 20 + 0 = 30
sub sum_backup_types {

    my $search_list_p = $_[0]; # pointer into the array that countains the list of the data to check
    my $ftp_connection_p = $_[1]; # pointer into the ftp connection object
    my $global_config_p = $_[2]; # pointer into the configuration
    my $log_body_p = $_[3]; # pointer into the e-mail body string

    my $error_flag = 0; # error flag
    # the counters
    my $full_counter = 0;
    my $incremental_counter = 0;
    my $differencial_counter = 0;
    my $total_counter = 0;

    my @file_list; # to contain the file list
    if ( $$global_config_p{local_db_history_file} ne "" ) { # if we are using a local db file and not reading the actual file list
        @file_list = read_write_history_file( $global_config_p, "", "", $log_body_p, "read_all_filenames" ); # read the entire db file
        if ( !defined $file_list[0] ) {
            $error_flag = 1; # mark error
        }
    }

    if ( $$global_config_p{backup_working} eq "local file system" and $$global_config_p{local_db_history_file} eq "") {
        opendir(dir_handle, "$$global_config_p{output_backup}") or $error_flag = 1; # open the local backup directory
        @file_list = readdir dir_handle; # read the file list
        closedir(dir_handle);# close the directory handle
    }
    elsif ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ){ # get the list from the ftp server
        if ( ! defined $ftp_connection_p ) { # if the ftp connection droped
            return 1; # exit imedicaly
        }
        my $ftp_directory = $$global_config_p{ftp_directory}; # prepare to change into the ftp remote directory
        # change into the upload directory
        if ( !$ftp_connection_p->cwd("$ftp_directory") ) {
            write_file_log($global_config_p, "Error - unable to acess the remote ftp directory [ $ftp_directory ]", $log_body_p); # write the error message into the script error file
            $error_flag = 1; # mark error
        }
        else { # if worked get the file list
            @file_list = $ftp_connection_p->ls;
        }
    }
    # continue if no errors occured acessing the backup list
    if ( $error_flag == 0 ) {
         foreach  my $file_name ( @file_list ) {
             if ( $file_name ne "." and $file_name ne ".." ) { # ignore the system file names . and ..
                 if ( $$global_config_p{backup_working} eq "ftp" and $$global_config_p{local_db_history_file} eq "" ) { # deal with anoying ftp servers that return "strange" names on the ls command, one example of this is the famouse windows war ftp deamon
                    $file_name = get_filedir_name_remote($file_name);
                 }
                 foreach my $my_backup_file_directory ( @$search_list_p ) { # compare filename against the list of what we want to search

                     my $my_crc = build_crc32("$my_backup_file_directory"); # build the crc32 number for this backup directory
                     if ( check_session_filename("$my_backup_file_directory","$file_name", "$my_crc", $global_config_p) eq "yes" ) {
                        # get the file metadata
                        my ($file_session_number, $file_type, $file_weekday, $file_date, $file_crc32 ) = split ( / /, get_filename_session_metadata($file_name, $global_config_p));
                        # start counting the types
                        switch: {
                            $file_type eq "full" && do {
                                            $full_counter++; # add to full
                                        };
                            $file_type eq "incremental" && do {
                                            $incremental_counter++; # add to incremental
                                        };
                            $file_type eq "differential" && do {
                                            $differencial_counter++; # add to differencial
                                        };
                        }
                     }
                 }
             }
         }
    }

    if ( $error_flag == 0 ) { # if no error occured
        $total_counter = $full_counter + $incremental_counter + $differencial_counter;
        return "$full_counter $incremental_counter $differencial_counter"; # return the values
    }
    else { # error was detected
        return undef; # return error
    }

}


#
# Will get all the configuration data from the main configuration hash, plus the
# dbi connection object and try to connect into the ftp server and login, if configured
# by the user on the configuration file it will retry the ftp connection for several
# times and change the passive mode setting
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 failure
sub ftp_connect_login {

    my $ftp_connection_p = $_[0]; # ftp dbi connection pointer
    my $global_config_p = $_[1]; # pointer into the configuration
    my $log_body_p = $_[2]; # pointer into the e-mail body string

    my $error_flag = 0; # error flag
    my $ftp_server = $$global_config_p{ftp_server}; # ftp server
    my $ftp_server_port = $$global_config_p{ftp_server_port}; # ftp server port
    my $ftp_passive_mode = $$global_config_p{ftp_passive_mode}; # passive mode
    my $ftp_passive_mode_real = $$global_config_p{ftp_passive_mode_real};   # The real parameter passed into the NET:FTP, here is 
    my $ftp_proxy = $$global_config_p{ftp_proxy};    # FTP Proxy, must work in open mode.
    my $ftp_timeout = $$global_config_p{ftp_timeout}; # ftp connection timeout
    my $ftp_username = $$global_config_p{ftp_username}; # ftp username
    my $ftp_password = $$global_config_p{ftp_password}; # ftp password
    my $ftp_debug_level_real = $$global_config_p{ftp_debug_level_real}; # ftp debug level
    my $ftp_connect_retries = $$global_config_p{ftp_connect_retries}; # # ftp connect retries

    # try a max of N times ftp connects before giving up
    for (my $connect_try=0; $connect_try < $ftp_connect_retries; $connect_try++) {
        if ( $connect_try == 3 ) { # if we are over 3 connections attempt to switch the ftp passive mode
            write_file_log($global_config_p, "Notice - reversing ftp passive mode", $log_body_p); # write the error message into the script error file
            if ( $$global_config_p{ftp_passive_mode} eq "no" ) { # passive mode no, switch to yes
                $ftp_passive_mode = "yes";
                $ftp_passive_mode_real = 1;
                $$global_config_p{ftp_passive_mode} = "yes";
                $$global_config_p{ftp_passive_mode_real} = 1; # The real parameter passed into the NET:FTP, here is [ 0 ; 1 ]
            }
            else { # passive mode yes switch to no
                $ftp_passive_mode = "no";
                $ftp_passive_mode_real = 0;
                $$global_config_p{ftp_passive_mode} = "no";
                $$global_config_p{ftp_passive_mode_real} = 0; # The real parameter passed into the NET:FTP, here is [ 0 ; 1 ]
            }
        }
        if ( $connect_try > 0 ) { # connect into the FTP server failed report that we are retrying
            write_file_log($global_config_p, "Warning - unable connect to the ftp server, retrying in 60 seconds", $log_body_p); # write the error message into the script error file
            sleep 60; # wait 60 seconds before attemping to connect
        }
        write_file_log($global_config_p, "Connect and login into the ftp server [ $ftp_server:$ftp_server_port ] ; FTP Proxy [  $ftp_proxy ] ; Passive mode [ $ftp_passive_mode ] ; FTP Timeout [ $ftp_timeout ]", $log_body_p); # write the error message into the script error file
        if ( $ftp_proxy ne "" ) { # if we have a firewall / proxy server
            $$ftp_connection_p = Net::FTP->new("$ftp_server", Debug => $ftp_debug_level_real, Port => $ftp_server_port, Passive => $ftp_passive_mode_real, Firewall => $ftp_proxy, Timeout => $ftp_timeout );
        }
        else { # if we don't have a firewall / proxy server
            $$ftp_connection_p = Net::FTP->new("$ftp_server", Debug => $ftp_debug_level_real, Port => $ftp_server_port, Passive => $ftp_passive_mode_real, Timeout => $ftp_timeout );
        }
        if ( defined $$ftp_connection_p ) { # test if the ftp connection worked
            $connect_try = $ftp_connect_retries + 1; # ftp conection worked, max the connect tries to exit the ftp connection retries
        }
    }
    if ( ! defined $$ftp_connection_p ) { # test if the ftp connection worked
        write_file_log($global_config_p, "Error - unable connect to the ftp server", $log_body_p); # write the error message into the script error file
        $error_flag = 1; # return failure
    }

    if ( $error_flag == 0 ) { # if managed to connect, attempt to login
        if ( !$$ftp_connection_p->login("$ftp_username", "$ftp_password") ) { # login into ftp server
            write_file_log($global_config_p, "Error - login into the ftp server [ username: $ftp_username ] ; [ password: ***** ]", $log_body_p); # write the error message into the script error file
            $error_flag = 1; # mark error
        }
    }

    return $error_flag; # return result
}


#
# Will return the current date, formated for mail MIME, example : "Wednesday, 5 May, 2004 09:34:37 -0000"
# Will return a date
# Explain: This date is in special format to be send via e-mail and to be understud by e-mail clients
sub get_mail_date {

    # Define the day of the week and the month of the year
    my @days   = ('Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday');
    my @months = ('January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December');

    # Get the current time and format the hour, minutes and seconds.
    my ($sec,$min,$hour,$mday,$mon,$year,$wday) = (localtime(time))[0,1,2,3,4,5,6];
    my $time = sprintf("%02d:%02d:%02d",$hour,$min,$sec);
    $year += 1900; # Add 1900 to the year to get the full 4 digit year

    # Format the date, in this kind "Wednesday, 5 May, 2004 09:34:37 -0000"
    my $date = "$days[$wday], $mday $months[$mon], $year $time -0000";

    return $date; # return the date

}


#
# Send a email (if configured)
# Will return 0 ; 1 ; 2 ; 3 ; 4
# Explain: 0 is sucess ; 1 2 3 4 are failures
sub send_email {

    my $my_message_subject = $_[0]; # e-mail message subject
    my $my_message_body = $_[1]; # e-mail message body, the actual email text
    my $x_mailer = $_[2]; # name and version of the program that send the e-mail
    my $global_config_p = $_[3]; # pointer into the configuration

    my $configuration_file_name = $$global_config_p{configuration_file_name};
    my $error_flag = 0; # to mark errors
    my $mail_to_addresses_valid = ""; # this will contain the mail addresses that were accepted by the mail server
    # mail configuration stuff
    my $mail_type = $$global_config_p{mail_type};
    my $mail_server = $$global_config_p{mail_server};
    my $mail_port = $$global_config_p{mail_port};
    my $mail_timeout = $$global_config_p{mail_timeout};
    my $mail_from_address = $$global_config_p{mail_from_address};
    my $mail_to_addresses = $$global_config_p{mail_to_addresses};
    my $mail_username = $$global_config_p{mail_username};
    my $mail_password = $$global_config_p{mail_password};
    my $mail_debug_level_real = $$global_config_p{mail_debug_level_real}; # debug level

    if ( $mail_type eq "smtp" ) {
        my $smtp_connection = ""; # this will countain the SMTP object that represents the connection into the mail server
        # connect into the smtp mail server
        $smtp_connection = Net::SMTP->new("$mail_server", Port => "$mail_port", Timeout => "$mail_timeout", Debug => "$mail_debug_level_real" );
        if ( ! defined $smtp_connection ) { # check if the SMTP connection is up
            $error_flag = 1; # return failure, fail to connect
        }
        elsif ( $mail_username ne "" ) { # If mail requires authentication 
            if ( !$smtp_connection->auth("$mail_username", "$mail_password") ) {
                    $smtp_connection->quit; # exit the smtp mail server
                    $error_flag = 2; # failed to authenticate
            }
        }
        if ( $error_flag == 0 ) { # manage to connect, attempt to send e-mail messages
            if ( !$smtp_connection->mail($mail_from_address) ) { # pass the from mail address
                $error_flag = 3; # return failure, invalid from e-mail address
            }
            else { # 
                foreach ( split(',', $mail_to_addresses) ) { # send each e-mail adresss
                    if ( $smtp_connection->recipient("$_", {SkipBad => 1}) ) { # test the e-mail address
                        if ( $mail_to_addresses_valid ne "" ) { # more than one e-mail recipient
                            $mail_to_addresses_valid = "$mail_to_addresses_valid, $_";
                        }
                        else { # first e-mail address
                            $mail_to_addresses_valid = "$_";
                        }
                    }
                }
                if ( $mail_to_addresses_valid ne "" ) { # if there is at least one valid email address in the "to" mail section
                    # Prepare the full mail message, including meta data (data, mail program, from, to, subject and body)
                    my $mail_data = "MIME-Version: 1.0" . "\n"
                         . "From: $mail_from_address" . "\n"
                         . "To: " . ( ref($mail_to_addresses_valid) ? join(';', @$mail_to_addresses_valid) : $mail_to_addresses_valid ) . "\n"
                         . "Date: " . get_mail_date() . "\n"
                         . "X-Mailer: " . $x_mailer . "\n"
                         . "Content-Type: text/plain; charset=iso-8859-1" . "\n"
                         . "Subject: " . $my_message_subject . "\n\n"
                         . "Beginning of e-mail report" . "\n\n"
                         . $my_message_body
                         . "End of e-mail report" . "\n\n";
                    if ( !$smtp_connection->data($mail_data) ) { # send the message
                        $error_flag = 5; # return failure, failed to send e-mail message
                    }
                }
                else {
                    $error_flag = 4; # return failure, there was no valid "to" mail address
                }
            }
            $smtp_connection->quit; # exit the smtp mail server
        }
    }

   # write apropriate messages into log file
    switch: {
        $error_flag == 1 && do {
                            write_file_log($global_config_p, "Warning - unable connect to the $mail_type mail server named [ $mail_server ]"); # write the error message into the script error file
                        };
        $error_flag == 2 && do {
                            write_file_log($global_config_p, "Warning - failure on mail authentication [ username: $mail_username ; password: ***** ], make shore that both the mail username and password are correct and that the mail server requires authentication"); # write the error message into the script error file
                        };
        $error_flag == 3 && do {
                            write_file_log($global_config_p, "Warning - the \"from\" e-mail address [ $mail_from_address ] is invalid"); # write the error message into the script error file
                        };
        $error_flag == 4 && do {
                            write_file_log($global_config_p, "Warning - there was no valid \"to\" e-mail address, no report was send"); # write the error message into the script error file
                        };
        $error_flag == 5 && do {
                            write_file_log($global_config_p, "Warning - failed to send the e-mail report, the connection with the mail server might have dropped"); # write the error message into the script error file
                        };
    }

    return $error_flag; # return the result

}


#
# Check if a specific file, directory or link are suppose to backup, this check is done
# by comparing the modification time with the previous list (if any is available of
# course).
# If the modification time is different or if it doesn't even exists and if the file
# extension isn't on the reject list then it's approved, else it's not.
# Will return: -1 ; 0 ; any positive number
# Explain: 0 is no don't backup ; any positive number is yes ; -1 if there was any error
sub check_add_to_backup {

    my $filedirlink_path = $_[0];
    my $backup_filedir_crc32 = $_[1]; # the crc32 number that defines the backup data path (directory or file)
    my $my_previous_directory = $_[2]; # the directory relative to where the backup starts
    my $dirfilelink_remain = $_[3]; # the remain of the path... filedirlink_path = backup_filedir_crc32 + dirfilelink_remain
    my $previous_file_list_p = $_[4]; # pointer into the hash that defines the list of the previous backup file list
    my $backup_data_type = $_[5]; # define the backup data type (directory or file)
    my $global_config_p = $_[6]; # pointer into the configuration 
    my $error_flag = ""; # mark error or status, by default it won't do backup

    if ( $backup_data_type ne "file" ) { # if user is doing a backup of a specficic file, then advanced file rejections are not used
        # check if the file type is not on the extextion rejection list
        if ( $$global_config_p{backup_extension_rejections} ne "" ) {
            my @my_split = split( /\./, $filedirlink_path); # split the file, in . parts
            my $size_value = scalar(@my_split) - 1 ; # read the array size, note array starts at zero size is always bigger
            if ( $size_value > 0 ) { # check if the file has extension
                my $file_ext = $my_split[$size_value]; # file or directory name
                @my_split = (); # clear the array
                @my_split = split(/,/,  $$global_config_p{backup_extension_rejections});
                foreach my $backup_rejection ( @my_split ) {
                    if ( $backup_rejection eq $file_ext ) { # if the file extension is on the reject list
                        $error_flag = 0; # don't do backup
                        last; # break and exit
                    }
                }
            }
        }

        # check if the file type is not on the filename rejection list, and if not already rejected
        if ( $error_flag eq "" and $$global_config_p{backup_filenames_rejections} ne "" ) {
            my @my_split = split(/,/,  $$global_config_p{backup_filenames_rejections});
            my $filedir_name = get_filedir_name("$filedirlink_path", $global_config_p);
            foreach my $backup_rejection_filename ( @my_split ) {
                if ( $backup_rejection_filename eq $filedir_name ) { # if the file extension is on the reject list
                    $error_flag = 0; # don't do backup
                    last; # break and exit
                }
            }
        }

        # check if the file size is in the backup size limit (in bytes), and if not already rejected
        if ( $error_flag eq "" and $$global_config_p{backup_file_size_limit} != 0 ) {
            # get the present file size (in bytes)
            my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$present_mtime,$ctime,$blksize,$blocks) = stat("$filedirlink_path");
            # check if the real file size is bigger that the configured limit
            if ( $size > $$global_config_p{backup_file_size_limit_bytes} ) {
                $error_flag = 0; # don't do backup
            }
        }
    }
    # check the previous file modification date/time with the present one
    if ( $error_flag eq "" and $$global_config_p{backup_mode} ne "full" ) { # only if not previous rejected or that no error occured
        my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$present_mtime,$ctime,$blksize,$blocks);
        if ( !( -l "$filedirlink_path") ) { # if it's a file or directory
            # get the present file modification time
            ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$present_mtime,$ctime,$blksize,$blocks) = stat("$filedirlink_path");
        }
        else { # it's a unix link
            # get the present file modification time
            ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$present_mtime,$ctime,$blksize,$blocks) = lstat("$filedirlink_path");
        }
        if ( defined $present_mtime ) { # if able to get the file modification time
            $error_flag = 0; # don't do backup
            # build the present backup management line and match it against the backup management file to see if anything changed
            my $rebuild_line = "$backup_filedir_crc32<$my_previous_directory<$dirfilelink_remain<$present_mtime";
            # if its a file and the line is not register on the management file (with the present modification data/time)
            if ( !(-d "$filedirlink_path") and !(exists $$previous_file_list_p{$rebuild_line}) ) {
                $error_flag = $present_mtime; # do the backup and update the file modification date/time
            }
        }
    }

    if ( $error_flag eq "" and $$global_config_p{backup_mode} eq "full" ) {
        $error_flag = 100; # since we are in full backup mode, this is not ever written in a file
    }

    if ( $error_flag eq "" ) { # detect possible bugs
        $error_flag = -1; # if $error_flag is still empty a bug occured
    }

    return $error_flag; # return the result

}


#
# Try to open a rejection file (always named simplebackup_rejection.txt), open it
# read the order and return it.
# If any error occures (problem reading the file or invalid order) it will return a
# string indicatind error.
# Will return: all ; files ; sub-directories ; error
# Explain: all             -> is order reject all
#          files           -> is order reject files but accept sub-directories
#          sub-directories -> is order reject sub-directories but accept files
#          touch           -> is not a order, it's just a marker to say to simplebackup
#                             backup this dir even if only the simplebackup_rejection.txt
#                             file exists... this is a "fix" to a stupid tar "feature",
#                             that prevents doing a backup in a empty directory.
#          error           -> indicates that there was a error reading the file
sub test_read_rejection_file {

    my $rejection_file = $_[0]; # full path into the simplebackup_rejection.txt file
    my $error_flag = 0; # mark errors
    my $return_value = "";
    my $counter = 0;

    # open file in read mode ( < )
    open(rejection_file_handle, "<$rejection_file") or $error_flag = 1;
    if ( $error_flag == 0 ) {
        # read the entire config file line by line
        while(<rejection_file_handle>) {
            $counter++; # this counter is used to prevent the file from having more than one line
            chomp($_); # remove \n (new line)
            $return_value = $_; # read the file
            last if $counter > 1; # only read one line, more than that and the file has a error or is of the wrong type
        }
        close(rejection_file_handle); # close the file
    }
    # test if any error occured
    if ( $error_flag != 0 or $counter > 1 or ($return_value ne "all" and $return_value ne "files" and $return_value ne "sub-directories" and $return_value ne "touch") ) {
        $return_value = "error"; # mark error
    }

    return $return_value; # return string

}


#
# Build the list of files that are suppose to be backup
# Will return: 0 ; 1
# Explain: 0 is success ; 1 is failure
sub build_present_backup_list {

    my $filedir_to_check = $_[0]; # file or directory path to check
    my $input_backup_crc32 = $_[1]; # the crc32 of the file or directory to backup
    my $my_previous_directory = $_[2]; # start backup directory... example if backup /home/miguel the start directory is /home
    my $my_backup_file_directoryname = $_[3]; # directory name to be backup
    my $previous_file_list_p = $_[4]; # pointer into the array that defines the list of the previous backup file list
    my $new_file_list_p = $_[5]; # pointer into the array that defines the new list backup file list
    my $file_separator = $_[6]; # file separator / for unix \ for windows
    my $backup_data_type = $_[7]; # if we are building a backup list of a directory (and subdirectories + files) or from a single file
    my $error_flag = $_[8]; # mark the error log 
    my $global_config_p = $_[9]; # pointer into the configuration
    my $log_body_p = $_[10]; # pointer into the e-mail body string

    my $rejection_file_order = ""; # this will contain the orders from a available simplebackup_rejection.txt file
    my $my_previous_directory_size = length($my_previous_directory); # will contain the string size of the start backup directoryp

    no warnings 'recursion'; # disable possible recursion warning messages from perl, this could append if you got a very deep directory list
    if ( $error_flag == 0 ) { # if there was any previous error the function will ABORT, backup will not sucessed
        my @filedir_list;
        if ( $backup_data_type eq "directory" ) { # it's a directory
            opendir(dir_handle, "$filedir_to_check") or $error_flag = 1; # open the load directory
            @filedir_list = readdir dir_handle; # read the file list
            closedir(dir_handle);# close the directory handle
            write_file_log($global_config_p, "Warning - unable to open the directory [ $filedir_to_check ]", $log_body_p) if ( $error_flag != 0 ); # write the error message into the script error file if unable to open the directory
            return 0 if $error_flag == 1; # error exit imediatly
        }
        else { # it's a file
            push (@filedir_list, "."); # emulate a directory entry
            push (@filedir_list, ".."); # emulate a directory entry
            push (@filedir_list, "$my_backup_file_directoryname"); # place the name of the file to check
        }
        if ( $error_flag == 0 ) {
            $filedir_to_check =~ s/\\+$//;  # delete any possible ending \ char
            $filedir_to_check =~ s/\/+$//;  # delete any possible ending / char
            if ( scalar (@filedir_list) == 2 and $$global_config_p{backup_format} ne "tar" and $$global_config_p{backup_format} ne "tar.Z" and $$global_config_p{backup_format} ne "tar.gz" and $$global_config_p{backup_format} ne "tar.bz2"   ) { # if the directory is empty, check if it's supose to be added into the current backup, notice in any tar kind of backups empty directories are ignore
                # break the full directory into the apropriate fields: /input_directory*directory_to_backup*dir
                my $remain_size = length($filedir_to_check);
                my $dirfilelink_remain = substr($filedir_to_check,$my_previous_directory_size,$remain_size);
                $dirfilelink_remain =~ s/^\\//;  # delete any possible leading \ char
                $dirfilelink_remain =~ s/^\///;  # delete any possible leading / char
                my $mtime = check_add_to_backup("$filedir_to_check", "$input_backup_crc32", "$my_previous_directory", "$dirfilelink_remain", $previous_file_list_p, $backup_data_type, $global_config_p ); # check if the file is suppost to be added into the backup
                if ( $mtime > 0 ) {
                    push (@$new_file_list_p, "$input_backup_crc32<$my_previous_directory<$dirfilelink_remain<dir"); # keep the empty directory for backup
                }
                elsif ( $mtime < 0 ) { # error getting the file/link modification time, notice mtime = 0 means that the file is not supose to be included in the backup
                    write_file_log($global_config_p, "Warning - unable to access the empty directory [ $filedir_to_check ]", $log_body_p); # write the error message into the script error file
                }
                return $error_flag; # exit imediatly
            }
            # first look for simplebackup_rejection.txt files if required by user
            if ( $$global_config_p{backup_directory_rejections} eq "yes" ) {
                foreach my $filedir_name ( @filedir_list ) { # check the entiry directory list
                    if ( $filedir_name eq "simplebackup_rejection.txt" ) {
                        $rejection_file_order = test_read_rejection_file("$filedir_to_check$file_separator$filedir_name");
                        if ( $rejection_file_order eq "error" ) { # if there was any error reading the file
                            write_file_log($global_config_p, "Warning - rejection file [ $filedir_to_check$file_separator$filedir_name ] is corrupted and his orders are being ignored", $log_body_p); # write the error message into the script error file
                        }
                    }
                }
            }
            if ( $rejection_file_order ne "all" ) { # continue if directory is not to be ignored (order by simplebackup_rejection.txt file)
                foreach my $filedir_name ( @filedir_list ) { # check the entiry directory list
                    if ( $filedir_name ne "." and $filedir_name ne ".." ) { # ignore the system file names . and ..
                        if ( -d "$filedir_to_check$file_separator$filedir_name" ) { # if it's a directory, go all over again
                            if ( $rejection_file_order ne "sub-directories" ) { # if sub-directories are not to be ignored (order by simplebackup_rejection.txt file)
                                $error_flag = build_present_backup_list("$filedir_to_check$file_separator$filedir_name", "$input_backup_crc32", "$my_previous_directory", $my_backup_file_directoryname, $previous_file_list_p, $new_file_list_p, $file_separator, $backup_data_type, $error_flag, $global_config_p, $log_body_p );
                            }
                        }
                        elsif ( $rejection_file_order ne "files" ) { # if it's a file or unix link and if files are not to be ignored (order by simplebackup_rejection.txt file)
                            my $filelink_full_path = "";
                            my $dirfilelink_remain = "";
                            # break the full directory into the apropriate fields: /input_directory*directory_to_backup*dir
                            if ( $backup_data_type ne "file" ) { # type is a directory
                                $filelink_full_path = "$filedir_to_check$file_separator$filedir_name";
                                my $remain_size = length($filelink_full_path);
                                $dirfilelink_remain = substr("$filelink_full_path",$my_previous_directory_size,$remain_size);
                            }
                            else { # we are doing a backup of a single file
                                $filelink_full_path = "$filedir_to_check";
                                $dirfilelink_remain = "$filedir_name";
                            }
                            $dirfilelink_remain =~ s/^\\//;  # delete any possible leading \ char
                            $dirfilelink_remain =~ s/^\///;  # delete any possible leading / char
                            my $mtime = check_add_to_backup("$filelink_full_path", "$input_backup_crc32", "$my_previous_directory", "$dirfilelink_remain", $previous_file_list_p, $backup_data_type, $global_config_p ); # check if the file is suppost to be added into the backup
                            if ( $mtime > 0 ) {
                                push (@$new_file_list_p, "$input_backup_crc32<$my_previous_directory<$dirfilelink_remain<$mtime"); # keep the empty directory for backup
                            }
                            elsif ( $mtime < 0 ) { # error getting the file/link modification time, notice mtime = 0 means that the file is not supose to be included in the backup
                                write_file_log($global_config_p, "Warning - unable to access the modification date/time of [ $filedir_to_check$file_separator$filedir_name ]", $log_body_p); # write the error message into the script error file
                            }
                        }
                        last if $error_flag != 0; # break cicle imediatly if any error was detected
                    }
                }
            }
        }
    }

    return $error_flag;
}


#
# Get the start of a full path plus the remain of the full path and return the full path
# Will return: the full path of something
sub rebuild_full_path {

    my $start_of_path = $_[0]; # start of path, example /home   or d:\
    my $remain_of_path = $_[1]; # remain of path  example: miguel
    my $global_config_p = $_[2]; # pointer into the configuration
    my $return_string = ""; # what is returned   example: /home/miguel  or d:\miguel

    my $file_separator = $$global_config_p{host_file_separator}; # file separator

    # deal with Operating System dependend stuff
    if ( $$global_config_p{host_os} eq "unix" and $start_of_path eq "/" ) { # unix type paths and unix path is /
        $start_of_path = "";
    }
    elsif ( $$global_config_p{host_os} ne "unix" ) { # microsoft type paths
        $start_of_path =~ s/\\$//;  # Delete trailing \ char
    }

    $return_string = "$start_of_path$file_separator$remain_of_path"; # rebuild the full path

    return $return_string;

}


#
# Read/Write the backup list of files into or from the backup management file or write a temporary file to pass
# into the apropriate archive commands.
# Notice partially read/writes file are not allowed, or it correctly reads/writes the full file or it doesn't.
# Will return: 0 ; 1 ; 2
# Explain: 0 is success ; 1 is failure ; 2 is failure (invalid operation, must be read, write or write_tmp)
sub read_write_management_file {

    my $previous_file_list_p = $_[0]; # pointer into the hash that defines the passed list backup file list
    my $present_file_list_p = $_[1]; # pointer into the array that defines the new list backup file list
    my $operation = $_[2]; # the operation, read or write ?
    my $backup_file_directoryname = $_[3]; # directory of backup
    my $tmp_management_file = $_[4]; # path into the temporary management file, only used in the write_tmp operation
    my $global_config_p = $_[5]; # pointer into the configuration
    my $log_body_p = $_[6]; # pointer into the e-mail body string
    my $my_file_separator = $$global_config_p{host_file_separator}; # file separator
    my $error_flag = 2; # to mark errors

    if ( $operation eq "read" ) { # read operation, read into the previous management file list
        $error_flag = 0;
        open(management_file_handle, "<$$global_config_p{backup_db_management_file}") or $error_flag = 1;
        if ( $error_flag == 0 ) {
            my $line_counter = 0; # counter the line number
            my $last_line = ""; # to keep a copy of the last read line
            while(<management_file_handle>) {
                $line_counter++; # add a line number
                chomp($_); # remove the new line chars
                if ( $line_counter == 1 and $_ ne "<-- start of management file -->" ) { # search if first line marks the start of file
                    $error_flag = 1; # mark error missing start of file line
                }
                # don't load into ram the start of file and end of file lines
                if ( $line_counter > 1 and $_ ne "<-- end of management file -->" ) {
                    # now check if the line is correct
                    my @my_split = split( /</, $_);
                    # check if there is ONLY one : char and if there is something like /asdas/sds:dir or /sdaa/sads:828271 or D:\asd\ss:dir  or d:\asd\ss:877661
                    if ( scalar(@my_split) == 4 and ($my_split[3] eq "dir" or ($my_split[3] =~ /^\d+$/ and $my_split[3] ne 0) ) ) {
                        $$previous_file_list_p{"$_"} = (); # load the line
                    }
                    else { # corrupted line
                        $error_flag = 1; # mark error
                    }
                }
                if ( $last_line eq "<-- end of management file -->" ) { # detect dupplicated end of file marker
                    $error_flag = 1; # mark error
                }
                $last_line = $_; # backup the read line
                last if $error_flag != 0; # break cicle if any error occured
            }
            close (management_file_handle); # close the file
            if ( $last_line ne "<-- end of management file -->" ) { # search if first line marks the start of file
                $error_flag = 1; # mark error missing end of file line
            }
        }
    }
    elsif ( $operation eq "write" ) { # write operation using the previous file list and the present file list (previously read or build)
        my @new_list; # this array the list that will be written into the management file, will contain the already
                      # backup files/directories (from the previous_backup_list) and the files/directories from the new
                      # list
        $error_flag = 0; # by defauld no error was found
        if ( (keys %$previous_file_list_p) == 0 ) {
            @new_list = @$present_file_list_p; # the previous list is empty so it's only required to add the new files
        }
        else { # build a new list based on the previous file/dir list and the new files/directories added
            # compare the previous list with the new and update files/dirs that are present on the new and add the
            # files that are only on the older
            my $add_file;
            my $some_previous_filedir; # will contain the relative path of files/directories on the previous file list
            my $some_previous_cddir; # will start of the full path into the previous file list
            my $some_present_filedir; # will contain the relative path of files/directories on the present file list
            my $some_present_cddir; # will start of the full path into the present file list
            my $trash1; # will contain the stat data that i don't care about
            my $trash2; # will contain the stat data that i don't care about
            #foreach my $some_previous_filedir_string ( @$previous_file_list_p ) { # read the previous list
            for my $some_previous_filedir_string ( keys %$previous_file_list_p ) { # read the previous list
                $add_file = "";
                ( $trash1, $some_previous_cddir, $some_previous_filedir, $trash2 ) = split(/</, $some_previous_filedir_string); # split the file/dir path and the stat data
                # rebuild the full path
                $some_previous_filedir = rebuild_full_path("$some_previous_cddir", "$some_previous_filedir", $global_config_p);
                foreach my $some_present_filedir_string ( @$present_file_list_p ) {
                    ( $trash1, $some_present_cddir, $some_present_filedir, $trash2 ) = split(/</, $some_present_filedir_string); # split the previous dir, file/dir path and the stat data
                    $some_present_filedir = rebuild_full_path("$some_present_cddir", "$some_present_filedir", $global_config_p); # rebuild the full path
                    if ( $some_present_filedir eq $some_previous_filedir ) {
                        $add_file = "$some_present_filedir_string"; # order for the new element to be added ( file/dir name + the stat date
                    }
                }
                $add_file = $some_previous_filedir_string if $add_file eq ""; # if file/dir wasn't found on the new backup list (not backuped)
                push @new_list, $add_file; # add the new file/dir into the management file list
            }
            # now search for new files/dirs on the present list that are not on the old list (meaning new files/dirs)
            foreach my $some_present_filedir_string ( @$present_file_list_p ) { # read the previous list
                $add_file = "yes"; # by default the file is added
                ( $trash1, $some_present_cddir, $some_present_filedir, $trash2 ) = split(/</, $some_present_filedir_string); # split the previous dir, file/dir path and the stat data
                # rebuild the full path
                $some_present_filedir = rebuild_full_path("$some_present_cddir", "$some_present_filedir", $global_config_p);
                for my $some_previous_filedir_string ( keys %$previous_file_list_p ) { # read the previous list
                    ( $trash1, $some_previous_cddir, $some_previous_filedir, $trash2 ) = split(/</, $some_previous_filedir_string); # split the file/dir path and the stat data
                    $some_previous_filedir = rebuild_full_path("$some_previous_cddir", "$some_previous_filedir", $global_config_p); # rebuild the full path 
                    if ( $some_present_filedir eq $some_previous_filedir ) {
                        $add_file = "no"; # found the file/dir so don't add it
                    }
                }
                if ( $add_file eq "yes" ) { # if file wasn't found on the new list
                    push @new_list, $some_present_filedir_string; # add the new file/dir into the management file list
                }
            }
        }
        @new_list = sort @new_list; # sort the list to help performance in future backups (using this list)
        open(management_file_handle, ">$$global_config_p{backup_db_management_file}") or $error_flag = 1;
        if ( $error_flag == 0 ) {
            #lock the file for exclusive script use
            flock (management_file_handle, LOCK_EX) or $error_flag = 1;
            print management_file_handle "<-- start of management file -->\n", or $error_flag = 1;
            foreach my $new_line ( @new_list ) {
                # write the log message
                print management_file_handle "$new_line\n", or $error_flag = 1;
                last if $error_flag != 0; # break imediatly if any error occured
            }
            if ( $error_flag == 0 ) { # print the close file line
                print management_file_handle "<-- end of management file -->", or $error_flag = 1;
            }
            close (management_file_handle); # close the file
        }
    }
    elsif ( $operation eq "write_tmp" ) { # write temporary file, used to pass into the archive commands
        $error_flag = 0; # valid option
        open(management_tmp_file_handle, ">$tmp_management_file") or $error_flag = 1;
        if ( $error_flag == 0 ) {
            #lock the file for exclusive script use
            flock (management_tmp_file_handle, LOCK_EX) or $error_flag = 1;
            foreach my $tmp_line ( @$present_file_list_p ) { # read the present backup list
                last if $error_flag != 0; # break imediatly if any error occured
                my ($backup_name, $start_dir, $file_dir, $file_type) = split ( /</, $tmp_line ); # get only the file name
                print management_tmp_file_handle "$file_dir\n", or $error_flag = 1;
            }
            close(management_tmp_file_handle); # close the file
            if ( $error_flag != 0 ) { # if there was a error or but the file contains no data, erase it
                unlink "$tmp_management_file"; # erase the temporary file, since it's empty or damaged
            }
        }
    }

    if ( $error_flag == 1 ) { # deal with errors
        my $my_file = "";
        if ( $operation eq "read" ) {
            $my_file = $$global_config_p{backup_db_management_file};
            @$previous_file_list_p = (); # clear the previous backup list
            write_file_log($global_config_p, "Error - unable to read the management file [ $my_file ]", $log_body_p); # write the error message into the script error file
        }
        elsif ( $operation eq "write" ) {
            $my_file = $$global_config_p{backup_db_management_file};
            write_file_log($global_config_p, "Error - unable to write into the management file [ $my_file ]", $log_body_p); # write the error message into the script error file
        }
        else {
            write_file_log($global_config_p, "Error - unable to write into the temporary backup list file [ $tmp_management_file ]", $log_body_p); # write the error message into the script error file
        }
    }

    return $error_flag;

}


#
# The passes the backup log into a external command using stdin, this command is
# configured by the user on the configuration file
# Will return: 0 ; 1
# Explain: 0 is success ; 1 is failure
sub export_log {

    my $log = $_[0]; # e-mail message body, the actual email text
    my $global_config_p = $_[1]; # pointer into the configuration
    my $pass_log_to_external_cmd = $$global_config_p{pass_log_to_external_cmd};
    my $error_flag = 0;

    open(command_handle, "| $pass_log_to_external_cmd") or $error_flag = 1;
    if ( $error_flag == 0 ) {
        print command_handle $log or $error_flag = 1; # export the log file into the external command
        close(command_handle); # close the command file handle
    }

    return $error_flag; # return the result
}


#
# The actual function that compresses the data into a file
# Will return: -1 ; any_other_positive_integer_numeric_value
# Explain: -1 is failure ; any other integer number number that represents the compressed file size
sub do_compress {

    my $archive_file_name = $_[0]; # compressed file name
    my $archive_file_tar_name = $_[1]; # temporary name for used in tar differential/incremental backups
    my $backup_file_directoryname = $_[2]; # file or directory to compress
    my $backup_present_file_list_p = $_[3]; # pointer into the backup list, if this is available the backup will be made depending on this
    my $previous_directory = $_[4]; # directory before the backup dir
    my $global_config_p = $_[5]; # pointer into the configuration
    my $log_body_p = $_[6]; # pointer into the e-mail body string
    my $compression_ratio = $$global_config_p{backup_real_ratio}; # the compression ratio
    my $null_device = $$global_config_p{host_null_device}; # operating system null device
    my $error_flag = 0; # mark errors
    my @compress_command;

    if ( ! chdir "$previous_directory" ) { # attempt to change into the previous directory, this is done to build a backup without a full path into the directory, example...   /home/miguel/documents will compress a file with just the directory documents and not the full path
        write_file_log($global_config_p, "Error - unable access the directory [ $previous_directory ] for backup, check permissions", $log_body_p); # write the error message into the script error file
        return -1; # return failure
    }

    # erase temporary files could be left over by some simplebackup previous crash or a user kill operation
    unlink "$archive_file_name" if -f "$archive_file_name"; # erase the temporary file could be left over by some user
    unlink "$archive_file_tar_name" if ( -f "$archive_file_tar_name" ); # erase the temporary file could be left over by some user

    if ( scalar(@$backup_present_file_list_p) == 0 ) { # if there is no backup list defined, then it's a standard full backup with everything
        switch: { # prepare backup command
            $$global_config_p{backup_format} eq "rar" && do { # rar type of files
                            @compress_command = ("rar", "a", "-inul", "-isnd", "-m$compression_ratio", "-md4096", "-r", "-s", "-dh", "-rr", "-ow", "\"$archive_file_name\"", "\"$backup_file_directoryname\"", "$null_device");                                      
                        };
            $$global_config_p{backup_format} eq "zip" && do { # tar type of files
                            # deal with Operating System dependend stuff and build the full bath into the backup file
                            if ( $$global_config_p{host_os} eq "unix" ) { # unix operating system (without -S parameter)
                                @compress_command = ("zip", "-r", "-q", "-$compression_ratio", "\"$archive_file_name\"", "\"$backup_file_directoryname\"", "$null_device");
                            }
                            else {# windows operating system (with -S parameter, which mean include hidden files)
                                @compress_command = ("zip", "-r", "-q", "-$compression_ratio", "-S", "\"$archive_file_name\"", "\"$backup_file_directoryname\"", "$null_device");
                            }
                        };
            $$global_config_p{backup_format} eq "tar" && do { # tar type of files
                            @compress_command = ("tar", "-cpf", "\"$archive_file_name\"", "\"$backup_file_directoryname\"", "$null_device");
                        };
            $$global_config_p{backup_format} eq "tar.Z" && do { # tar + compress type of files
                            @compress_command = ("tar", "-cpf", "-", "\"$backup_file_directoryname\"", "|", "compress", ">", "\"$archive_file_name\"", "$null_device");
                        };
            $$global_config_p{backup_format} eq "tar.gz" && do { # tar + gzip type of files
                            @compress_command = ("tar", "-cpf", "-", "\"$backup_file_directoryname\"", "|", "gzip", "-$compression_ratio", ">", "\"$archive_file_name\"", "$null_device");
                        };
            $$global_config_p{backup_format} eq "tar.bz2" && do { # tar + bzip2 type of files
                            @compress_command = ("tar", "-cpf", "-", "\"$backup_file_directoryname\"", "|", "bzip2", "-$compression_ratio", ">", "\"$archive_file_name\"", "$null_device");
                        };
        }
        write_file_log($global_config_p, "Backup into temporary file using command [ @compress_command ]", $log_body_p); # write the error message into the script error file
        system("@compress_command") == 0 or $error_flag = -1; # do backup and if we return 1 something failed
    }
    else { # backup using a backup list

        my @not_relevant;
        my $tmp_management_file = get_filedir_name($$global_config_p{configuration_file_name}, $global_config_p);
        $tmp_management_file = "$tmp_management_file" . ".tmp"; # add the tmp extension
        $tmp_management_file = rebuild_full_path("$$global_config_p{temporary_dir}", "$tmp_management_file", $global_config_p); # rebuild the full path pointing into the path defined by temporary_dir
        # write a temporary file list, that is actualy passed into the archive command
        if ( read_write_management_file(\@not_relevant, $backup_present_file_list_p, "write_tmp", $backup_file_directoryname, $tmp_management_file, $global_config_p, $log_body_p ) != 0 ) {
            $error_flag = -1; # mark error
        }
        if ( $error_flag == 0 ) {
            switch: { # prepare backup command
                $$global_config_p{backup_format} eq "rar" && do { # rar type of files
                                # deal with Operating System dependend stuff and build the full bath into the backup file
                                if ( $$global_config_p{host_os} eq "unix" ) { # unix operating system
                                    @compress_command = ("cat", "\"$tmp_management_file\"", "|", "rar", "a", "-inul", "-isnd", "-m$compression_ratio", "-md4096", "-r", "-s", "-dh", "-rr", "-ow", "\"$archive_file_name\"", "@", "$null_device");
                                }
                                else {# windows operating system
                                    @compress_command = ("type", "\"$tmp_management_file\"", "|", "rar", "a", "-inul", "-isnd", "-m$compression_ratio", "-md4096", "-r", "-s", "-dh", "-rr", "-ow", "\"$archive_file_name\"", "@", "$null_device");
                                }
                            };
                            $$global_config_p{backup_format} eq "zip" && do { # tar type of files
                                # deal with Operating System dependend stuff and build the full bath into the backup file
                                if ( $$global_config_p{host_os} eq "unix" ) { # unix operating system (without -S parameter)
                                    @compress_command = ("cat", "\"$tmp_management_file\"", "|", "zip", "-r", "-q", "-$compression_ratio", "-@", "\"$archive_file_name\"", "$null_device");
                                }
                                else {# windows operating system (with -S parameter, which mean include hidden files)
                                    @compress_command = ("type", "\"$tmp_management_file\"", "|", "zip", "-r", "-q", "-$compression_ratio", "-S", "-@", "\"$archive_file_name\"", "$null_device");
                                }
                            };
                            ( $$global_config_p{backup_format} eq "tar" or $$global_config_p{backup_format} eq "tar.Z" or $$global_config_p{backup_format} eq "tar.gz" or $$global_config_p{backup_format} eq "tar.bz2" )&& do { # tar + bzip2 type of files
                                write_file_log($global_config_p, "Warning - Your backup configuration does not allow empty directories to be backup", $log_body_p); # write the error message into the script error file
                                if ( $^O =~ "linux" or $^O =~ "Linux" or $^O =~ "LINUX" or $^O =~ "MSWin32" or $^O =~ "MSWIN32" ) { # deal with different tar command  gnu tar uses -T (linux and windows) other generaly use -I
                                    @compress_command = ("tar", "-cpf", "\"$archive_file_tar_name\"", "-T", "\"$tmp_management_file\"", "$null_device");
                                }
                                else {
                                    @compress_command = ("tar", "-cpf", "\"$archive_file_tar_name\"", "-I", "\"$tmp_management_file\"", "$null_device");
                                }
                            };
            }
            write_file_log($global_config_p, "Backup into temporary file using a command [ @compress_command ]", $log_body_p); # write the error message into the script error file
            system("@compress_command") == 0 or $error_flag = -1; # do backup and if we return -1 something failed
            if ( $error_flag == 0 ) {
                switch: { # prepare compress command (for a tar backups )
                    $$global_config_p{backup_format} eq "tar.Z" && do { # compress type of files
                        @compress_command = ("compress", "\"$archive_file_tar_name\"", "$null_device");
                    };
                    $$global_config_p{backup_format} eq "tar.gz" && do { # gzip type of files
                        @compress_command = ("gzip", "-$compression_ratio", "\"$archive_file_tar_name\"", "$null_device");
                    };
                    $$global_config_p{backup_format} eq "tar.bz2" && do { # bzip2 type of files
                        @compress_command = ("bzip2", "-$compression_ratio", "\"$archive_file_tar_name\"", "$null_device");
                    };
                }
                if ( $$global_config_p{backup_format} eq "tar.Z" or $$global_config_p{backup_format} eq "tar.gz" or $$global_config_p{backup_format} eq "tar.bz2" ) {
                    write_file_log($global_config_p, "Compressing tar archive using a command [ @compress_command ]", $log_body_p); # write the error message into the script error file
                    system("@compress_command") == 0 or $error_flag = -1; # do backup and if we return 1 something failed
                }
            }
        }
        # clear left over files
        unlink "$archive_file_tar_name" if ( -f "$archive_file_tar_name" and $$global_config_p{backup_format} ne "tar" );
        unlink "$tmp_management_file" if -f "$tmp_management_file"; # erase the temporary management file
    }

    if ( $error_flag == 0 ) { # if no error occured during compression, get the file size
        my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$present_mtime,$ctime,$blksize,$blocks) = stat("$archive_file_name");
        $error_flag = $size; # return the backup file size
    }

    return $error_flag; # return the result or the file size

}


#
# Receive any string and return the corresponding CRC32 number
# The crc32 algorithm used is based in the Swissknife library (crc32.pm module)
# from Wolfgang Fleischmann wfl\@ebi.ac.uk
# Will return: a integer (crc32 number)
# Explain: The integer is a crc32 number
sub build_crc32 {

    my $the_string = $_[0]; # the string that we are going to convert into a CRC32 number
    $the_string=~tr/a-z/A-Z/;
    my @the_string = split //, $the_string;
    my @crc32_table=();

    # Build the Crc32 table
    my ($i,$j,$crc_table);
    my $CRC32_POLYNOMIAL = 0xEDB88320;
    for ($i=0; $i <= 255; $i++){
        $crc_table = $i;
        for ($j=8; $j>0; $j--){
            if ($crc_table & 1){
                $crc_table = (($crc_table >> 1) & 0x7FFFFFFF) ^ $CRC32_POLYNOMIAL;
            } 
            else {
                $crc_table = ($crc_table >> 1) & 0x7FFFFFFF;
            }
        }
        $crc32_table[$i]=$crc_table;
    }

    # Build the Crc32 number
    my $crc = 0xFFFFFFFF;
    my $temp1;
    my $temp2;
    while (my $c=shift @the_string){
	my $cc=ord($c);
	$temp1 = ($crc >> 8) & 0x00FFFFFF;
	$temp2 = $crc32_table[($crc ^ $cc) & 0xFF ];
	$crc = $temp1 ^ $temp2;
    }

    return $crc; # return the crc32 number
}


#
# Do the actual backup, calling what ever auxiliary function's required
# Will return: 0 ; 1
# Explain: 0 is sucess ; 1 failure
sub do_backup {

    my $script_version_int = $_[0]; # script version
    my $last_update_int = $_[1]; # script last update date
    my $global_config_p = $_[2]; # pointer into the configuration
    my $error_flag = 0; # to mark errors
    my $session_number = ""; # backup session name
    my $ftp_connection; # to keep the NET:FTP connection object
    my $log_body = ""; # will contain all LOG MESSAGES and will be used (if configured)
                       # to send the full backup log via e-mail
    my $machine_hostname = $$global_config_p{hostname}; # get the hostname
    my $backup_filename_final = ""; # will contain the final backup file name
    my $backup_filename_temp_fortar = ""; # for advanced backups using tar
    my $backup_filename_temp = ""; # will contain the temporary backup file name
    my $backup_filename_final_fpath = ""; # will contain the final backup file name and full path
    my $backup_filename_temp_fortar_fpath = ""; # for advanced backups using tar and full path
    my $backup_filename_temp_fpath = ""; # will contain the temporary backup file name and full path
    my %backup_previous_file_list; # keep passed the backup list for incremental and differental backups
    my @backup_present_file_list; # present backup list, is used in incremental, differencial and might be used in full backups
    my $backup_files_count = 0; # the number of backup files that where created in the present backup session
    my $backup_files_sumsize = 0; # the total size in bytes of the backup files that where created in the present backup session
    my $all_backup_file_size = 0; # to keep the size of all compressed backup file files (in bytes)
    my $my_file_separator = $$global_config_p{host_file_separator}; # file separator
    my $backup_mode = $$global_config_p{backup_mode}; # the backup mode used
    my $backup_real_mode = $$global_config_p{backup_real_mode}; # the real backup mode used
    my $backup_working = $$global_config_p{backup_working}; # how is the backup working, ftp, local file system, etc
    my $host_type = $$global_config_p{host_os}; # the configurated operating system.... unix...etc
    my $config_file = $$global_config_p{configuration_file}; # put it in order to easy put it inside a text string
    my $log_file = $$global_config_p{log_file}; # put it in order to easy put it inside a text string
    my @my_input_backup_split = split(/</,  $$global_config_p{input_backup}); # split each backup directorie in a array

    my @input_backup_done; # contains the list of the actual backups that where done
    # start logging
    write_file_log($global_config_p, "Starting Backup ( simplebackup.pl v$script_version_int - $last_update_int )",\$log_body); # write the message into the script error file
    write_file_log($global_config_p, "Hostname is [ $machine_hostname ]", \$log_body); # write the message into the script error file
    write_file_log($global_config_p, "Using configuration file [ $config_file ]", \$log_body); # write the message into the script error file
    write_file_log($global_config_p, "Using log file [ $log_file ]", \$log_body); # write the message into the script error file
    write_file_log($global_config_p, "The configured backup mode is [ $backup_mode ]", \$log_body); # write the message into the script error file
    write_file_log($global_config_p, "Today's backup mode is [ $backup_real_mode ]", \$log_body); # write the message into the script error file
    write_file_log($global_config_p, "Backup is working in [ $backup_working ]", \$log_body); # write the message into the script error file
    write_file_log($global_config_p, "Selected host operating system is [ $host_type ], detected is [ $^O ]", \$log_body); # write the message into the script error file
    # if we are using a database file
    if ( $$global_config_p{local_db_history_file} ne "" ) {
        my $local_db_history_file = $$global_config_p{local_db_history_file};
        write_file_log($global_config_p, "Backup sessions are managed using the history file [ $local_db_history_file ]", \$log_body); # write the message into the script error file
    }
    # if we are doing a incremental or differential backup
    if ( $$global_config_p{backup_mode} ne "full" ) {
        my $local_backup_db_management_file = $$global_config_p{backup_db_management_file};
        write_file_log($global_config_p, "Backup data is managed using the management file [ $local_backup_db_management_file ]", \$log_body); # write the message into the script error file
    }

    if ( $$global_config_p{backup_working} eq "ftp" ) { # connect into the ftp server
        $error_flag = ftp_connect_login(\$ftp_connection, $global_config_p, \$log_body);
    }
    if ( $error_flag == 0 ) { # only enter if if no error occured.... ftp connect error for example
        # detecting if we need to do a full backup, if the backup files (defined by input_backup) is out of sync with the backup list (defined by input_backup) or if there are any incremental/differencial backup  files under differencial/incremental backup mode (the reverse)
        my $present_backup_count = sum_backup_types(\@my_input_backup_split, $ftp_connection, $global_config_p, \$log_body);
        if ( defined $present_backup_count ) {
            my ($full_count, $incremental_count, $differencial_count, $all_count ) = split ( / /, $present_backup_count); # split the counts into the apropriate fields
            if ( $$global_config_p{backup_mode} ne "full" and $$global_config_p{backup_real_mode} ne "full" and ($full_count != scalar(@my_input_backup_split) or ($$global_config_p{backup_mode} eq "differential" and $incremental_count != 0) or ($$global_config_p{backup_mode} eq "incremental" and $differencial_count != 0) ) ) { # detect (if required) if the full backup files are different from the backup data
                $$global_config_p{backup_real_mode} = "full"; # reset the backup type to FULL
                $$global_config_p{keep_last_n_files} = 0; # doing a full backup, keep only the present backup and delete all the rest
                write_file_log($global_config_p, "Warning - the backup files are out of sync with the backup list or the backup mode, forcing a full backup", \$log_body); # write the message into the script error file
            }
            # If there is a management file and we are implementing a full backup (in incremental or differential backup types),
            # erase the file to always force the full backup, even if it fails by some reason
            if ( -f $$global_config_p{backup_db_management_file} and $$global_config_p{backup_mode} ne "full" and $$global_config_p{backup_real_mode} eq "full" ) {
                write_file_log($global_config_p, "Erasing the backup management file", \$log_body); # write the message into the script error file
                unlink $$global_config_p{backup_db_management_file}; # this is done to always force a full backup
            }
            # read the backup management list (if required) to build incremental/differencial backups
            if ( $$global_config_p{backup_real_mode} ne "full" and -f "$$global_config_p{backup_db_management_file}" ) { # use a backup list file if we are not working in full backup mode
                write_file_log($global_config_p, "Reading backup management file", \$log_body); # write the message into the script error file
                if ( read_write_management_file(\%backup_previous_file_list, "not_used", "read", "not_used", "not_used", $global_config_p, \$log_body ) != 0 ) { # read the file into a hash
                    $$global_config_p{backup_real_mode} = "full"; # reset the backup type to FULL
                    $$global_config_p{keep_last_n_files} = 0; # doing a full backup, keep only the present backup and delete all the rest
                    unlink "$$global_config_p{backup_db_management_file}"; # attempt to erase the file, since it's damaged
                    write_file_log($global_config_p, "Warning - failed to read the backup management file, attempting to erase the file and forcing a full backup", \$log_body); # write the message into the script error file
                }
            }
            $session_number = get_present_session_number($ftp_connection, $global_config_p, \$log_body); # get the backup session number
            if ( $session_number == 0 ) { # check if there was no error
                write_file_log($global_config_p, "Error - unable to get a new session id number for the backup", \$log_body); # write the error message into the script error file
                $error_flag = 1; # return failure
            }
            else {
                write_file_log($global_config_p, "The backup session number is [ $session_number ]",\$log_body); # write the message into the script error file
            }
        }
    }
    if ( defined $ftp_connection and $$global_config_p{backup_working} eq "ftp" ) { # disconnect the ftp server, if connected before
        write_file_log($global_config_p, "Disconnecting from ftp server",\$log_body); # write the error message into the script error file
        $ftp_connection->quit() if $ftp_connection; # disconnect from ftp server
    }
    if ( $error_flag == 0 ) { # do the backups if no error occured

        # prepare the generic metadata date, this date is used in all backup filenames
        my @days = ('sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday');
        # Get the current time and format the hour, minutes and seconds.
        my ($sec,$min,$hour,$mday,$mon,$year,$wday) = (localtime(time))[0,1,2,3,4,5,6];
        $year += 1900; # Add 1900 to the year to get the full 4 digit year
        $mon++; # months start at 0 not at 1
        # Format the date, in this kind "Wednesday 05-03-2004"
        $mday = "0$mday" if ($mday <= 9); # add zero if required
        $mon = "0$mon" if ($mon <= 9); # add zero if required
        my $meta_date = "$days[$wday] $mday-$mon-$year";
        # Compress each backup directory and move the apropriate file into his final location
        foreach my $my_backup_file_directory( @my_input_backup_split ) { # deal with each backup directory
           my $backup_data_type = "directory"; # by default the backup data is a directory
           if ( ! -d $my_backup_file_directory ) {
               $backup_data_type = "file"; # we are doing a backup of a file
           }
           write_file_log($global_config_p, "Processing backup $backup_data_type [ $my_backup_file_directory ]",\$log_body); # write the error message into the script error file
           my $input_backup_crc32 = build_crc32("$my_backup_file_directory"); # build the crc32 number for this backup directory
           my $my_backup_file_directory_name = get_filedir_name("$my_backup_file_directory", $global_config_p); # get the backup directory name
           my $dont_backup = "no"; # do backup (compress) by default
           if ( $backup_files_count == 0 ) { # only enter here once, before doing ANY read of the backup directories
               # if there is a available the script/batch file/command declare by run_before_backup  RUN IT !
               if ( $$global_config_p{run_before_backup} ne "" ) {
                   my $run_before_backup_command = $$global_config_p{run_before_backup};
                   write_file_log($global_config_p, "Executing before backup procedure using command [ $run_before_backup_command ]", \$log_body); # write the error message into the script error file
                   system("$run_before_backup_command") == 0 or $error_flag = 1; # do backup and if we return 1 something failed
                   if ( $error_flag != 0 ) {
                       write_file_log($global_config_p, "Error - before backup procedure failed, backup will not continue", \$log_body); # write the error message into the script error file
                   }
                   else {
                       sleep 60; # sleep for 60 seconds in order to make it saffer to do backups, in windows some processes (outlook) do not free thier files impediatly after ending
                   }
               }
           }
           my $previous_directory = get_previous_filedir_name("$my_backup_file_directory", $global_config_p); # get the directory path up the previous directory of the one to backup
           # build the backup list if no error occured and if required
           my @backup_present_file_list_tmp; # this will countain the backup file list for the present backup session only
           if ( $error_flag == 0 and $$global_config_p{build_backup_list} eq "yes" ) {
                write_file_log($global_config_p, "Building backup list", \$log_body); # write the message into the script error file
                if ( build_present_backup_list("$my_backup_file_directory", "$input_backup_crc32", "$previous_directory", "$my_backup_file_directory_name", \%backup_previous_file_list, \@backup_present_file_list_tmp, $$global_config_p{host_file_separator}, $backup_data_type, 0, \%global_config, \$log_body ) != 0 ) { # read the file into a array
                    write_file_log($global_config_p, "Error - unable to build the backup list, check permissions and if the file system is not corrupted",\$log_body); # write the error message into the script error file
                    $error_flag = 1; # return failure
                }
                elsif ( scalar(@backup_present_file_list_tmp) == 0 ) { # check if the backup list is not empty
                    $dont_backup = "yes"; # abort backup
                    write_file_log($global_config_p, "Notice - there is no new data, backup will not continue for $backup_data_type [ $my_backup_file_directory ]",\$log_body); # write the error message into the script error file
                }
            }
            if ( $error_flag == 0 and $dont_backup eq "no" ) { # if backup is required for this directory and if no error occured
                # build the backup file name (if no error occured)
                if ( $$global_config_p{hostname_on_backup_files} ne "yes" ) { # if we are not including the hostname on the backup file names
                    $backup_filename_final = "$my_backup_file_directory_name." . "$session_number " . $$global_config_p{backup_real_mode} . " $meta_date crc$input_backup_crc32." . $$global_config_p{backup_format};
                    $backup_filename_temp_fortar = "_tmp_" . "$my_backup_file_directory_name." . "$session_number " . $$global_config_p{backup_real_mode} . " $meta_date crc$input_backup_crc32." . "tar";
                }
                else { # if we are including the hostname on the backup file names
                    $backup_filename_final = $$global_config_p{hostname} . ".$my_backup_file_directory_name." . "$session_number " . $$global_config_p{backup_real_mode} . " $meta_date crc$input_backup_crc32." . $$global_config_p{backup_format};
                    $backup_filename_temp_fortar = "_tmp_" . $$global_config_p{hostname} . ".$my_backup_file_directory_name." . "$session_number " . $$global_config_p{backup_real_mode} . " $meta_date crc$input_backup_crc32." . "tar";
                }
                $backup_filename_temp = "_tmp_$backup_filename_final";
                # deal with the temporary filenames with full path
                my $temporary_dir = "";
                if ( $$global_config_p{host_os} eq "unix" ) { # unix type paths
                    if ( $$global_config_p{temporary_dir} ne "/" ) { # if the temporary data is / (root) BIG mistake but it's up to the user
                        $temporary_dir = $$global_config_p{temporary_dir};
                    }
                    else {
                        $temporary_dir = "";
                    }
                }
                else { # microsoft type paths
                    $temporary_dir = $$global_config_p{temporary_dir};
                    $temporary_dir =~ s/\\$//;  # Delete trailing \ char
                }
                # now build the full path into the temporary files
                $backup_filename_temp_fpath = "$temporary_dir$my_file_separator$backup_filename_temp";
                $backup_filename_temp_fortar_fpath = "$temporary_dir$my_file_separator$backup_filename_temp_fortar";
                # Compress the directory
                my $backup_files_sumsize_current = do_compress($backup_filename_temp_fpath, $backup_filename_temp_fortar_fpath, $my_backup_file_directory_name, \@backup_present_file_list_tmp, $previous_directory, $global_config_p, \$log_body);
                if ( $backup_files_sumsize_current < 0 ) { # detect if there was a error compressing the data
                    $error_flag = 1; # mark error
                    write_file_log($global_config_p, "Error - during backup compress process, check permissions, if you have the appropriate operating system commands, if no file (in the directory to be backup) is being locked by a program or operating system, and if there is enough free space on temporary path, indicated by [ temporary_dir ]", \$log_body); # write the error message into the script error file
                }
                if ( $error_flag == 0 and $$global_config_p{build_backup_list} eq "yes" ) { # copy the present backup list into the final backup list if required
                    foreach my $my_backup_entry ( @backup_present_file_list_tmp ) {
                        push (@backup_present_file_list, $my_backup_entry);
                    }
                }
                # if the compress is ok, move the backup file into the final location
                if ( $error_flag == 0 ) {
                    # build final backup with the full path and file name
                    my $backup_path = "";
                    if ( $$global_config_p{backup_working} eq "local file system" ) { # if the directory is on the local file system
                        $backup_path = $$global_config_p{output_backup};
                    }
                    elsif ( $$global_config_p{backup_working} eq "ftp" ) { # if the directory is on the ftp server
                        $backup_path = $$global_config_p{ftp_directory};
                    }
                    $backup_path =~ s/[\\\/]$// if "$backup_path" ne "/";;  # Delete trailing \ or / char same thing kind of but $ represents end of line
                    # deal with Operating System dependend stuff and build the full bath into the backup file
                    if ( $backup_path eq "/" and ($$global_config_p{host_os} eq "unix" or $$global_config_p{backup_working} eq "ftp") ) { # unix type paths or ftp backup type
                        $backup_path = ""; # extra security to deal with / path's
                    }
                    if ( $$global_config_p{backup_working} eq "ftp" ) { # if working in FTP
                        $backup_filename_final_fpath = "$backup_path/$backup_filename_final"; # build a full path file
                    }
                    else {
                        $backup_filename_final_fpath = "$backup_path$my_file_separator$backup_filename_final"; #  build a full path file
                    }
                    if ( $error_flag == 0 ) { # if all is ok
                       if ( $backup_files_count == 0 ) { # only enter here once, before doing ANY write into the tape device
                           # check if it's required to execute any system command before writing into tape
                           if ( $$global_config_p{backup_working} eq "tape" and $$global_config_p{tape_before_backup} ne "" ) {
                               my @tape_before_backup = $$global_config_p{tape_before_backup};
                               write_file_log($global_config_p, "Executing before writing to tape using command [ @tape_before_backup ]", \$log_body); # write the error message into the script error file
                               system(@tape_before_backup) == 0 or $error_flag = 1; # execute the do before writing to tape and if we return 1 something failed
                               if ( $error_flag != 0 ) {
                                   write_file_log($global_config_p, "Error - before writing to tape procedure failed", \$log_body); # write the error message into the script error file
                               }
                           }
                        }
                        if ( $error_flag == 0 and $$global_config_p{backup_working} eq "ftp" ) { # connect into the ftp server
                            sleep 20; # sleep for 20 seconds in order to prevent a overkill of the ftp server with a series of fast login and logoffs, this  could occured if the backup data is small and we have pleanty of different directories and files to backup
                            $error_flag = ftp_connect_login(\$ftp_connection, $global_config_p, \$log_body);
                        }
                        if ( $error_flag == 0 ) { # if no error occured
                            write_file_log($global_config_p, "Moving temporary backup file [ $backup_filename_temp_fpath ] into the final backup file [ $backup_filename_final_fpath ]", \$log_body); # write the error message into the script error file
                            # move compress backup file from the temporary location into the final backup destination
                            my $move_result = move_file("$backup_filename_temp_fpath", "$backup_filename_final_fpath", $ftp_connection, $global_config_p, \$log_body);
                            if ( $move_result != 0 ) {
                                $error_flag = 1; # mark error
                                write_file_log($global_config_p, "Error - moving temporary backup file [ $backup_filename_temp_fpath ] into the final backup file [ $backup_filename_final_fpath ]", \$log_body); # write the error message into the script error file
                            }
                            else { # moving of the file into the final location is a sucess
                                $backup_files_count++; # increment the total of backup files that where created
                                push (@input_backup_done, $my_backup_file_directory); # add the backup directory into the valid done list
                                $backup_files_sumsize = $backup_files_sumsize + $backup_files_sumsize_current; # file move works increment the backup files size
                            }
                            if ( defined $ftp_connection and $$global_config_p{backup_working} eq "ftp" ) { # disconnect the ftp server, if connected before
                                write_file_log($global_config_p, "Disconnecting from ftp server",\$log_body); # write the error message into the script error file
                                $ftp_connection->quit() if $ftp_connection; # disconnect from ftp server
                            }
                            # if we are using a local db file and not reading the actual file list, PS: $move_result == 2 means that the upload into the ftp server failed and the partialy uploaded file also failed to be erase
                            if ( $$global_config_p{local_db_history_file} ne "" and ($move_result == 0 or $move_result == 2) ) { 
                                if ( ! defined read_write_history_file( $global_config_p, "$backup_filename_final", "$backup_files_sumsize", \$log_body, "add" ) ) { # add the file into the db history file
                                   $error_flag = 1; # error adding entry to history file mark error
                                }
                            }
                        }
                    }
                }
                if ( -f "$backup_filename_temp_fpath" ) { # erase the temporary file if one is still available
                    write_file_log($global_config_p, "Erasing temporary backup file [ $backup_filename_temp_fpath ]", \$log_body); # write the error message into the script error file
                    unlink "$backup_filename_temp_fpath"; # erase temporary file
                }
            }
            last if ( $error_flag != 0 ); # break backup cycle if any error occured
        }
    }
    # if we did any backup finish some stuff
    if ( $backup_files_count > 0 ) {
        if ( $$global_config_p{backup_working} eq "ftp" ) { # connect into the ftp server
            $error_flag = ftp_connect_login(\$ftp_connection, $global_config_p, \$log_body);
        }
        # write a new management file, if there is any data inside of it, and if we did a backup
        # (while in differencial or incremental mode), if the real mode is not differencial
        # and of course if there was no previous error
        if ( $error_flag == 0 and $$global_config_p{build_backup_list} eq "yes" and $$global_config_p{backup_mode} ne "full" and $$global_config_p{backup_real_mode} ne "differential" ) {
            write_file_log($global_config_p, "Writing backup management file", \$log_body); # write the message into the script error file
            if ( read_write_management_file(\%backup_previous_file_list, \@backup_present_file_list, "write", "not_used", "not_used", $global_config_p, \$log_body ) != 0 ) { # write the array into a file
                unlink "$$global_config_p{backup_db_management_file}"; # attempt to erase the file, since it's damaged
                write_file_log($global_config_p, "Warning - failed to write into the backup management file, attempting to erase the file", \$log_body); # write the message into the script error file
            }
        }
        # clear previous backups (if required)
        if ( clear_previous_backups($session_number, \@input_backup_done, $ftp_connection, $global_config_p, \$log_body) != 0 ) {
            $error_flag = 1; # mark error
        }
        # get the total size of all backups, only do this if the backup runned (with or without sucess)
        $all_backup_file_size = sum_all_backups($ftp_connection, $global_config_p, \$log_body);
        if ( $all_backup_file_size < 0 ) { # detect if there was a error reading the size of the backup files
            $error_flag = 1; # mark error
            write_file_log($global_config_p, "Error - unable to get the size of all the compressed backup files", \$log_body); # write the error message into the script error file
        }
    }
    if ( defined $ftp_connection and $$global_config_p{backup_working} eq "ftp" ) { # disconnect the ftp server, if connected before
        write_file_log($global_config_p, "Disconnecting from ftp server", \$log_body); # write the error message into the script error file
        $ftp_connection->quit() if $ftp_connection; # disconnect from ftp server
    }
    # check if there is a available the script/batch file/command declare by run_after_backup  RUN IT !
    if ( $$global_config_p{run_after_backup} ne "" ) {
        my $special_error_flag = 0;
        my $run_after_backup_command = $$global_config_p{run_after_backup};
        write_file_log($global_config_p, "Executing after backup procedure using command [ $run_after_backup_command ]", \$log_body); # write the error message into the script error file
        system($run_after_backup_command) == 0 or $special_error_flag = 1; # do backup and if we return 1 something failed
        if ( $special_error_flag > 0 ) {
            write_file_log($global_config_p, "Warning - after backup procedure failed", \$log_body); # write the error message into the script error file
        }
    }
    # check if it's required to execute any system command after writing into tape, this of course if any use of the tape was done in the first place
    if ( $backup_files_count > 0 and $$global_config_p{backup_working} eq "tape" and $$global_config_p{tape_after_backup} ne "" ) {
        my @tape_after_backup = $$global_config_p{tape_after_backup};
        my $special_error_flag = 0;
        write_file_log($global_config_p, "Executing after writing to tape using command [ @tape_after_backup ]", \$log_body); # write the error message into the script error file
        system(@tape_after_backup) == 0 or $special_error_flag = 1; # execute the do before writing to tape and if we return 1 something failed
        if ( $special_error_flag != 0 ) {
            write_file_log($global_config_p, "Warning - after writing to tape procedure failed", \$log_body); # write the error message into the script error file                        
        }
    }
    # write the final backup messages and send the e-mail report (if configured by user)
    my $my_configuration_file_name = $$global_config_p{configuration_file_name}; # get the configuration file name, something.conf  , not /path/something.conf or c:\path\something.conf
    my $my_full_date = (localtime(time())); # present date in the default format, to use in the e-mail last line message
    if ( $error_flag == 0 ) { # write the final message on the log/email report, sucess or failure, error_flag == 0 here means sucess in backup

        if ( $backup_files_count > 0 ) { # control sitiations where the backup did not start (backup rules inibit this)
            # convert the size of the backup temporary file, into kbytes and mbytes, since it's in bytes, rounding them up
            my $backup_files_sumsize_kbytes = sprintf("%.2f",$backup_files_sumsize/1024); # 1024 bytes = 1 Kbyte
            my $backup_files_sumsize_mbytes = sprintf("%.2f",$backup_files_sumsize_kbytes/1024); # 1024 kbytes = 1 Mbyte
            # write the message reporting the backup data file path, $backup_file_final is deal above inside this function, and only if we backup had sucess
            if ( $$global_config_p{backup_working} eq "local file system" ) { # if local file system mode
                write_file_log($global_config_p, "A total of [ $backup_files_count ] backup files were written and occupy [ $backup_files_sumsize_mbytes Mbytes / $backup_files_sumsize_kbytes Kbytes ]", \$log_body);
            }
            elsif ( $$global_config_p{backup_working} eq "ftp" ) { # ftp backup mode
                my $my_backup_working = $$global_config_p{backup_working}; # backup mode ftp, scp, etc
                my $my_remote_server_name = $$global_config_p{ftp_server}; # remote server name, only ftp is supported so far
                my $my_remote_server_port = $$global_config_p{ftp_server_port}; # remote server port, only ftp is supported so far
                my $my_remote_username = $$global_config_p{ftp_username}; # remote username, only ftp username is supported so far
                write_file_log($global_config_p, "A total of [ $backup_files_count ] backup files were written and occupy [ $backup_files_sumsize_mbytes Mbytes / $backup_files_sumsize_kbytes Kbytes ], located in the remote $my_backup_working server [ $my_remote_server_name:$my_remote_server_port ; username: $my_remote_username ]", \$log_body);
            }
            elsif ( $$global_config_p{backup_working} eq "tape" ) { # tape backup mode
                my $my_tape_device = $$global_config_p{tape_device}; # tape get the device
                write_file_log($global_config_p, "A total of [ $backup_files_count ] backup files were written and occupy [ $backup_files_sumsize_mbytes Mbytes / $backup_files_sumsize_kbytes Kbytes ], located in tape device [ $my_tape_device ]", \$log_body);
            }
            if ( $all_backup_file_size > 0 ) { # if there wasn't a error getting the size of the backup files and if the size if over zero
                # convert the size of all backup files, into kbytes and mbytes, since it's in bytes, rounding them up
                my $all_backup_file_size_kbytes = sprintf("%.2f",$all_backup_file_size/1024); # 1024 bytes = 1 Kbyte
                my $all_backup_file_size_mbytes = sprintf("%.2f",$all_backup_file_size_kbytes/1024); # 1024 kbytes = 1 Mbyte
                write_file_log($global_config_p, "The size of all backup sessions related to this backup configuration occupy [ $all_backup_file_size_mbytes Mbytes / $all_backup_file_size_kbytes Kbytes ]", \$log_body);
            }
        }
        if ( $$global_config_p{pass_log_to_external_cmd} ne "" ) { # if there is any available external command to accept the log
            write_file_log($global_config_p, "Calling external command and passing the backup log by the stdin, notice if the e-mail report is used this will not be reported here", \$log_body); # write the error message into the script error file
            my $ext_log_body = "$log_body" . "$my_full_date - End of Backup in ++ sucess ++\n\n";
            if ( export_log( $ext_log_body, $global_config_p ) != 0 ) { # there was a error
                my $pass_log_to_external_cmd = $$global_config_p{pass_log_to_external_cmd};
                write_file_log($global_config_p, "Warning - there was a problem calling [ $pass_log_to_external_cmd ]", \$log_body); # write the error message into the script error file
            }
        }
        if ( $$global_config_p{mail_server} ne "" ) { # if mail server is configured then we can send e-mail read_config() functions tests previoulys
            write_file_log($global_config_p, "Sending e-mail backup report", \$log_body); # write the error message into the script error file
            my $email_log_body = "$log_body" . "$my_full_date - End of Backup in ++ sucess ++\n\n";
            send_email("Simplebackup report [ $machine_hostname ; $my_configuration_file_name ; $session_number ]. Backup ended in ++ sucess ++.", $email_log_body, "simplebackup.pl v$script_version_int - $last_update_int", $global_config_p);
        }
        write_file_log($global_config_p, "End of Backup in ++ sucess ++\n\n", \$log_body); # write the error message into the script error file
    }
    else { # if the backup failed
        if ( $all_backup_file_size > 0 ) { # if there wasn't a error getting the size of the backup files and if the size if over zero
            # convert the size of all backup files, into kbytes and mbytes, since it's in bytes, rounding them up
            my $all_backup_file_size_kbytes =  sprintf("%.2f",$all_backup_file_size/1024); # 1024 bytes = 1 Kbyte
            my $all_backup_file_size_mbytes =  sprintf("%.2f",$all_backup_file_size_kbytes/1024); # 1024 kbytes = 1 Mbyte
            write_file_log($global_config_p, "The size of all backup sessions related to this configuration occupy [ $all_backup_file_size_mbytes Mbytes / $all_backup_file_size_kbytes Kbytes ]", \$log_body);
        }
        if ( $$global_config_p{pass_log_to_external_cmd} ne "" ) { # if there is any available external command to accept the log
            write_file_log($global_config_p, "Calling external command and passing the backup log by the stdin, notice if the e-mail report is used this will not be reported here", \$log_body); # write the error message into the script error file
            my $ext_log_body = "$log_body" . "$my_full_date - End of Backup in -- failure --\n\n";
            if ( export_log( $ext_log_body, $global_config_p ) != 0 ) { # there was a error
                my $pass_log_to_external_cmd = $$global_config_p{pass_log_to_external_cmd};
                write_file_log($global_config_p, "Warning - there was a problem calling [ $pass_log_to_external_cmd ]", \$log_body); # write the error message into the script error file
            }
        }
        if ( $$global_config_p{mail_server} ne "" ) { # if mail server is configured then we can send e-mail read_config() functions tests previoulys
            write_file_log($global_config_p, "Sending e-mail backup report", \$log_body); # write the error message into the script error file
            my $email_log_body = "$log_body" . "$my_full_date - End of Backup in -- failure --\n\n";
            send_email("Simplebackup report [ $machine_hostname ; $my_configuration_file_name ; $session_number ]. Backup ended in -- failure --.", $email_log_body, "simplebackup.pl v$script_version_int - $last_update_int", $global_config_p);
        }
        write_file_log($global_config_p, "End of Backup in -- failure --\n\n", \$log_body); # write the error message into the script error file
    }

    return $error_flag; # return the backup result
}


##########################
# Main
##########################

# test if the commands/parameters are correct
if ( test_commands(@ARGV) != 0 ) {
    show_commands($script_version, $script_last_update, $script_license, $script_authors, $script_email, $script_site); # show the valid commands
    exit 0; # exit from script if the arguments are wrong or there are no arguments
}

# create/edit configuration
if ( $ARGV[0] eq "--econf" ) { # create/edit configuration
    exit create_edit_config($ARGV[1], $script_version);
}

# Read the configuration file
if ( read_config($ARGV[1], \%global_config) != 0 ) {
    exit 1; # exit from script if there are any problems on the config file
}

if ( $ARGV[0] eq "--tconf" ) { # show configuration
    show_config(\%global_config);
    print "** Configuration file has passed the tests **\n";

    exit 0; # exit without erros, this is just to show the configuration not to do the backup
}

# Do the backup
if ( do_backup($script_version, $script_last_update, \%global_config) != 0 ) {
    exit 1; # backup failure
};

exit 0; # all worked
